@@ -30,6 +30,7 @@
 #include <QSize>
 #include <QPoint>
 #include <QPointF>
+#include <QPointer>
 
 class KActionCollection;
 class QRect;
@@ -47,7 +48,7 @@ class KoCanvasControllerProxyObject;
  * Flake does not provide a canvas, the application will have to
  * implement a canvas themselves. You canvas can be QWidget-based
  * or something we haven't invented yet -- as long the class that holds the canvas
- * imlements KoCanvasController, tools, scrolling and zooming will work.
+ * implements KoCanvasController, tools, scrolling and zooming will work.
  *
  * A KoCanvasController implementation acts as a decorator around the canvas widget
  * and provides a way to scroll the canvas, allows the canvas to be centered
@@ -69,16 +70,9 @@ class KoCanvasControllerProxyObject;
 class KRITAFLAKE_EXPORT KoCanvasController
 {
 public:
-    /// An enum to alter the positioning and size of the canvas inside the canvas controller
-    enum CanvasMode {
-        AlignTop,     ///< canvas is top aligned if smaller than the viewport
-        Centered,     ///< canvas is centered if smaller than the viewport
-        Infinite,     ///< canvas is never smaller than the viewport
-        Spreadsheet   ///< same as Infinite, but supports right-to-left layouts
-    };
 
     // proxy QObject: use this to connect to slots and signals.
-    KoCanvasControllerProxyObject *proxyObject;
+    QPointer<KoCanvasControllerProxyObject> proxyObject;
 
     /**
      * Constructor.
@@ -99,15 +93,6 @@ public:
      */
     virtual void setMargin(int margin);
 
-    /**
-     * Sets the how the canvas behaves if the zoomed document becomes smaller than the viewport.
-     * @param mode the new canvas mode, CanvasMode::Centered is the default value
-     */
-    virtual void setCanvasMode(KoCanvasController::CanvasMode mode);
-
-    /// Returns the current canvas mode
-    virtual KoCanvasController::CanvasMode canvasMode() const;
-
     /**
      * compatibility with QAbstractScrollArea
      */
@@ -118,15 +103,6 @@ public:
      */
     virtual QSize viewportSize() const = 0;
 
-    /**
-     * Set the shadow option -- by default the canvas controller draws
-     * a black shadow around the canvas widget, which you may or may
-     * not want.
-     *
-     * @param drawShadow if true, the shadow is drawn, if false, not
-     */
-    virtual void setDrawShadow(bool drawShadow) = 0;
-
     /**
      * Set the new canvas to be shown as a child
      * Calling this will emit canvasRemoved() if there was a canvas before, and will emit
@@ -254,6 +230,26 @@ public:
      */
     virtual void pan(const QPoint &distance) = 0;
 
+    /**
+     * Move the canvas up. This behaves the same as \sa pan() with a positive y coordinate.
+     */
+    virtual void panUp() = 0;
+
+    /**
+     * Move the canvas down. This behaves the same as \sa pan() with a negative y coordinate.
+     */
+    virtual void panDown() = 0;
+
+    /**
+     * Move the canvas to the left. This behaves the same as \sa pan() with a positive x coordinate.
+     */
+    virtual void panLeft() = 0;
+
+    /**
+     * Move the canvas to the right. This behaves the same as \sa pan() with a negative x coordinate.
+     */
+    virtual void panRight() = 0;
+
     /**
      * Get the position of the scrollbar
      */
@@ -265,6 +261,11 @@ public:
      */
     virtual void setScrollBarValue(const QPoint &value) = 0;
 
+    /**
+     * Update the range of scroll bars
+     */
+    virtual void resetScrollBars() = 0;
+
     /**
      * Called when the size of your document in view coordinates (pixels) changes, for instance when zooming.
      *
@@ -292,10 +293,10 @@ public:
     virtual void setVastScrolling(qreal factor) = 0;
 
    /**
-     * Returns the action collection for the canvas
-     * @returns action collection for this canvas, can be 0
+     * Returns the action collection for the window
+     * @returns action collection for this window, can be 0
      */
-    virtual KActionCollection* actionCollection() const;
+    KActionCollection* actionCollection() const;
 
     QPoint documentOffset() const;
 
@@ -449,7 +450,6 @@ public:
 
     void scrollContentsBy(int /*dx*/, int /*dy*/) override {}
     QSize viewportSize() const override { return QSize(); }
-    void setDrawShadow(bool /*drawShadow*/) override {}
     void setCanvas(KoCanvasBase *canvas) override {Q_UNUSED(canvas)}
     KoCanvasBase *canvas() const override {return 0;}
     int visibleHeight() const override {return 0;}
@@ -466,13 +466,17 @@ public:
     void setPreferredCenter(const QPointF &/*viewPoint*/) override {}
     QPointF preferredCenter() const override {return QPointF();}
     void pan(const QPoint &/*distance*/) override {}
+    void panUp() override {}
+    void panDown() override {}
+    void panLeft() override {}
+    void panRight() override {}
     QPoint scrollBarValue() const override {return QPoint();}
     void setScrollBarValue(const QPoint &/*value*/) override {}
+    void resetScrollBars() override {}
     void updateDocumentSize(const QSize &/*sz*/, bool /*recalculateCenter*/) override {}
     void setZoomWithWheel(bool /*zoom*/) override {}
     void setVastScrolling(qreal /*factor*/) override {}
     QPointF currentCursorPosition() const override { return QPointF(); }
-
 };
 
 #endif
