@@ -54,6 +54,7 @@
 #include <KoClipPath.h>
 #include <KoClipMask.h>
 #include <KoXmlNS.h>
+#include <QXmlSimpleReader>
 
 #include "SvgUtil.h"
 #include "SvgShape.h"
@@ -66,8 +67,71 @@
 #include <KoVectorPatternBackground.h>
 #include <KoMarker.h>
 
+#include <text/KoSvgTextShape.h>
+#include <text/KoSvgTextChunkShape.h>
+
+#include "kis_dom_utils.h"
+
+#include "kis_algebra_2d.h"
 #include "kis_debug.h"
 #include "kis_global.h"
+#include <algorithm>
+
+
+struct SvgParser::DeferredUseStore {
+    struct El {
+        El(const KoXmlElement* ue, const QString& key) :
+            m_useElement(ue), m_key(key) {
+        }
+        const KoXmlElement* m_useElement;
+        QString m_key;
+    };
+    DeferredUseStore(SvgParser* p) :
+        m_parse(p) {
+    }
+
+    void add(const KoXmlElement* useE, const QString& key) {
+        m_uses.push_back(El(useE, key));
+    }
+    bool empty() const {
+        return m_uses.empty();
+    }
+
+    void checkPendingUse(const KoXmlElement &b, QList<KoShape*>& shapes) {
+        KoShape* shape = 0;
+        const QString id = b.attribute("id");
+
+        if (id.isEmpty())
+            return;
+
+        // debugFlake << "Checking id: " << id;
+        auto i = std::partition(m_uses.begin(), m_uses.end(),
+                                [&](const El& e) -> bool {return e.m_key != id;});
+
+        while (i != m_uses.end()) {
+            const El& el = m_uses.back();
+            if (m_parse->m_context.hasDefinition(el.m_key)) {
+                // debugFlake << "Found pending use for id: " << el.m_key;
+                shape = m_parse->resolveUse(*(el.m_useElement), el.m_key);
+                if (shape) {
+                    shapes.append(shape);
+                }
+            }
+            m_uses.pop_back();
+        }
+    }
+
+    ~DeferredUseStore() {
+        while (!m_uses.empty()) {
+            const El& el = m_uses.back();
+            debugFlake << "WARNING: could not find path in <use xlink:href=\"#xxxxx\" expression. Losing data here. Key:"
+                       << el.m_key;
+            m_uses.pop_back();
+        }
+    }
+    SvgParser* m_parse;
+    std::vector<El> m_uses;
+};
 
 
 SvgParser::SvgParser(KoDocumentResourceManager *documentResourceManager)
@@ -81,6 +145,44 @@ SvgParser::~SvgParser()
     qDeleteAll(m_symbols);
 }
 
+KoXmlDocument SvgParser::createDocumentFromSvg(QIODevice *device, QString *errorMsg, int *errorLine, int *errorColumn)
+{
+    QXmlInputSource source(device);
+    return createDocumentFromSvg(&source, errorMsg, errorLine, errorColumn);
+}
+
+KoXmlDocument SvgParser::createDocumentFromSvg(const QByteArray &data, QString *errorMsg, int *errorLine, int *errorColumn)
+{
+    QXmlInputSource source;
+    source.setData(data);
+
+    return createDocumentFromSvg(&source, errorMsg, errorLine, errorColumn);
+}
+
+KoXmlDocument SvgParser::createDocumentFromSvg(const QString &data, QString *errorMsg, int *errorLine, int *errorColumn)
+{
+    QXmlInputSource source;
+    source.setData(data);
+
+    return createDocumentFromSvg(&source, errorMsg, errorLine, errorColumn);
+}
+
+KoXmlDocument SvgParser::createDocumentFromSvg(QXmlInputSource *source, QString *errorMsg, int *errorLine, int *errorColumn)
+{
+    // we should read all spaces to parse text node correctly
+    QXmlSimpleReader reader;
+    reader.setFeature("http://qt-project.org/xml/features/report-whitespace-only-CharData", true);
+    reader.setFeature("http://xml.org/sax/features/namespaces", false);
+    reader.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
+
+    QDomDocument doc;
+    if (!doc.setContent(source, &reader, errorMsg, errorLine, errorColumn)) {
+        return QDomDocument();
+    }
+
+    return doc;
+}
+
 void SvgParser::setXmlBaseDir(const QString &baseDir)
 {
     m_context.setInitialXmlBaseDir(baseDir);
@@ -89,7 +191,9 @@ void SvgParser::setXmlBaseDir(const QString &baseDir)
         [this](const QString &name) {
             const QString fileName = m_context.xmlBaseDir() + QDir::separator() + name;
             QFile file(fileName);
-            KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(file.exists(), QByteArray());
+            if (!file.exists()) {
+                return QByteArray();
+            }
             file.open(QIODevice::ReadOnly);
             return file.readAll();
         });
@@ -99,6 +203,7 @@ void SvgParser::setResolution(const QRectF boundsInPixels, qreal pixelsPerInch)
 {
     KIS_ASSERT(!m_context.currentGC());
     m_context.pushGraphicsContext();
+    m_context.currentGC()->isResolutionFrame = true;
     m_context.currentGC()->pixelsPerInch = pixelsPerInch;
 
     const qreal scale = 72.0 / pixelsPerInch;
@@ -107,6 +212,13 @@ void SvgParser::setResolution(const QRectF boundsInPixels, qreal pixelsPerInch)
     m_context.currentGC()->matrix = t;
 }
 
+void SvgParser::setForcedFontSizeResolution(qreal value)
+{
+    if (qFuzzyCompare(value, 0.0)) return;
+
+    m_context.currentGC()->forcedFontSizeCoeff = 72.0 / value;
+}
+
 QList<KoShape*> SvgParser::shapes() const
 {
     return m_shapes;
@@ -306,7 +418,7 @@ SvgGradientHelper* SvgParser::parseGradient(const KoXmlElement &e)
         }
         gradHelper.setGradient(g);
     } else {
-        qDebug() << "WARNING: Failed to parse gradient with tag" << e.tagName();
+        debugFlake << "WARNING: Failed to parse gradient with tag" << e.tagName();
     }
 
     // handle spread method
@@ -415,7 +527,7 @@ QSharedPointer<KoVectorPatternBackground> SvgParser::parsePattern(const KoXmlEle
     }
 
     /**
-     * In Krita shapes X,Y coordinates are baked into the the shape global transform, but
+     * In Krita shapes X,Y coordinates are baked into the shape global transform, but
      * the pattern should be painted in "user" coordinates. Therefore, we should handle
      * this offfset separately.
      *
@@ -594,7 +706,7 @@ bool SvgParser::parseSymbol(const KoXmlElement &e)
 
     if (id.isEmpty()) return false;
 
-    KoSvgSymbol *svgSymbol = new KoSvgSymbol();
+    QScopedPointer<KoSvgSymbol> svgSymbol(new KoSvgSymbol());
 
     // ensure that the clip path is loaded in local coordinates system
     m_context.pushGraphicsContext(e, false);
@@ -603,24 +715,23 @@ bool SvgParser::parseSymbol(const KoXmlElement &e)
 
     QString title = e.firstChildElement("title").toElement().text();
 
-    KoShape *symbolShape = parseGroup(e);
+    QScopedPointer<KoShape> symbolShape(parseGroup(e));
 
     m_context.popGraphicsContext();
 
     if (!symbolShape) return false;
 
-    svgSymbol->shape = symbolShape;
+    svgSymbol->shape = symbolShape.take();
     svgSymbol->title = title;
     svgSymbol->id = id;
     if (title.isEmpty()) svgSymbol->title = id;
 
     if (svgSymbol->shape->boundingRect() == QRectF(0.0, 0.0, 0.0, 0.0)) {
-        warnFlake << "Symbol" << id << "seems to be empty, discarding";
-        delete svgSymbol;
+        debugFlake << "Symbol" << id << "seems to be empty, discarding";
         return false;
     }
 
-    m_symbols << svgSymbol;
+    m_symbols << svgSymbol.take();
 
     return true;
 }
@@ -708,13 +819,7 @@ void SvgParser::applyCurrentStyle(KoShape *shape, const QPointF &shapeToOriginal
 {
     if (!shape) return;
 
-    SvgGraphicsContext *gc = m_context.currentGC();
-    KIS_ASSERT(gc);
-
-    if (!dynamic_cast<KoShapeGroup*>(shape)) {
-        applyFillStyle(shape);
-        applyStrokeStyle(shape);
-    }
+    applyCurrentBasicStyle(shape);
 
     if (KoPathShape *pathShape = dynamic_cast<KoPathShape*>(shape)) {
         applyMarkers(pathShape);
@@ -724,6 +829,20 @@ void SvgParser::applyCurrentStyle(KoShape *shape, const QPointF &shapeToOriginal
     applyClipping(shape, shapeToOriginalUserCoordinates);
     applyMaskClipping(shape, shapeToOriginalUserCoordinates);
 
+}
+
+void SvgParser::applyCurrentBasicStyle(KoShape *shape)
+{
+    if (!shape) return;
+
+    SvgGraphicsContext *gc = m_context.currentGC();
+    KIS_ASSERT(gc);
+
+    if (!dynamic_cast<KoShapeGroup*>(shape)) {
+        applyFillStyle(shape);
+        applyStrokeStyle(shape);
+    }
+
     if (!gc->display || !gc->visible) {
         /**
          * WARNING: here is a small inconsistency with the standard:
@@ -740,6 +859,7 @@ void SvgParser::applyCurrentStyle(KoShape *shape, const QPointF &shapeToOriginal
     shape->setTransparency(1.0 - gc->opacity);
 }
 
+
 void SvgParser::applyStyle(KoShape *obj, const KoXmlElement &e, const QPointF &shapeToOriginalUserCoordinates)
 {
     applyStyle(obj, m_context.styleParser().collectStyles(e), shapeToOriginalUserCoordinates);
@@ -818,6 +938,46 @@ QGradient* prepareGradientForShape(const SvgGradientHelper *gradient,
 
             resultGradient = KoFlake::cloneGradient(gradient->gradient());
             *transform = gradient->transform() * gc->matrix * shape->transformation().inverted();
+
+            const QRectF outlineRect = shape->outlineRect();
+            if (outlineRect.isEmpty()) return resultGradient;
+
+            /**
+             * If shape outline rect is valid, convert the gradient into OBB mode by
+             * doing some magic conversions: we compensate non-uniform size of the shape
+             * by applying an additional pre-transform
+             */
+
+            QRadialGradient *rgradient = static_cast<QRadialGradient*>(resultGradient);
+
+            const qreal maxDimension = KisAlgebra2D::maxDimension(outlineRect);
+            const QRectF uniformSize(outlineRect.topLeft(), QSizeF(maxDimension, maxDimension));
+
+            const QTransform uniformizeTransform =
+                    QTransform::fromTranslate(-outlineRect.x(), -outlineRect.y()) *
+                    QTransform::fromScale(maxDimension / shape->outlineRect().width(),
+                                          maxDimension / shape->outlineRect().height()) *
+                    QTransform::fromTranslate(outlineRect.x(), outlineRect.y());
+
+            const QPointF centerLocal = transform->map(rgradient->center());
+            const QPointF focalLocal = transform->map(rgradient->focalPoint());
+
+            const QPointF centerOBB = KisAlgebra2D::absoluteToRelative(centerLocal, uniformSize);
+            const QPointF focalOBB = KisAlgebra2D::absoluteToRelative(focalLocal, uniformSize);
+
+            rgradient->setCenter(centerOBB);
+            rgradient->setFocalPoint(focalOBB);
+
+            const qreal centerRadiusOBB = KisAlgebra2D::absoluteToRelative(rgradient->centerRadius(), uniformSize);
+            const qreal focalRadiusOBB = KisAlgebra2D::absoluteToRelative(rgradient->focalRadius(), uniformSize);
+
+            rgradient->setCenterRadius(centerRadiusOBB);
+            rgradient->setFocalRadius(focalRadiusOBB);
+
+            rgradient->setCoordinateMode(QGradient::ObjectBoundingMode);
+
+            // Warning: should it really be pre-multiplication?
+            *transform = uniformizeTransform * gradient->transform();
         }
     }
 
@@ -939,7 +1099,7 @@ void SvgParser::applyFilter(KoShape *shape)
 
     // parse filter region
     QRectF bound(shape->position(), shape->size());
-    // work on bounding box without viewbox tranformation applied
+    // work on bounding box without viewbox transformation applied
     // so user space coordinates of bounding box and filter region match up
     bound = gc->viewboxTransform.inverted().mapRect(bound);
 
@@ -1129,40 +1289,50 @@ void SvgParser::applyMaskClipping(KoShape *shape, const QPointF &shapeToOriginal
     shape->setClipMask(clipMask);
 }
 
-KoShape* SvgParser::parseUse(const KoXmlElement &e)
+KoShape* SvgParser::parseUse(const KoXmlElement &e, DeferredUseStore* deferredUseStore)
+{
+    QString href = e.attribute("xlink:href");
+    if (href.isEmpty())
+        return 0;
+
+    QString key = href.mid(1);
+    const bool gotDef = m_context.hasDefinition(key);
+    if (gotDef) {
+        return resolveUse(e, key);
+    } else if (deferredUseStore) {
+        deferredUseStore->add(&e, key);
+        return 0;
+    }
+    debugFlake << "WARNING: Did not find reference for svg 'use' element. Skipping. Id: "
+             << key;
+    return 0;
+}
+
+KoShape* SvgParser::resolveUse(const KoXmlElement &e, const QString& key)
 {
     KoShape *result = 0;
 
-    QString id = e.attribute("xlink:href");
-    if (!id.isEmpty()) {
+    SvgGraphicsContext *gc = m_context.pushGraphicsContext(e);
 
-        QString key = id.mid(1);
-        if (m_context.hasDefinition(key)) {
+    // TODO: parse 'width' and 'height' as well
+    gc->matrix.translate(parseUnitX(e.attribute("x", "0")), parseUnitY(e.attribute("y", "0")));
 
-            SvgGraphicsContext *gc = m_context.pushGraphicsContext(e);
-
-            // TODO: parse 'width' and 'hight' as well
-            gc->matrix.translate(parseUnitX(e.attribute("x", "0")), parseUnitY(e.attribute("y", "0")));
-
-            const KoXmlElement &referencedElement = m_context.definition(key);
-            result = parseGroup(e, referencedElement);
-
-            m_context.popGraphicsContext();
-        }
-    }
+    const KoXmlElement &referencedElement = m_context.definition(key);
+    result = parseGroup(e, referencedElement);
 
+    m_context.popGraphicsContext();
     return result;
 }
 
-void SvgParser::addToGroup(QList<KoShape*> shapes, KoShapeGroup *group)
+void SvgParser::addToGroup(QList<KoShape*> shapes, KoShapeContainer *group)
 {
     m_shapes += shapes;
 
     if (!group || shapes.isEmpty())
         return;
 
-    // not clipped, but inherit transform
-    KoShapeGroupCommand cmd(group, shapes, false, true, false);
+    // not normalized
+    KoShapeGroupCommand cmd(group, shapes, false);
     cmd.redo();
 }
 
@@ -1178,8 +1348,33 @@ QList<KoShape*> SvgParser::parseSvg(const KoXmlElement &e, QSizeF *fragmentSize)
 
     const QString w = e.attribute("width");
     const QString h = e.attribute("height");
-    const qreal width = w.isEmpty() ? 666.0 : parseUnitX(w);
-    const qreal height = h.isEmpty() ? 555.0 : parseUnitY(h);
+
+    qreal width = w.isEmpty() ? 666.0 : parseUnitX(w);
+    qreal height = h.isEmpty() ? 555.0 : parseUnitY(h);
+
+    if (w.isEmpty() || h.isEmpty()) {
+        QRectF viewRect;
+        QTransform viewTransform_unused;
+        QRectF fakeBoundingRect(0.0, 0.0, 1.0, 1.0);
+
+        if (SvgUtil::parseViewBox(e, fakeBoundingRect,
+                                  &viewRect, &viewTransform_unused)) {
+
+            QSizeF estimatedSize = viewRect.size();
+
+            if (estimatedSize.isValid()) {
+
+                if (!w.isEmpty()) {
+                    estimatedSize = QSizeF(width, width * estimatedSize.height() / estimatedSize.width());
+                } else if (!h.isEmpty()) {
+                    estimatedSize = QSizeF(height * estimatedSize.width() / estimatedSize.height(), height);
+                }
+
+                width = estimatedSize.width();
+                height = estimatedSize.height();
+            }
+        }
+    }
 
     QSizeF svgFragmentSize(QSizeF(width, height));
 
@@ -1238,7 +1433,7 @@ void SvgParser::applyViewBoxTransform(const KoXmlElement &element)
     QRectF viewRect = gc->currentBoundingBox;
     QTransform viewTransform;
 
-    if (SvgUtil::parseViewBox(gc, element, gc->currentBoundingBox,
+    if (SvgUtil::parseViewBox(element, gc->currentBoundingBox,
                               &viewRect, &viewTransform)) {
 
         gc->matrix = viewTransform * gc->matrix;
@@ -1294,7 +1489,7 @@ KoShape* SvgParser::parseGroup(const KoXmlElement &b, const KoXmlElement &overri
     if (!overrideChildrenFrom.isNull()) {
         // we upload styles from both: <use> and <defs>
         uploadStyleToContext(overrideChildrenFrom);
-        childShapes = parseSingleElement(overrideChildrenFrom);
+        childShapes = parseSingleElement(overrideChildrenFrom, 0);
     } else {
         childShapes = parseContainer(b);
     }
@@ -1311,17 +1506,118 @@ KoShape* SvgParser::parseGroup(const KoXmlElement &b, const KoXmlElement &overri
     return group;
 }
 
-QList<KoShape*> SvgParser::parseContainer(const KoXmlElement &e)
+KoShape* SvgParser::parseTextNode(const KoXmlText &e)
+{
+    QScopedPointer<KoSvgTextChunkShape> textChunk(new KoSvgTextChunkShape());
+    textChunk->setZIndex(m_context.nextZIndex());
+
+    if (!textChunk->loadSvgTextNode(e, m_context)) {
+        return 0;
+    }
+
+    textChunk->applyAbsoluteTransformation(m_context.currentGC()->matrix);
+    applyCurrentBasicStyle(textChunk.data()); // apply style to this group after size is set
+
+    return textChunk.take();
+}
+
+KoXmlText getTheOnlyTextChild(const KoXmlElement &e)
+{
+    KoXmlNode firstChild = e.firstChild();
+    return !firstChild.isNull() && firstChild == e.lastChild() && firstChild.isText() ?
+                firstChild.toText() : KoXmlText();
+}
+
+KoShape *SvgParser::parseTextElement(const KoXmlElement &e, KoSvgTextShape *mergeIntoShape)
+{
+    KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(e.tagName() == "text" || e.tagName() == "tspan", 0);
+    KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(m_isInsideTextSubtree || e.tagName() == "text", 0);
+    KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(e.tagName() == "text" || !mergeIntoShape, 0);
+
+    KoSvgTextShape *rootTextShape  = 0;
+
+    if (e.tagName() == "text") {
+        // XXX: Shapes need to be created by their factories
+        if (mergeIntoShape) {
+            rootTextShape = mergeIntoShape;
+        } else {
+            rootTextShape = new KoSvgTextShape();
+            const QString useRichText = e.attribute("krita:useRichText", "true");
+            rootTextShape->setRichTextPreferred(useRichText != "false");
+        }
+    }
+
+    if (rootTextShape) {
+        m_isInsideTextSubtree = true;
+    }
+
+    m_context.pushGraphicsContext(e);
+    uploadStyleToContext(e);
+
+    KoSvgTextChunkShape *textChunk = rootTextShape ? rootTextShape : new KoSvgTextChunkShape();
+    textChunk->setZIndex(m_context.nextZIndex());
+
+    textChunk->loadSvg(e, m_context);
+
+    // 1) apply transformation only in case we are not overriding the shape!
+    // 2) the transformation should be applied *before* the shape is added to the group!
+    if (!mergeIntoShape) {
+        // groups should also have their own coordinate system!
+        textChunk->applyAbsoluteTransformation(m_context.currentGC()->matrix);
+        const QPointF extraOffset = extraShapeOffset(textChunk, m_context.currentGC()->matrix);
+
+        // handle id
+        applyId(e.attribute("id"), textChunk);
+        applyCurrentStyle(textChunk, extraOffset); // apply style to this group after size is set
+    } else {
+        m_context.currentGC()->matrix = mergeIntoShape->absoluteTransformation(0);
+        applyCurrentBasicStyle(textChunk);
+    }
+
+    KoXmlText onlyTextChild = getTheOnlyTextChild(e);
+    if (!onlyTextChild.isNull()) {
+        textChunk->loadSvgTextNode(onlyTextChild, m_context);
+    } else {
+        QList<KoShape*> childShapes = parseContainer(e, true);
+        addToGroup(childShapes, textChunk);
+    }
+
+    m_context.popGraphicsContext();
+
+    textChunk->normalizeCharTransformations();
+
+    if (rootTextShape) {
+        textChunk->simplifyFillStrokeInheritance();
+
+        m_isInsideTextSubtree = false;
+        rootTextShape->relayout();
+    }
+
+    return textChunk;
+}
+
+QList<KoShape*> SvgParser::parseContainer(const KoXmlElement &e, bool parseTextNodes)
 {
     QList<KoShape*> shapes;
 
     // are we parsing a switch container
     bool isSwitch = e.tagName() == "switch";
 
+    DeferredUseStore deferredUseStore(this);
+
     for (KoXmlNode n = e.firstChild(); !n.isNull(); n = n.nextSibling()) {
         KoXmlElement b = n.toElement();
-        if (b.isNull())
+        if (b.isNull()) {
+            if (parseTextNodes && n.isText()) {
+                KoShape *shape = parseTextNode(n.toText());
+
+                if (shape) {
+                    shapes += shape;
+                }
+            }
+
             continue;
+        }
 
         if (isSwitch) {
             // if we are parsing a switch check the requiredFeatures, requiredExtensions
@@ -1335,27 +1631,35 @@ QList<KoShape*> SvgParser::parseContainer(const KoXmlElement &e)
                 continue;
             }
             if (b.hasAttribute("systemLanguage")) {
-                // not implemeted yet
+                // not implemented yet
             }
         }
 
-        QList<KoShape*> currentShapes = parseSingleElement(b);
+        QList<KoShape*> currentShapes = parseSingleElement(b, &deferredUseStore);
         shapes.append(currentShapes);
 
         // if we are parsing a switch, stop after the first supported element
         if (isSwitch && !currentShapes.isEmpty())
             break;
     }
-
     return shapes;
 }
 
-QList<KoShape*> SvgParser::parseSingleElement(const KoXmlElement &b)
+void SvgParser::parseDefsElement(const KoXmlElement &e)
+{
+    KIS_SAFE_ASSERT_RECOVER_RETURN(e.tagName() == "defs");
+    parseSingleElement(e);
+}
+
+QList<KoShape*> SvgParser::parseSingleElement(const KoXmlElement &b, DeferredUseStore* deferredUseStore)
 {
     QList<KoShape*> shapes;
 
     // save definition for later instantiation with 'use'
     m_context.addDefinition(b);
+    if (deferredUseStore) {
+        deferredUseStore->checkPendingUse(b, shapes);
+    }
 
     if (b.tagName() == "svg") {
         shapes += parseSvg(b);
@@ -1375,7 +1679,8 @@ QList<KoShape*> SvgParser::parseSingleElement(const KoXmlElement &b)
              */
             KoShape *defsShape = parseGroup(b);
             defsShape->setVisible(false);
-            delete defsShape;
+            m_defsShapes << defsShape; // TODO: where to delete the shape!?
+
         }
     } else if (b.tagName() == "linearGradient" || b.tagName() == "radialGradient") {
     } else if (b.tagName() == "pattern") {
@@ -1391,6 +1696,10 @@ QList<KoShape*> SvgParser::parseSingleElement(const KoXmlElement &b)
         parseSymbol(b);
     } else if (b.tagName() == "style") {
         m_context.addStyleSheet(b);
+    } else if (b.tagName() == "text" ||
+               b.tagName() == "tspan") {
+
+        shapes += parseTextElement(b);
     } else if (b.tagName() == "rect" ||
                b.tagName() == "ellipse" ||
                b.tagName() == "circle" ||
@@ -1398,13 +1707,15 @@ QList<KoShape*> SvgParser::parseSingleElement(const KoXmlElement &b)
                b.tagName() == "polyline" ||
                b.tagName() == "polygon" ||
                b.tagName() == "path" ||
-               b.tagName() == "image" ||
-               b.tagName() == "text") {
+               b.tagName() == "image") {
         KoShape *shape = createObjectDirect(b);
         if (shape)
             shapes.append(shape);
     } else if (b.tagName() == "use") {
-        shapes += parseUse(b);
+        KoShape* s = parseUse(b, deferredUseStore);
+        if (s) {
+            shapes += s;
+        }
     } else if (b.tagName() == "color-profile") {
         m_context.parseProfile(b);
     } else {
@@ -1444,18 +1755,14 @@ KoShape * SvgParser::createPath(const KoXmlElement &element)
             path->clear();
 
             bool bFirst = true;
-            QString points = element.attribute("points").simplified();
-            points.replace(',', ' ');
-            points.remove('\r');
-            points.remove('\n');
-            QStringList pointList = points.split(' ', QString::SkipEmptyParts);
+            QStringList pointList = SvgUtil::simplifyList(element.attribute("points"));
             for (QStringList::Iterator it = pointList.begin(); it != pointList.end(); ++it) {
                 QPointF point;
-                point.setX(SvgUtil::fromUserSpace((*it).toDouble()));
+                point.setX(SvgUtil::fromUserSpace(KisDomUtils::toDouble(*it)));
                 ++it;
                 if (it == pointList.end())
                     break;
-                point.setY(SvgUtil::fromUserSpace((*it).toDouble()));
+                point.setY(SvgUtil::fromUserSpace(KisDomUtils::toDouble(*it)));
                 if (bFirst) {
                     path->moveTo(point);
                     bFirst = false;
@@ -1596,14 +1903,15 @@ KoShape *SvgParser::createShape(const QString &shapeID)
         debugFlake << "Could not create Default shape for shape id" << shapeID;
         return 0;
     }
-    if (shape->shapeId().isEmpty())
+    if (shape->shapeId().isEmpty()) {
         shape->setShapeId(factory->id());
+    }
 
-    // reset tranformation that might come from the default shape
+    // reset transformation that might come from the default shape
     shape->setTransformation(QTransform());
 
     // reset border
-    KoShapeStrokeModelSP oldStroke = shape->stroke();
+    // ??? KoShapeStrokeModelSP oldStroke = shape->stroke();
     shape->setStroke(KoShapeStrokeModelSP());
 
     // reset fill
