@@ -22,6 +22,7 @@
 
 #include <QUuid>
 #include <KoColorSpaceConstants.h>
+#include <KoProperties.h>
 
 #include "kis_painter.h"
 #include "kis_image.h"
@@ -39,17 +40,20 @@
 #include "commands/kis_image_layer_move_command.h"
 #include "commands/kis_image_change_layers_command.h"
 #include "commands_new/kis_activate_selection_mask_command.h"
+#include "commands/kis_image_change_visibility_command.h"
 #include "kis_abstract_projection_plane.h"
 #include "kis_processing_applicator.h"
 #include "kis_image_animation_interface.h"
 #include "kis_keyframe_channel.h"
 #include "kis_command_utils.h"
-#include "kis_processing_applicator.h"
 #include "commands_new/kis_change_projection_color_command.h"
 #include "kis_layer_properties_icons.h"
 #include "lazybrush/kis_colorize_mask.h"
 #include "commands/kis_node_property_list_command.h"
+#include "commands/kis_node_compositeop_command.h"
 #include <KisDelayedUpdateNodeInterface.h>
+#include "krita_utils.h"
+#include "kis_image_signal_router.h"
 
 
 namespace KisLayerUtils {
@@ -83,6 +87,8 @@ namespace KisLayerUtils {
 
         SwitchFrameCommand::SharedStorageSP storage;
         QSet<int> frames;
+        bool useInTimeline = false;
+        bool enableOnionSkins = false;
 
         virtual KisNodeList allSrcNodes() = 0;
 
@@ -102,6 +108,14 @@ namespace KisLayerUtils {
             frames =
                 fetchLayerFramesRecursive(prevLayer) |
                 fetchLayerFramesRecursive(currLayer);
+
+            useInTimeline = prevLayer->useInTimeline() || currLayer->useInTimeline();
+
+            const KisPaintLayer *paintLayer = qobject_cast<KisPaintLayer*>(currLayer.data());
+            if (paintLayer) enableOnionSkins |= paintLayer->onionSkinEnabled();
+
+            paintLayer = qobject_cast<KisPaintLayer*>(prevLayer.data());
+            if (paintLayer) enableOnionSkins |= paintLayer->onionSkinEnabled();
         }
 
         KisLayerSP prevLayer;
@@ -123,10 +137,17 @@ namespace KisLayerUtils {
         {
             foreach (KisNodeSP node, mergedNodes) {
                 frames |= fetchLayerFramesRecursive(node);
+                useInTimeline |= node->useInTimeline();
+
+                const KisPaintLayer *paintLayer = qobject_cast<KisPaintLayer*>(node.data());
+                if (paintLayer) {
+                    enableOnionSkins |= paintLayer->onionSkinEnabled();
+                }
             }
         }
 
         KisNodeList mergedNodes;
+        bool nodesCompositingVaries = false;
 
         KisNodeList allSrcNodes() override {
             return mergedNodes;
@@ -197,6 +218,44 @@ namespace KisLayerUtils {
         MergeDownInfoBaseSP m_info;
     };
 
+    struct DisableExtraCompositing : public KisCommandUtils::AggregateCommand {
+        DisableExtraCompositing(MergeMultipleInfoSP info) : m_info(info) {}
+
+        void populateChildCommands() override {
+            /**
+             * We disable extra compositing only in case all the layers have
+             * the same compositing properties, therefore, we can just sum them using
+             * Normal blend mode
+             */
+            if (m_info->nodesCompositingVaries) return;
+
+            // we should disable dirty requests on **redo only**, otherwise
+            // the state of the layers will not be recovered on undo
+            m_info->image->disableDirtyRequests();
+
+            Q_FOREACH (KisNodeSP node, m_info->allSrcNodes()) {
+                if (node->compositeOpId() != COMPOSITE_OVER) {
+                    addCommand(new KisNodeCompositeOpCommand(node, node->compositeOpId(), COMPOSITE_OVER));
+                }
+
+                if (KisLayerPropertiesIcons::nodeProperty(node, KisLayerPropertiesIcons::inheritAlpha, false).toBool()) {
+
+                    KisBaseNode::PropertyList props = node->sectionModelProperties();
+                    KisLayerPropertiesIcons::setNodeProperty(&props,
+                                                             KisLayerPropertiesIcons::inheritAlpha,
+                                                             false);
+
+                    addCommand(new KisNodePropertyListCommand(node, props));
+                }
+            }
+
+            m_info->image->enableDirtyRequests();
+        }
+
+    private:
+        MergeMultipleInfoSP m_info;
+    };
+
     struct DisablePassThroughForHeadsOnly : public KisCommandUtils::AggregateCommand {
         DisablePassThroughForHeadsOnly(MergeDownInfoBaseSP info, bool skipIfDstIsGroup = false)
             : m_info(info),
@@ -281,14 +340,7 @@ namespace KisLayerUtils {
 
         void redo() override {
             foreach (KisNodeSP node, m_info->allSrcNodes()) {
-                recursiveApplyNodes(node,
-                    [] (KisNodeSP node) {
-                        KisDelayedUpdateNodeInterface *delayedUpdate =
-                            dynamic_cast<KisDelayedUpdateNodeInterface*>(node.data());
-                        if (delayedUpdate) {
-                            delayedUpdate->forceUpdateTimedNode();
-                        }
-                    });
+                forceAllDelayedNodesUpdate(node);
             }
         }
 
@@ -346,13 +398,20 @@ namespace KisLayerUtils {
         CreateMergedLayer(MergeDownInfoSP info) : m_info(info) {}
 
         void populateChildCommands() override {
-            // actual merging done by KisLayer::createMergedLayer (or specialized decendant)
+            // actual merging done by KisLayer::createMergedLayer (or specialized descendant)
             m_info->dstNode = m_info->currLayer->createMergedLayerTemplate(m_info->prevLayer);
 
             if (m_info->frames.size() > 0) {
                 m_info->dstNode->enableAnimation();
                 m_info->dstNode->getKeyframeChannel(KisKeyframeChannel::Content.id(), true);
             }
+
+            m_info->dstNode->setUseInTimeline(m_info->useInTimeline);
+
+            KisPaintLayer *dstPaintLayer = qobject_cast<KisPaintLayer*>(m_info->dstNode.data());
+            if (dstPaintLayer) {
+                dstPaintLayer->setOnionSkinEnabled(m_info->enableOnionSkins);
+            }
         }
 
     private:
@@ -366,7 +425,6 @@ namespace KisLayerUtils {
 
         void populateChildCommands() override {
             QString mergedLayerName;
-
             if (m_name.isEmpty()){
                 const QString mergedLayerSuffix = i18n("Merged");
                 mergedLayerName = m_info->mergedNodes.first()->name();
@@ -379,21 +437,32 @@ namespace KisLayerUtils {
                 mergedLayerName = m_name;
             }
 
-            m_info->dstNode = new KisPaintLayer(m_info->image, mergedLayerName, OPACITY_OPAQUE_U8);
+            KisPaintLayer *dstPaintLayer = new KisPaintLayer(m_info->image, mergedLayerName, OPACITY_OPAQUE_U8);
+            m_info->dstNode = dstPaintLayer;
 
             if (m_info->frames.size() > 0) {
                 m_info->dstNode->enableAnimation();
                 m_info->dstNode->getKeyframeChannel(KisKeyframeChannel::Content.id(), true);
             }
 
+
+            auto channelFlagsLazy = [](KisNodeSP node) {
+                KisLayer *layer = dynamic_cast<KisLayer*>(node.data());
+                return layer ? layer->channelFlags() : QBitArray();
+            };
+
             QString compositeOpId;
             QBitArray channelFlags;
             bool compositionVaries = false;
+            bool isFirstCycle = true;
 
             foreach (KisNodeSP node, m_info->allSrcNodes()) {
-                if (compositeOpId.isEmpty()) {
+                if (isFirstCycle) {
                     compositeOpId = node->compositeOpId();
-                } else if (compositeOpId != node->compositeOpId()) {
+                    channelFlags = channelFlagsLazy(node);
+                    isFirstCycle = false;
+                } else if (compositeOpId != node->compositeOpId() ||
+                           channelFlags != channelFlagsLazy(node)) {
                     compositionVaries = true;
                     break;
                 }
@@ -413,6 +482,11 @@ namespace KisLayerUtils {
                     m_info->dstLayer()->setChannelFlags(channelFlags);
                 }
             }
+
+            m_info->nodesCompositingVaries = compositionVaries;
+
+            m_info->dstNode->setUseInTimeline(m_info->useInTimeline);
+            dstPaintLayer->setOnionSkinEnabled(m_info->enableOnionSkins);
         }
 
     private:
@@ -424,7 +498,7 @@ namespace KisLayerUtils {
         MergeLayers(MergeDownInfoSP info) : m_info(info) {}
 
         void populateChildCommands() override {
-            // actual merging done by KisLayer::createMergedLayer (or specialized decendant)
+            // actual merging done by KisLayer::createMergedLayer (or specialized descendant)
             m_info->currLayer->fillMergedLayerTemplate(m_info->dstLayer(), m_info->prevLayer);
         }
 
@@ -490,9 +564,9 @@ namespace KisLayerUtils {
     {
     }
 
-    void KeepNodesSelectedCommand::end() {
+    void KeepNodesSelectedCommand::partB() {
         KisImageSignalType type;
-        if (isFinalizing()) {
+        if (getState() == State::FINALIZING) {
             type = ComplexNodeReselectionSignal(m_activeAfter, m_selectedAfter);
         } else {
             type = ComplexNodeReselectionSignal(m_activeBefore, m_selectedBefore);
@@ -500,8 +574,17 @@ namespace KisLayerUtils {
         m_image->signalRouter()->emitNotification(type);
     }
 
-    KisLayerSP constructDefaultLayer(KisImageSP image) {
-        return new KisPaintLayer(image.data(), image->nextLayerName(), OPACITY_OPAQUE_U8, image->colorSpace());
+    SelectGlobalSelectionMask::SelectGlobalSelectionMask(KisImageSP image)
+        : m_image(image)
+    {
+    }
+
+    void SelectGlobalSelectionMask::redo() {
+
+        KisImageSignalType type =
+                ComplexNodeReselectionSignal(m_image->rootLayer()->selectionMask(), KisNodeList());
+        m_image->signalRouter()->emitNotification(type);
+
     }
 
     RemoveNodeHelper::~RemoveNodeHelper()
@@ -520,13 +603,22 @@ namespace KisLayerUtils {
     void RemoveNodeHelper::safeRemoveMultipleNodes(KisNodeList nodes, KisImageSP image) {
         const bool lastLayer = scanForLastLayer(image, nodes);
 
+        auto isNodeWeird = [] (KisNodeSP node) {
+            const bool normalCompositeMode = node->compositeOpId() == COMPOSITE_OVER;
+
+            KisLayer *layer = dynamic_cast<KisLayer*>(node.data());
+            const bool hasInheritAlpha = layer && layer->alphaChannelDisabled();
+            return !normalCompositeMode && !hasInheritAlpha;
+        };
+
         while (!nodes.isEmpty()) {
             KisNodeList::iterator it = nodes.begin();
 
             while (it != nodes.end()) {
                 if (!checkIsSourceForClone(*it, nodes)) {
                     KisNodeSP node = *it;
-                    addCommandImpl(new KisImageLayerRemoveCommand(image, node, false, true));
+
+                    addCommandImpl(new KisImageLayerRemoveCommand(image, node, !isNodeWeird(node), true));
                     it = nodes.erase(it);
                 } else {
                     ++it;
@@ -535,7 +627,7 @@ namespace KisLayerUtils {
         }
 
         if (lastLayer) {
-            KisLayerSP newLayer = constructDefaultLayer(image);
+            KisLayerSP newLayer = new KisPaintLayer(image.data(), image->nextLayerName(), OPACITY_OPAQUE_U8, image->colorSpace());
             addCommandImpl(new KisImageLayerAddCommand(image, newLayer,
                                                        image->root(),
                                                        KisNodeSP(),
@@ -631,12 +723,47 @@ namespace KisLayerUtils {
                 putAfter = nodesToDelete.last();
             }
 
-            // Add the new merged node on top of the active node -- checking
-            // whether the parent is going to be deleted
-            parent = putAfter->parent();
-            while (parent && nodesToDelete.contains(parent)) {
-                parent = parent->parent();
+            // Add the new merged node on top of the active node
+            // -- checking all parents if they are included in nodesToDelete
+            // Not every descendant is included in nodesToDelete even if in fact
+            //   they are going to be deleted, so we need to check it.
+            // If we consider the path from root to the putAfter node,
+            //    if there are any nodes marked for deletion, any node afterwards
+            //    is going to be deleted, too.
+            //   example:      root . . . . . ! ! . . ! ! ! ! . . . . putAfter
+            //   it should be: root . . . . . ! ! ! ! ! ! ! ! ! ! ! ! !putAfter
+            //   and here:     root . . . . X ! ! . . ! ! ! ! . . . . putAfter
+            //   you can see which node is "the perfect ancestor"
+            //   (marked X; called "parent" in the function arguments).
+            //   and here:     root . . . . . O ! . . ! ! ! ! . . . . putAfter
+            //   you can see which node is "the topmost deleted ancestor" (marked 'O')
+
+            KisNodeSP node = putAfter->parent();
+            bool foundDeletedAncestor = false;
+            KisNodeSP topmostAncestorToDelete = nullptr;
+
+            while (node) {
+
+                if (nodesToDelete.contains(node)
+                        && !nodesToDelete.contains(node->parent())) {
+                    foundDeletedAncestor = true;
+                    topmostAncestorToDelete = node;
+                    // Here node is to be deleted and its parent is not,
+                    // so its parent is the one of the first not deleted (="perfect") ancestors.
+                    // We need the one that is closest to the top (root)
+                }
+
+                node = node->parent();
             }
+
+            if (foundDeletedAncestor) {
+                parent = topmostAncestorToDelete->parent();
+                putAfter = topmostAncestorToDelete;
+            }
+            else {
+                parent = putAfter->parent(); // putAfter (and none of its ancestors) is to be deleted, so its parent is the first not deleted ancestor
+            }
+
         }
 
         void populateChildCommands() override {
@@ -658,20 +785,12 @@ namespace KisLayerUtils {
 
             }
             else {
-                if (parent == m_putAfter->parent()) {
-                    addCommand(new KisImageLayerAddCommand(m_info->image,
-                                                           m_info->dstNode,
-                                                           parent,
-                                                           m_putAfter,
-                                                           true, false));
-                }
-                else {
-                    addCommand(new KisImageLayerAddCommand(m_info->image,
-                                                           m_info->dstNode,
-                                                           parent,
-                                                           parent->lastChild(),
-                                                           true, false));
-                }
+                addCommand(new KisImageLayerAddCommand(m_info->image,
+                                                       m_info->dstNode,
+                                                       parent,
+                                                       m_putAfter,
+                                                       true, false));
+
 
                 /**
                  * We can merge selection masks, in this case dstLayer is not defined!
@@ -682,7 +801,18 @@ namespace KisLayerUtils {
                                            m_info->selectionMasks);
                 }
 
-                safeRemoveMultipleNodes(m_info->allSrcNodes(), m_info->image);
+                KisNodeList safeNodesToDelete = m_info->allSrcNodes();
+                for (KisNodeList::iterator it = safeNodesToDelete.begin(); it != safeNodesToDelete.end(); ++it) {
+                    KisNodeSP node = *it;
+                    if (node->userLocked() && node->visible()) {
+                        addCommand(new KisImageChangeVisibilityCommand(false, node));
+                    }
+                }
+
+                KritaUtils::filterContainer<KisNodeList>(safeNodesToDelete, [](KisNodeSP node) {
+                  return !node->userLocked();
+                });
+                safeRemoveMultipleNodes(safeNodesToDelete, m_info->image);
             }
 
 
@@ -720,7 +850,7 @@ namespace KisLayerUtils {
 
     SwitchFrameCommand::~SwitchFrameCommand() {}
 
-    void SwitchFrameCommand::init() {
+    void SwitchFrameCommand::partA() {
         KisImageAnimationInterface *interface = m_image->animationInterface();
         const int currentTime = interface->currentTime();
         if (currentTime == m_newTime) {
@@ -732,7 +862,7 @@ namespace KisLayerUtils {
         interface->saveAndResetCurrentTime(m_newTime, &m_storage->value);
     }
 
-    void SwitchFrameCommand::end() {
+    void SwitchFrameCommand::partB() {
         KisImageAnimationInterface *interface = m_image->animationInterface();
         const int currentTime = interface->currentTime();
         if (currentTime == m_storage->value) {
@@ -749,11 +879,27 @@ namespace KisLayerUtils {
         void populateChildCommands() override {
             KUndo2Command *cmd = new KisCommandUtils::SkipFirstRedoWrapper();
             KisKeyframeChannel *channel = m_info->dstNode->getKeyframeChannel(KisKeyframeChannel::Content.id());
-            channel->addKeyframe(m_frame, cmd);
+            KisKeyframeSP keyframe = channel->addKeyframe(m_frame, cmd);
+
+            applyKeyframeColorLabel(keyframe);
 
             addCommand(cmd);
         }
 
+        void applyKeyframeColorLabel(KisKeyframeSP dstKeyframe) {
+            Q_FOREACH(KisNodeSP srcNode, m_info->allSrcNodes()) {
+                Q_FOREACH(KisKeyframeChannel *channel, srcNode->keyframeChannels().values()) {
+                    KisKeyframeSP keyframe = channel->keyframeAt(m_frame);
+                    if (!keyframe.isNull() && keyframe->colorLabel() != 0) {
+                        dstKeyframe->setColorLabel(keyframe->colorLabel());
+                        return;
+                    }
+                }
+            }
+
+            dstKeyframe->setColorLabel(0);
+        }
+
     private:
         MergeDownInfoBaseSP m_info;
         int m_frame;
@@ -833,7 +979,12 @@ namespace KisLayerUtils {
             applicator.applyCommand(new FillSelectionMasks(info));
             applicator.applyCommand(new CreateMergedLayer(info), KisStrokeJobData::BARRIER);
 
-            // in two-layer mode we disable pass trhough only when the destination layer
+            // NOTE: shape layer may have emitted spontaneous jobs during layer creation,
+            //       wait for them to complete!
+            applicator.applyCommand(new RefreshDelayedUpdateLayers(info), KisStrokeJobData::BARRIER);
+            applicator.applyCommand(new KUndo2Command(), KisStrokeJobData::BARRIER);
+
+            // in two-layer mode we disable pass through only when the destination layer
             // is not a group layer
             applicator.applyCommand(new DisablePassThroughForHeadsOnly(info, true));
             applicator.applyCommand(new KUndo2Command(), KisStrokeJobData::BARRIER);
@@ -948,8 +1099,7 @@ namespace KisLayerUtils {
         while (it != nodes.end()) {
             if ((!allowMasks && !qobject_cast<KisLayer*>(it->data())) ||
                 checkIsChildOf(*it, nodes)) {
-
-                qDebug() << "Skipping node" << ppVar((*it)->name());
+                //qDebug() << "Skipping node" << ppVar((*it)->name());
                 it = nodes.erase(it);
             } else {
                 ++it;
@@ -1048,7 +1198,7 @@ namespace KisLayerUtils {
         int putAfterIndex = -1;
 
         Q_FOREACH(KisNodeSP node, nodes) {
-            if (node->visible()) {
+            if (node->visible() || node->userLocked()) {
                 visibleNodes << node;
             } else {
                 *invisibleNodes << node;
@@ -1094,7 +1244,7 @@ namespace KisLayerUtils {
         filterMergableNodes(mergedNodes);
         {
             KisNodeList tempNodes;
-            qSwap(mergedNodes, tempNodes);
+            std::swap(mergedNodes, tempNodes);
             sortMergableNodes(image->root(), tempNodes, mergedNodes);
         }
 
@@ -1116,6 +1266,14 @@ namespace KisLayerUtils {
         mergedNodes = filterInvisibleNodes(originalNodes, &invisibleNodes, &putAfter);
 
         if (!invisibleNodes.isEmpty()) {
+            /* If the putAfter node is invisible,
+             * we should instead pick one of the nodes
+             * to be merged to avoid a null putAfter.
+             */
+            if (!putAfter->visible()){
+                putAfter = mergedNodes.first();
+            }
+
             applicator.applyCommand(
                 new SimpleRemoveLayers(invisibleNodes,
                                        image),
@@ -1136,6 +1294,8 @@ namespace KisLayerUtils {
             applicator.applyCommand(new KeepMergedNodesSelected(info, putAfter, false));
             applicator.applyCommand(new FillSelectionMasks(info));
             applicator.applyCommand(new CreateMergedLayerMultiple(info, layerName), KisStrokeJobData::BARRIER);
+            applicator.applyCommand(new DisableExtraCompositing(info));
+            applicator.applyCommand(new KUndo2Command(), KisStrokeJobData::BARRIER);
 
             if (info->frames.size() > 0) {
                 foreach (int frame, info->frames) {
@@ -1164,6 +1324,7 @@ namespace KisLayerUtils {
                                             KisStrokeJobData::SEQUENTIAL,
                                         KisStrokeJobData::EXCLUSIVE);
             }
+
             applicator.applyCommand(new KeepMergedNodesSelected(info, putAfter, true));
         }
 
@@ -1286,12 +1447,17 @@ namespace KisLayerUtils {
         mergeMultipleLayersImpl(image, mergedNodes, layer, true, kundo2_i18n("Flatten Layer"));
     }
 
-    void flattenImage(KisImageSP image)
+    void flattenImage(KisImageSP image, KisNodeSP activeNode)
     {
+        if (!activeNode) {
+            activeNode = image->root()->lastChild();
+        }
+
+
         KisNodeList mergedNodes;
         mergedNodes << image->root();
 
-        mergeMultipleLayersImpl(image, mergedNodes, 0, true, kundo2_i18n("Flatten Image"));
+        mergeMultipleLayersImpl(image, mergedNodes, activeNode, true, kundo2_i18n("Flatten Image"));
     }
 
     KisSimpleUpdateCommand::KisSimpleUpdateCommand(KisNodeList nodes, bool finalize, KUndo2Command *parent)
@@ -1299,7 +1465,7 @@ namespace KisLayerUtils {
           m_nodes(nodes)
     {
     }
-    void KisSimpleUpdateCommand::end()
+    void KisSimpleUpdateCommand::partB()
     {
         updateNodes(m_nodes);
     }
@@ -1310,17 +1476,6 @@ namespace KisLayerUtils {
         }
     }
 
-    void recursiveApplyNodes(KisNodeSP node, std::function<void(KisNodeSP)> func)
-    {
-        func(node);
-
-        node = node->firstChild();
-        while (node) {
-            recursiveApplyNodes(node, func);
-            node = node->nextSibling();
-        }
-    }
-
     KisNodeSP recursiveFindNode(KisNodeSP node, std::function<bool(KisNodeSP)> func)
     {
         if (func(node)) {
@@ -1344,7 +1499,44 @@ namespace KisLayerUtils {
         return recursiveFindNode(root,
             [uuid] (KisNodeSP node) {
                 return node->uuid() == uuid;
+        });
+    }
+
+    void forceAllDelayedNodesUpdate(KisNodeSP root)
+    {
+        KisLayerUtils::recursiveApplyNodes(root,
+        [] (KisNodeSP node) {
+            KisDelayedUpdateNodeInterface *delayedUpdate =
+                    dynamic_cast<KisDelayedUpdateNodeInterface*>(node.data());
+            if (delayedUpdate) {
+                delayedUpdate->forceUpdateTimedNode();
+            }
+        });
+    }
+
+    bool hasDelayedNodeWithUpdates(KisNodeSP root)
+    {
+        return recursiveFindNode(root,
+            [] (KisNodeSP node) {
+                KisDelayedUpdateNodeInterface *delayedUpdate =
+                    dynamic_cast<KisDelayedUpdateNodeInterface*>(node.data());
+
+                return delayedUpdate ? delayedUpdate->hasPendingTimedUpdates() : false;
             });
     }
 
+    KisImageSP findImageByHierarchy(KisNodeSP node)
+    {
+        while (node) {
+            const KisLayer *layer = dynamic_cast<const KisLayer*>(node.data());
+            if (layer) {
+                return layer->image();
+            }
+
+            node = node->parent();
+        }
+
+        return 0;
+    }
+
 }
