@@ -100,13 +100,15 @@ public:
     KisDefaultBoundsBaseSP defaultBounds;
     QScopedPointer<KisPaintDeviceStrategy> basicStrategy;
     QScopedPointer<KisPaintDeviceWrappedStrategy> wrappedStrategy;
+    QMutex m_wrappedStrategyMutex;
+
     QScopedPointer<KisPaintDeviceFramesInterface> framesInterface;
     bool isProjectionDevice;
 
     KisPaintDeviceStrategy* currentStrategy();
 
     void init(const KoColorSpace *cs, const quint8 *defaultPixel);
-    KUndo2Command* convertColorSpace(const KoColorSpace * dstColorSpace, KoColorConversionTransformation::Intent renderingIntent, KoColorConversionTransformation::ConversionFlags conversionFlags);
+    void convertColorSpace(const KoColorSpace * dstColorSpace, KoColorConversionTransformation::Intent renderingIntent, KoColorConversionTransformation::ConversionFlags conversionFlags, KUndo2Command *parentCommand);
     bool assignProfile(const KoColorProfile * profile);
 
     inline const KoColorSpace* colorSpace() const
@@ -159,13 +161,13 @@ public:
             if (m_data) {
                 m_data->prepareClone(rhs->currentNonLodData(), true);
             } else {
-                m_data = toQShared(new KisPaintDeviceData(rhs->currentNonLodData(), true));
+                m_data = toQShared(new KisPaintDeviceData(q, rhs->currentNonLodData(), true));
             }
         } else {
             if (m_data && !rhs->m_data) {
                 m_data.clear();
             } else if (!m_data && rhs->m_data) {
-                m_data = toQShared(new KisPaintDeviceData(rhs->m_data.data(), true));
+                m_data = toQShared(new KisPaintDeviceData(q, rhs->m_data.data(), true));
             } else if (m_data && rhs->m_data) {
                 m_data->prepareClone(rhs->m_data.data(), true);
             }
@@ -175,7 +177,7 @@ public:
                 FramesHash::const_iterator end = rhs->m_frames.constEnd();
 
                 for (; it != end; ++it) {
-                    DataSP data = toQShared(new KisPaintDeviceData(it.value().data(), true));
+                    DataSP data = toQShared(new KisPaintDeviceData(q, it.value().data(), true));
                     m_frames.insert(it.key(), data);
                 }
             }
@@ -183,7 +185,7 @@ public:
         }
 
         if (rhs->m_lodData) {
-            m_lodData.reset(new KisPaintDeviceData(rhs->m_lodData.data(), true));
+            m_lodData.reset(new KisPaintDeviceData(q, rhs->m_lodData.data(), true));
         }
     }
 
@@ -290,17 +292,17 @@ public:
              * new frame and clear m_data to make the "background" for
              * the areas where there is no frame at all.
              */
-            data = toQShared(new Data(m_data.data(), true));
+            data = toQShared(new Data(q, m_data.data(), true));
             m_data->dataManager()->clear();
             m_data->cache()->invalidate();
             initialFrame = true;
 
         } else if (copy) {
             DataSP srcData = m_frames[copySrc];
-            data = toQShared(new Data(srcData.data(), true));
+            data = toQShared(new Data(q, srcData.data(), true));
         } else {
             DataSP srcData = m_frames.begin().value();
-            data = toQShared(new Data(srcData.data(), false));
+            data = toQShared(new Data(q, srcData.data(), false));
         }
 
         if (!initialFrame && !copy) {
@@ -405,8 +407,46 @@ public:
     void uploadLodDataStruct(LodDataStruct *dst);
     QRegion regionForLodSyncing() const;
 
+    void updateLodDataManager(KisDataManager *srcDataManager,
+                              KisDataManager *dstDataManager, const QPoint &srcOffset, const QPoint &dstOffset,
+                              const QRect &originalRect, int lod);
+
+    void generateLodCloneDevice(KisPaintDeviceSP dst, const QRect &originalRect, int lod);
+
     void tesingFetchLodDevice(KisPaintDeviceSP targetDevice);
 
+
+private:
+    qint64 estimateDataSize(Data *data) const {
+        const QRect &rc = data->dataManager()->extent();
+        return rc.width() * rc.height() * data->colorSpace()->pixelSize();
+    }
+
+public:
+
+    void estimateMemoryStats(qint64 &imageData, qint64 &temporaryData, qint64 &lodData) const {
+        imageData = 0;
+        temporaryData = 0;
+        lodData = 0;
+
+        if (m_data) {
+            imageData += estimateDataSize(m_data.data());
+        }
+
+        if (m_lodData) {
+            lodData += estimateDataSize(m_lodData.data());
+        }
+
+        if (m_externalFrameData) {
+            temporaryData += estimateDataSize(m_externalFrameData.data());
+        }
+
+        Q_FOREACH (DataSP value, m_frames.values()) {
+            imageData += estimateDataSize(value.data());
+        }
+    }
+
+
 private:
 
     QRegion syncWholeDevice(Data *srcData);
@@ -447,7 +487,7 @@ private:
             if (!m_externalFrameData) {
                 QMutexLocker l(&m_dataSwitchLock);
                 if (!m_externalFrameData) {
-                    m_externalFrameData.reset(new Data(m_data.data(), false));
+                    m_externalFrameData.reset(new Data(q, m_data.data(), false));
                 }
             }
             data = m_externalFrameData.data();
@@ -463,7 +503,7 @@ private:
 
             QMutexLocker l(&m_dataSwitchLock);
             if (!m_lodData) {
-                m_lodData.reset(new Data(srcData, false));
+                m_lodData.reset(new Data(q, srcData, false));
             }
         }
     }
@@ -556,9 +596,16 @@ KisPaintDevice::Private::KisPaintDeviceStrategy* KisPaintDevice::Private::curren
         return basicStrategy.data();
     }
 
-    QRect wrapRect = defaultBounds->bounds();
+    const QRect wrapRect = defaultBounds->bounds();
+
     if (!wrappedStrategy || wrappedStrategy->wrapRect() != wrapRect) {
-        wrappedStrategy.reset(new KisPaintDeviceWrappedStrategy(wrapRect, q, this));
+        QMutexLocker locker(&m_wrappedStrategyMutex);
+
+        if (!wrappedStrategy) {
+            wrappedStrategy.reset(new KisPaintDeviceWrappedStrategy(wrapRect, q, this));
+        }  else if (wrappedStrategy->wrapRect() != wrapRect) {
+            wrappedStrategy->setWrapRect(wrapRect);
+        }
     }
 
     return wrappedStrategy.data();
@@ -609,9 +656,11 @@ QRegion KisPaintDevice::Private::regionForLodSyncing() const
 
 KisPaintDevice::LodDataStruct* KisPaintDevice::Private::createLodDataStruct(int newLod)
 {
+    KIS_SAFE_ASSERT_RECOVER_NOOP(newLod > 0);
+
     Data *srcData = currentNonLodData();
 
-    Data *lodData = new Data(srcData, false);
+    Data *lodData = new Data(q, srcData, false);
     LodDataStruct *lodStruct = new LodDataStructImpl(lodData);
 
     int expectedX = KisLodTransform::coordToLodCoord(srcData->x(), newLod);
@@ -642,15 +691,13 @@ KisPaintDevice::LodDataStruct* KisPaintDevice::Private::createLodDataStruct(int
     return lodStruct;
 }
 
-void KisPaintDevice::Private::updateLodDataStruct(LodDataStruct *_dst, const QRect &originalRect)
+void KisPaintDevice::Private::updateLodDataManager(KisDataManager *srcDataManager,
+                                                   KisDataManager *dstDataManager,
+                                                   const QPoint &srcOffset,
+                                                   const QPoint &dstOffset,
+                                                   const QRect &originalRect,
+                                                   int lod)
 {
-    LodDataStructImpl *dst = dynamic_cast<LodDataStructImpl*>(_dst);
-    KIS_SAFE_ASSERT_RECOVER_RETURN(dst);
-
-    Data *lodData = dst->lodData.data();
-    Data *srcData = currentNonLodData();
-
-    const int lod = lodData->levelOfDetail();
     const int srcStepSize = 1 << lod;
 
     KIS_ASSERT_RECOVER_RETURN(lod > 0);
@@ -661,7 +708,7 @@ void KisPaintDevice::Private::updateLodDataStruct(LodDataStruct *_dst, const QRe
 
     KIS_ASSERT_RECOVER_NOOP(srcRect.width() / srcStepSize == dstRect.width());
 
-    const int pixelSize = srcData->dataManager()->pixelSize();
+    const int pixelSize = srcDataManager->pixelSize();
 
     int rowsAccumulated = 0;
     int columnsAccumulated = 0;
@@ -690,14 +737,14 @@ void KisPaintDevice::Private::updateLodDataStruct(LodDataStruct *_dst, const QRe
         weights[srcCellSize - 1] = averageWeight - extraWeight;
     }
 
-    InternalSequentialConstIterator srcIntIt(StrategyPolicy(currentStrategy(), srcData->dataManager().data(), srcData->x(), srcData->y()), srcRect);
-    InternalSequentialIterator dstIntIt(StrategyPolicy(currentStrategy(), lodData->dataManager().data(), lodData->x(), lodData->y()), dstRect);
+    InternalSequentialConstIterator srcIntIt(StrategyPolicy(currentStrategy(), srcDataManager, srcOffset.x(), srcOffset.y()), srcRect);
+    InternalSequentialIterator dstIntIt(StrategyPolicy(currentStrategy(), dstDataManager, dstOffset.x(), dstOffset.y()), dstRect);
 
     int rowsRemaining = srcRect.height();
     while (rowsRemaining > 0) {
 
         int colsRemaining = srcRect.width();
-        while (colsRemaining > 0) {
+        while (colsRemaining > 0 && srcIntIt.nextPixel()) {
 
             memcpy(blendDataPtr, srcIntIt.rawDataConst(), pixelSize);
             blendDataPtr += pixelSize;
@@ -708,7 +755,6 @@ void KisPaintDevice::Private::updateLodDataStruct(LodDataStruct *_dst, const QRe
                 columnsAccumulated = 0;
             }
 
-            srcIntIt.nextPixel();
             colsRemaining--;
         }
 
@@ -718,11 +764,13 @@ void KisPaintDevice::Private::updateLodDataStruct(LodDataStruct *_dst, const QRe
 
             // blend and write the final data
             blendDataPtr = blendData.data();
-            for (int i = 0; i < dstRect.width(); i++) {
-                mixOp->mixColors(blendDataPtr, weights.data(), srcCellSize, dstIntIt.rawData());
 
+            int colsRemaining = dstRect.width();
+            while (colsRemaining > 0 && dstIntIt.nextPixel()) {
+                mixOp->mixColors(blendDataPtr, weights.data(), srcCellSize, dstIntIt.rawData());
                 blendDataPtr += srcCellStride;
-                dstIntIt.nextPixel();
+
+                colsRemaining--;
             }
 
             // reset counters
@@ -738,6 +786,33 @@ void KisPaintDevice::Private::updateLodDataStruct(LodDataStruct *_dst, const QRe
     }
 }
 
+void KisPaintDevice::Private::updateLodDataStruct(LodDataStruct *_dst, const QRect &originalRect)
+{
+    LodDataStructImpl *dst = dynamic_cast<LodDataStructImpl*>(_dst);
+    KIS_SAFE_ASSERT_RECOVER_RETURN(dst);
+
+    Data *lodData = dst->lodData.data();
+    Data *srcData = currentNonLodData();
+
+    const int lod = lodData->levelOfDetail();
+
+    updateLodDataManager(srcData->dataManager().data(), lodData->dataManager().data(),
+                         QPoint(srcData->x(), srcData->y()),
+                         QPoint(lodData->x(), lodData->y()),
+                         originalRect, lod);
+}
+
+void KisPaintDevice::Private::generateLodCloneDevice(KisPaintDeviceSP dst, const QRect &originalRect, int lod)
+{
+    KIS_SAFE_ASSERT_RECOVER_RETURN(fastBitBltPossible(dst));
+
+    Data *srcData = currentNonLodData();
+    updateLodDataManager(srcData->dataManager().data(), dst->dataManager().data(),
+                         QPoint(srcData->x(), srcData->y()),
+                         QPoint(dst->x(), dst->y()),
+                         originalRect, lod);
+}
+
 void KisPaintDevice::Private::uploadLodDataStruct(LodDataStruct *_dst)
 {
     LodDataStructImpl *dst = dynamic_cast<LodDataStructImpl*>(_dst);
@@ -796,7 +871,7 @@ void KisPaintDevice::Private::uploadFrameData(DataSP srcData, DataSP dstData)
 
         KUndo2Command tempCommand;
 
-        srcData = toQShared(new Data(srcData.data(), true));
+        srcData = toQShared(new Data(q, srcData.data(), true));
         srcData->convertDataColorSpace(dstData->colorSpace(),
                                        KoColorConversionTransformation::internalRenderingIntent(),
                                        KoColorConversionTransformation::internalConversionFlags(),
@@ -820,14 +895,15 @@ void KisPaintDevice::Private::tesingFetchLodDevice(KisPaintDeviceSP targetDevice
     transferFromData(data, targetDevice);
 }
 
-KUndo2Command* KisPaintDevice::Private::convertColorSpace(const KoColorSpace * dstColorSpace, KoColorConversionTransformation::Intent renderingIntent, KoColorConversionTransformation::ConversionFlags conversionFlags)
+void KisPaintDevice::Private::convertColorSpace(const KoColorSpace * dstColorSpace, KoColorConversionTransformation::Intent renderingIntent, KoColorConversionTransformation::ConversionFlags conversionFlags, KUndo2Command *parentCommand)
 {
 
     class DeviceChangeColorSpaceCommand : public KUndo2Command
     {
     public:
-        DeviceChangeColorSpaceCommand(KisPaintDeviceSP device)
-            : m_firstRun(true),
+        DeviceChangeColorSpaceCommand(KisPaintDeviceSP device, KUndo2Command *parent = 0)
+            : KUndo2Command(parent),
+              m_firstRun(true),
               m_device(device)
         {
         }
@@ -862,25 +938,19 @@ KUndo2Command* KisPaintDevice::Private::convertColorSpace(const KoColorSpace * d
     };
 
 
-    KUndo2Command *parentCommand = new DeviceChangeColorSpaceCommand(q);
+    QList<Data*> dataObjects = allDataObjects();
+    if (dataObjects.isEmpty()) return;
 
-    QList<Data*> dataObjects = allDataObjects();;
+    KUndo2Command *mainCommand =
+        parentCommand ? new DeviceChangeColorSpaceCommand(q, parentCommand) : 0;
 
     Q_FOREACH (Data *data, dataObjects) {
         if (!data) continue;
 
-        data->convertDataColorSpace(dstColorSpace, renderingIntent, conversionFlags, parentCommand);
+        data->convertDataColorSpace(dstColorSpace, renderingIntent, conversionFlags, mainCommand);
     }
 
-    if (!parentCommand->childCount()) {
-        delete parentCommand;
-        parentCommand = 0;
-    } else {
-        q->emitColorSpaceChanged();
-    }
-
-    return parentCommand;
-
+    q->emitColorSpaceChanged();
 }
 
 bool KisPaintDevice::Private::assignProfile(const KoColorProfile * profile)
@@ -891,7 +961,7 @@ bool KisPaintDevice::Private::assignProfile(const KoColorProfile * profile)
         KoColorSpaceRegistry::instance()->colorSpace(colorSpace()->colorModelId().id(), colorSpace()->colorDepthId().id(), profile);
     if (!dstColorSpace) return false;
 
-    QList<Data*> dataObjects = allDataObjects();;
+    QList<Data*> dataObjects = allDataObjects();
     Q_FOREACH (Data *data, dataObjects) {
         if (!data) continue;
         data->assignColorSpace(dstColorSpace);
@@ -904,7 +974,7 @@ bool KisPaintDevice::Private::assignProfile(const KoColorProfile * profile)
 
 void KisPaintDevice::Private::init(const KoColorSpace *cs, const quint8 *defaultPixel)
 {
-    QList<Data*> dataObjects = allDataObjects();;
+    QList<Data*> dataObjects = allDataObjects();
     Q_FOREACH (Data *data, dataObjects) {
         if (!data) continue;
 
@@ -953,30 +1023,35 @@ void KisPaintDevice::init(const KoColorSpace *colorSpace,
     setParentNode(parent);
 }
 
-KisPaintDevice::KisPaintDevice(const KisPaintDevice& rhs, bool copyFrames, KisNode *newParentNode)
+KisPaintDevice::KisPaintDevice(const KisPaintDevice& rhs, KritaUtils::DeviceCopyMode copyMode, KisNode *newParentNode)
     : QObject()
     , KisShared()
     , m_d(new Private(this))
 {
     if (this != &rhs) {
-        // temporary def. bounds object for the initialization phase only
-        m_d->defaultBounds = m_d->transitionalDefaultBounds;
-
-        // copy data objects with or without frames
-        m_d->cloneAllDataObjects(rhs.m_d, copyFrames);
-
-        if (copyFrames) {
-            KIS_ASSERT_RECOVER_RETURN(rhs.m_d->framesInterface);
-            KIS_ASSERT_RECOVER_RETURN(rhs.m_d->contentChannel);
-            m_d->framesInterface.reset(new KisPaintDeviceFramesInterface(this));
-            m_d->contentChannel.reset(new KisRasterKeyframeChannel(*rhs.m_d->contentChannel.data(), newParentNode, this));
-        }
-
-        setDefaultBounds(rhs.m_d->defaultBounds);
-        setParentNode(0);
+        makeFullCopyFrom(rhs, copyMode, newParentNode);
     }
 }
 
+void KisPaintDevice::makeFullCopyFrom(const KisPaintDevice &rhs, KritaUtils::DeviceCopyMode copyMode, KisNode *newParentNode)
+{
+    // temporary def. bounds object for the initialization phase only
+    m_d->defaultBounds = m_d->transitionalDefaultBounds;
+
+    // copy data objects with or without frames
+    m_d->cloneAllDataObjects(rhs.m_d, copyMode == KritaUtils::CopyAllFrames);
+
+    if (copyMode == KritaUtils::CopyAllFrames && rhs.m_d->framesInterface) {
+        KIS_ASSERT_RECOVER_RETURN(rhs.m_d->framesInterface);
+        KIS_ASSERT_RECOVER_RETURN(rhs.m_d->contentChannel);
+        m_d->framesInterface.reset(new KisPaintDeviceFramesInterface(this));
+        m_d->contentChannel.reset(new KisRasterKeyframeChannel(*rhs.m_d->contentChannel.data(), newParentNode, this));
+    }
+
+    setDefaultBounds(rhs.m_d->defaultBounds);
+    setParentNode(newParentNode);
+}
+
 KisPaintDevice::~KisPaintDevice()
 {
     delete m_d;
@@ -1055,6 +1130,11 @@ int KisPaintDevice::sequenceNumber() const
     return m_d->cache()->sequenceNumber();
 }
 
+void KisPaintDevice::estimateMemoryStats(qint64 &imageData, qint64 &temporaryData, qint64 &lodData) const
+{
+    m_d->estimateMemoryStats(imageData, temporaryData, lodData);
+}
+
 void KisPaintDevice::setParentNode(KisNodeWSP parent)
 {
     m_d->parent = parent;
@@ -1187,6 +1267,10 @@ QRect calculateExactBoundsImpl(const KisPaintDevice *device, const QRect &startR
 {
     if (startRect == endRect) return startRect;
 
+    // the passed extent might have weird invalid structure that
+    // can overflow integer precision when calling startRect.right()
+    if (!startRect.isValid()) return QRect();
+
     // Solution n°2
     int  x, y, w, h;
     int boundLeft, boundTop, boundRight, boundBottom;
@@ -1392,7 +1476,7 @@ void KisPaintDevice::clear(const QRect & rc)
 
 void KisPaintDevice::fill(const QRect & rc, const KoColor &color)
 {
-    Q_ASSERT(*color.colorSpace() == *colorSpace());
+    KIS_ASSERT_RECOVER_RETURN(*color.colorSpace() == *colorSpace());
     m_d->currentStrategy()->fill(rc, color.data());
 }
 
@@ -1427,10 +1511,9 @@ void KisPaintDevice::emitProfileChanged()
     emit profileChanged(m_d->colorSpace()->profile());
 }
 
-KUndo2Command* KisPaintDevice::convertTo(const KoColorSpace * dstColorSpace, KoColorConversionTransformation::Intent renderingIntent, KoColorConversionTransformation::ConversionFlags conversionFlags)
+void KisPaintDevice::convertTo(const KoColorSpace * dstColorSpace, KoColorConversionTransformation::Intent renderingIntent, KoColorConversionTransformation::ConversionFlags conversionFlags, KUndo2Command *parentCommand)
 {
-    KUndo2Command *command = m_d->convertColorSpace(dstColorSpace, renderingIntent, conversionFlags);
-    return command;
+    m_d->convertColorSpace(dstColorSpace, renderingIntent, conversionFlags, parentCommand);
 }
 
 bool KisPaintDevice::setProfile(const KoColorProfile * profile)
@@ -1465,7 +1548,7 @@ void KisPaintDevice::convertFromQImage(const QImage& _image, const KoColorProfil
 
             writeBytes(dstData, offsetX, offsetY, image.width(), image.height());
             delete[] dstData;
-        } catch (std::bad_alloc) {
+        } catch (const std::bad_alloc&) {
             warnKrita << "KisPaintDevice::convertFromQImage: Could not allocate" << image.width() * image.height() * pixelSize() << "bytes";
             return;
         }
@@ -1511,7 +1594,7 @@ QImage KisPaintDevice::convertToQImage(const KoColorProfile *dstProfile, qint32
     quint8 *data = 0;
     try {
         data = new quint8 [w * h * pixelSize()];
-    } catch (std::bad_alloc) {
+    } catch (const std::bad_alloc&) {
         warnKrita << "KisPaintDevice::convertToQImage std::bad_alloc for " << w << " * " << h << " * " << pixelSize();
         //delete[] data; // data is not allocated, so don't free it
         return QImage();
@@ -1715,28 +1798,33 @@ KisRandomSubAccessorSP KisPaintDevice::createRandomSubAccessor() const
 void KisPaintDevice::clearSelection(KisSelectionSP selection)
 {
     const KoColorSpace *colorSpace = m_d->colorSpace();
-    QRect r = selection->selectedExactRect() & m_d->defaultBounds->bounds();
+    const QRect r = selection->selectedExactRect();
 
     if (r.isValid()) {
 
-        KisHLineIteratorSP devIt = createHLineIteratorNG(r.x(), r.y(), r.width());
-        KisHLineConstIteratorSP selectionIt = selection->projection()->createHLineConstIteratorNG(r.x(), r.y(), r.width());
+        {
+            KisHLineIteratorSP devIt = createHLineIteratorNG(r.x(), r.y(), r.width());
+            KisHLineConstIteratorSP selectionIt = selection->projection()->createHLineConstIteratorNG(r.x(), r.y(), r.width());
 
-        const KoColor defaultPixel = this->defaultPixel();
-        bool transparentDefault = (defaultPixel.opacityU8() == OPACITY_TRANSPARENT_U8);
-        for (qint32 y = 0; y < r.height(); y++) {
+            const KoColor defaultPixel = this->defaultPixel();
+            bool transparentDefault = (defaultPixel.opacityU8() == OPACITY_TRANSPARENT_U8);
+            for (qint32 y = 0; y < r.height(); y++) {
 
-            do {
-                // XXX: Optimize by using stretches
-                colorSpace->applyInverseAlphaU8Mask(devIt->rawData(), selectionIt->rawDataConst(), 1);
-                if (transparentDefault && colorSpace->opacityU8(devIt->rawData()) == OPACITY_TRANSPARENT_U8) {
-                    memcpy(devIt->rawData(), defaultPixel.data(), colorSpace->pixelSize());
-                }
-            } while (devIt->nextPixel() && selectionIt->nextPixel());
-            devIt->nextRow();
-            selectionIt->nextRow();
+                do {
+                    // XXX: Optimize by using stretches
+                    colorSpace->applyInverseAlphaU8Mask(devIt->rawData(), selectionIt->rawDataConst(), 1);
+                    if (transparentDefault && colorSpace->opacityU8(devIt->rawData()) == OPACITY_TRANSPARENT_U8) {
+                        memcpy(devIt->rawData(), defaultPixel.data(), colorSpace->pixelSize());
+                    }
+                } while (devIt->nextPixel() && selectionIt->nextPixel());
+                devIt->nextRow();
+                selectionIt->nextRow();
+            }
         }
+
+        // purge() must be executed **after** all iterators have been destroyed!
         m_d->dataManager()->purge(r.translated(-m_d->x(), -m_d->y()));
+
         setDirty(r);
     }
 }
@@ -1880,8 +1968,10 @@ KisRasterKeyframeChannel *KisPaintDevice::createKeyframeChannel(const KoID &id)
 
 KisRasterKeyframeChannel* KisPaintDevice::keyframeChannel() const
 {
-    Q_ASSERT(m_d->contentChannel);
-    return m_d->contentChannel.data();
+    if (m_d->contentChannel) {
+        return m_d->contentChannel.data();
+    }
+    return 0;
 }
 
 const KoColorSpace* KisPaintDevice::colorSpace() const
@@ -1932,7 +2022,7 @@ QVector<qint32> KisPaintDevice::channelSizes() const
 {
     QVector<qint32> sizes;
     QList<KoChannelInfo*> channels = colorSpace()->channels();
-    qSort(channels);
+    std::sort(channels.begin(), channels.end());
 
     Q_FOREACH (KoChannelInfo * channelInfo, channels) {
         sizes.append(channelInfo->size());
@@ -1974,12 +2064,28 @@ void KisPaintDevice::uploadLodDataStruct(LodDataStruct *dst)
     m_d->uploadLodDataStruct(dst);
 }
 
+void KisPaintDevice::generateLodCloneDevice(KisPaintDeviceSP dst, const QRect &originalRect, int lod)
+{
+    m_d->generateLodCloneDevice(dst, originalRect, lod);
+}
+
 
 KisPaintDeviceFramesInterface* KisPaintDevice::framesInterface()
 {
     return m_d->framesInterface.data();
 }
 
+void KisPaintDevice::debugPaintDevice(const QString &basename) const
+{
+    static int i = 0;
+    QString filename = QString ("%1_%2.png").arg(QString::number(i), 6, QChar('0')).arg(basename);
+
+    QImage image = convertToQImage(0);
+    image.save(filename);
+
+    i++;
+}
+
 /******************************************************************/
 /*               KisPaintDeviceFramesInterface                    */
 /******************************************************************/
