@@ -54,6 +54,13 @@ class KisPaintDeviceFramesInterface;
 
 typedef KisSharedPtr<KisDataManager> KisDataManagerSP;
 
+namespace KritaUtils {
+enum DeviceCopyMode {
+    CopySnapshot = 0,
+    CopyAllFrames
+};
+}
+
 
 /**
  * A paint device contains the actual pixel data and offers methods
@@ -93,14 +100,16 @@ public:
     /**
      * Creates a copy of this device.
      *
-     * If \p copyFrames is false, the newly created device clones the
+     * If \p copyMode is CopySnapshot, the newly created device clones the
      * current frame of \p rhs only (default and efficient
-     * behavior). If \p copyFrames is true, the new device is a deep
+     * behavior). If \p copyFrames is CopyAllFrames, the new device is a deep
      * copy of the source with all the frames included.
      */
-    KisPaintDevice(const KisPaintDevice& rhs, bool copyFrames = false, KisNode *newParentNode = 0);
+    KisPaintDevice(const KisPaintDevice& rhs, KritaUtils::DeviceCopyMode copyMode = KritaUtils::CopySnapshot, KisNode *newParentNode = 0);
     ~KisPaintDevice() override;
 
+    void makeFullCopyFrom(const KisPaintDevice& rhs, KritaUtils::DeviceCopyMode copyMode = KritaUtils::CopySnapshot, KisNode *newParentNode = 0);
+
 protected:
     /**
      * A special constructor for usage in KisPixelSelection. It allows
@@ -190,7 +199,7 @@ public:
      */
     QRect extent() const;
 
-    /// Convience method for the above
+    /// Convenience method for the above
     void extent(qint32 &x, qint32 &y, qint32 &w, qint32 &h) const;
 
     /**
@@ -222,7 +231,7 @@ public:
     QRect exactBoundsAmortized() const;
 
     /**
-     * Retuns exact rectangle of the paint device that contains
+     * Returns exact rectangle of the paint device that contains
      * non-default pixels. For paint devices with fully transparent
      * default pixel is equivalent to exactBounds().
      *
@@ -259,7 +268,7 @@ public:
      */
     void crop(qint32 x, qint32 y, qint32 w, qint32 h);
 
-    /// Convience method for the above
+    /// Convenience method for the above
     void crop(const QRect & r);
 
     /**
@@ -310,7 +319,7 @@ public:
 public:
 
     /**
-     * Prepares the device for fastBitBlt opreration. It clears
+     * Prepares the device for fastBitBlt operation. It clears
      * the device, switches x,y shifts and colorspace if needed.
      * After this call fastBitBltPossible will return true.
      * May be used for initialization of temporary devices.
@@ -355,7 +364,7 @@ protected:
     /**
      * Checks whether a src paint device can be used as source
      * of fast bitBlt operation. The result of the check may
-     * depend on whether color spaces coinside, whether there is
+     * depend on whether color spaces coincide, whether there is
      * any shift of tiles between the devices and etc.
      *
      * WARNING: This check must be done <i>before</i> performing any
@@ -473,12 +482,11 @@ public:
 
     /**
      * Converts the paint device to a different colorspace
-     *
-     * @return a command that can be used to undo the conversion.
      */
-    KUndo2Command* convertTo(const KoColorSpace * dstColorSpace,
-                             KoColorConversionTransformation::Intent renderingIntent = KoColorConversionTransformation::internalRenderingIntent(),
-                             KoColorConversionTransformation::ConversionFlags conversionFlags = KoColorConversionTransformation::internalConversionFlags());
+    void convertTo(const KoColorSpace * dstColorSpace,
+                   KoColorConversionTransformation::Intent renderingIntent = KoColorConversionTransformation::internalRenderingIntent(),
+                   KoColorConversionTransformation::ConversionFlags conversionFlags = KoColorConversionTransformation::internalConversionFlags(),
+                   KUndo2Command *parentCommand = 0);
 
     /**
      * Changes the profile of the colorspace of this paint device to the given
@@ -488,20 +496,25 @@ public:
 
     /**
      * Fill this paint device with the data from image; starting at (offsetX, offsetY)
-     * @param srcProfileName name of the RGB profile to interpret the image as. 0 is interpreted as sRGB
+     * @param image the image
+     * @param profile name of the RGB profile to interpret the image as. 0 is interpreted as sRGB
+     * @param offsetX x offset
+     * @param offsetY y offset
      */
     void convertFromQImage(const QImage& image, const KoColorProfile *profile, qint32 offsetX = 0, qint32 offsetY = 0);
 
     /**
      * Create an RGBA QImage from a rectangle in the paint device.
      *
+     * @param dstProfile RGB profile to use in conversion. May be 0, in which
+     * case it's up to the color strategy to choose a profile (most
+     * like sRGB).
      * @param x Left coordinate of the rectangle
      * @param y Top coordinate of the rectangle
      * @param w Width of the rectangle in pixels
      * @param h Height of the rectangle in pixels
-     * @param dstProfile RGB profile to use in conversion. May be 0, in which
-     * case it's up to the color strategy to choose a profile (most
-     * like sRGB).
+     * @param renderingIntent Rendering intent
+     * @param conversionFlags Conversion flags
      */
     QImage convertToQImage(const KoColorProfile *dstProfile, qint32 x, qint32 y, qint32 w, qint32 h,
                            KoColorConversionTransformation::Intent renderingIntent = KoColorConversionTransformation::internalRenderingIntent(),
@@ -522,6 +535,8 @@ public:
      * @param dstProfile RGB profile to use in conversion. May be 0, in which
      * case it's up to the color strategy to choose a profile (most
      * like sRGB).
+     * @param renderingIntent Rendering intent
+     * @param conversionFlags Conversion flags
      */
     QImage convertToQImage(const KoColorProfile *  dstProfile,
                            KoColorConversionTransformation::Intent renderingIntent = KoColorConversionTransformation::internalRenderingIntent(),
@@ -532,9 +547,10 @@ public:
      * the aspect ratio. The width and height of the returned device
      * won't exceed \p maxw and \p maxw, but they may be smaller.
      *
-     * @param maxw: maximum width
-     * @param maxh: maximum height
-     * @param rect: only this rect will be used for the thumbnail
+     * @param w maximum width
+     * @param h maximum height
+     * @param rect only this rect will be used for the thumbnail
+     * @param outputRect output rectangle
      *
      */
     KisPaintDeviceSP createThumbnailDevice(qint32 w, qint32 h, QRect rect = QRect(), QRect outputRect = QRect()) const;
@@ -549,6 +565,8 @@ public:
      * @param maxh: maximum height
      * @param rect: only this rect will be used for the thumbnail
      * @param oversample: ratio used for antialiasing
+     * @param renderingIntent Rendering intent
+     * @param conversionFlags Conversion flags
      */
     QImage createThumbnail(qint32 maxw, qint32 maxh, QRect rect, qreal oversample = 1,
                            KoColorConversionTransformation::Intent renderingIntent = KoColorConversionTransformation::internalRenderingIntent(),
@@ -598,7 +616,7 @@ public:
      */
     bool setPixel(qint32 x, qint32 y, const QColor& c);
 
-    /// Convience method for the above
+    /// Convenience method for the above
     bool setPixel(qint32 x, qint32 y, const KoColor& kc);
 
     /**
@@ -618,7 +636,7 @@ public:
      * 2) Fill it with the desired color or data
      *
      * 3) Create a KisPainter and set all the properties of the
-     *    trasaction: selection, compositeOp, opacity and etc.
+     *    transaction: selection, compositeOp, opacity and etc.
      *
      * 4) Paint a newly created paint device over the destination
      *    device.
@@ -713,8 +731,7 @@ public:
     /**
      * Create a keyframe channel for the content on this device.
      * @param id identifier for the channel
-     * @param node the parent node for the channel
-     * @return keyframe channel
+     * @return keyframe channel or 0 if there is not one
      */
     KisRasterKeyframeChannel *createKeyframeChannel(const KoID &id);
 
@@ -725,6 +742,12 @@ public:
      */
     KisPaintDeviceFramesInterface* framesInterface();
 
+    /**
+     * @brief debugPaintDevice save the current paint device to a numbered PNG image
+     * @param basename the basename for the file.
+     */
+    void debugPaintDevice(const QString &basename) const;
+
 public:
 
     /**
@@ -761,6 +784,9 @@ public:
      */
     int sequenceNumber() const;
 
+
+    void estimateMemoryStats(qint64 &imageData, qint64 &temporaryData, qint64 &lodData) const;
+
 public:
 
     KisHLineIteratorSP createHLineIteratorNG(qint32 x, qint32 y, qint32 w);
@@ -776,20 +802,25 @@ public:
      * Create an iterator that will "artificially" extend the paint device with the
      * value of the border when trying to access values outside the range of data.
      *
-     * @param rc indicates the rectangle that truly contains data
+     * @param x x of top left corner
+     * @param y y of top left corner
+     * @param w width of the border
+     * @param _dataWidth indicates the rectangle that truly contains data
      */
     KisRepeatHLineConstIteratorSP createRepeatHLineConstIterator(qint32 x, qint32 y, qint32 w, const QRect& _dataWidth) const;
     /**
      * Create an iterator that will "artificially" extend the paint device with the
      * value of the border when trying to access values outside the range of data.
      *
-     * @param rc indicates the rectangle that trully contains data
+     * @param x x of top left corner
+     * @param y y of top left corner
+     * @param h height of the border
+     * @param _dataWidth indicates the rectangle that truly contains data
      */
     KisRepeatVLineConstIteratorSP createRepeatVLineConstIterator(qint32 x, qint32 y, qint32 h, const QRect& _dataWidth) const;
 
     /**
      * This function create a random accessor which can easily access to sub pixel values.
-     * @param selection an up-to-date selection that has the same origin as the paint device
      */
     KisRandomSubAccessorSP createRandomSubAccessor() const;
 
@@ -832,6 +863,8 @@ public:
     void updateLodDataStruct(LodDataStruct *dst, const QRect &srcRect);
     void uploadLodDataStruct(LodDataStruct *dst);
 
+    void generateLodCloneDevice(KisPaintDeviceSP dst, const QRect &originalRect, int lod);
+
     void setProjectionDevice(bool value);
     void tesingFetchLodDevice(KisPaintDeviceSP targetDevice);
 
