@@ -8,6 +8,7 @@
  *                2003-2011 Boudewijn Rempt <boud@valdyas.org>
  *                2004 Clarence Dang <dang@kde.org>
  *                2011 José Luis Vergara <pentalis@gmail.com>
+ *                2017 L. E. Segovia <leo.segovia@siggraph.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -33,8 +34,9 @@
 #include <QApplication>
 #include <QBuffer>
 #include <QByteArray>
-#include <QDesktopServices>
+#include <QStandardPaths>
 #include <QDesktopWidget>
+#include <QDesktopServices>
 #include <QGridLayout>
 #include <QMainWindow>
 #include <QMenu>
@@ -58,7 +60,6 @@
 #include <KoCanvasController.h>
 #include <KoCompositeOp.h>
 #include <KoDockRegistry.h>
-#include <KoDockWidgetTitleBar.h>
 #include <KoProperties.h>
 #include <KoResourceItemChooserSync.h>
 #include <KoSelection.h>
@@ -69,7 +70,6 @@
 #include <KoZoomHandler.h>
 #include <KoPluginLoader.h>
 #include <KoDocumentInfo.h>
-#include <KoGlobal.h>
 #include <KoColorSpaceRegistry.h>
 
 #include "input/kis_input_manager.h"
@@ -83,6 +83,7 @@
 #include "kis_canvas_controls_manager.h"
 #include "kis_canvas_resource_provider.h"
 #include "kis_composite_progress_proxy.h"
+#include <KoProgressUpdater.h>
 #include "kis_config.h"
 #include "kis_config_notifier.h"
 #include "kis_control_frame.h"
@@ -92,6 +93,7 @@
 #include "kis_filter_manager.h"
 #include "kis_group_layer.h"
 #include <kis_image.h>
+#include <kis_image_barrier_locker.h>
 #include "kis_image_manager.h"
 #include <kis_layer.h>
 #include "kis_mainwindow_observer.h"
@@ -101,15 +103,16 @@
 #include "kis_node_commands_adapter.h"
 #include "kis_node.h"
 #include "kis_node_manager.h"
-#include "kis_painting_assistants_manager.h"
+#include "KisDecorationsManager.h"
 #include <kis_paint_layer.h>
 #include "kis_paintop_box.h"
 #include <brushengine/kis_paintop_preset.h>
 #include "KisPart.h"
 #include "KisPrintJob.h"
-#include "kis_progress_widget.h"
-#include "kis_resource_server_provider.h"
+#include <KoUpdater.h>
+#include "KisResourceServerProvider.h"
 #include "kis_selection.h"
+#include "kis_selection_mask.h"
 #include "kis_selection_manager.h"
 #include "kis_shape_controller.h"
 #include "kis_shape_layer.h"
@@ -123,11 +126,12 @@
 #include "kis_zoom_manager.h"
 #include "widgets/kis_floating_message.h"
 #include "kis_signal_auto_connection.h"
-#include "kis_script_manager.h"
 #include "kis_icon_utils.h"
 #include "kis_guides_manager.h"
 #include "kis_derived_resources.h"
 #include "dialogs/kis_delayed_save_dialog.h"
+#include <KisMainWindow.h>
+#include "kis_signals_blocker.h"
 
 
 class BlockingUserInputEventFilter : public QObject
@@ -185,17 +189,10 @@ public:
         , actionCollection(_actionCollection)
         , mirrorManager(_q)
         , inputManager(_q)
-        , scriptManager(_q)
         , actionAuthor(0)
+        , showPixelGrid(0)
     {
-        canvasResourceManager.addDerivedResourceConverter(toQShared(new KisCompositeOpResourceConverter));
-        canvasResourceManager.addDerivedResourceConverter(toQShared(new KisEffectiveCompositeOpResourceConverter));
-        canvasResourceManager.addDerivedResourceConverter(toQShared(new KisOpacityResourceConverter));
-        canvasResourceManager.addDerivedResourceConverter(toQShared(new KisFlowResourceConverter));
-        canvasResourceManager.addDerivedResourceConverter(toQShared(new KisSizeResourceConverter));
-        canvasResourceManager.addDerivedResourceConverter(toQShared(new KisLodAvailabilityResourceConverter));
-        canvasResourceManager.addDerivedResourceConverter(toQShared(new KisEraserModeResourceConverter));
-        canvasResourceManager.addResourceUpdateMediator(toQShared(new KisPresetUpdateMediator));
+        KisViewManager::initializeResourceManager(&canvasResourceManager);
     }
 
 public:
@@ -217,17 +214,23 @@ public:
     KisAction *zoomOut;
     KisAction *softProof;
     KisAction *gamutCheck;
+    KisAction *toggleFgBg;
+    KisAction *resetFgBg;
 
     KisSelectionManager selectionManager;
     KisGuidesManager guidesManager;
     KisStatusBar statusBar;
+    QPointer<KoUpdater> persistentImageProgressUpdater;
+
+    QScopedPointer<KoProgressUpdater> persistentUnthreadedProgressUpdaterRouter;
+    QPointer<KoUpdater> persistentUnthreadedProgressUpdater;
 
     KisControlFrame controlFrame;
     KisNodeManager nodeManager;
     KisImageManager imageManager;
     KisGridManager gridManager;
     KisCanvasControlsManager canvasControlsManager;
-    KisPaintingAssistantsManager paintingAssistantsManager;
+    KisDecorationsManager paintingAssistantsManager;
     BlockingUserInputEventFilter blockingEventFilter;
     KisActionManager actionManager;
     QMainWindow* mainWindow;
@@ -235,22 +238,24 @@ public:
     bool showFloatingMessage;
     QPointer<KisView> currentImageView;
     KisCanvasResourceProvider canvasResourceProvider;
-    KoCanvasResourceManager canvasResourceManager;
+    KoCanvasResourceProvider canvasResourceManager;
     KisSignalCompressor guiUpdateCompressor;
     KActionCollection *actionCollection;
     KisMirrorManager mirrorManager;
     KisInputManager inputManager;
 
     KisSignalAutoConnectionsStore viewConnections;
-    KisScriptManager scriptManager;
     KSelectAction *actionAuthor; // Select action for author profile.
+    KisAction *showPixelGrid;
 
     QByteArray canvasState;
+#if QT_VERSION < QT_VERSION_CHECK(5, 10, 0)
+    QFlags<Qt::WindowState> windowFlags;
+#endif
 
     bool blockUntilOperationsFinishedImpl(KisImageSP image, bool force);
 };
 
-
 KisViewManager::KisViewManager(QWidget *parent, KActionCollection *_actionCollection)
     : d(new KisViewManagerPrivate(this, _actionCollection, parent))
 {
@@ -264,8 +269,26 @@ KisViewManager::KisViewManager(QWidget *parent, KActionCollection *_actionCollec
 
     // These initialization functions must wait until KisViewManager ctor is complete.
     d->statusBar.setup();
+    d->persistentImageProgressUpdater =
+            d->statusBar.progressUpdater()->startSubtask(1, "", true);
+    // reset state to "completed"
+    d->persistentImageProgressUpdater->setRange(0,100);
+    d->persistentImageProgressUpdater->setValue(100);
+
+    d->persistentUnthreadedProgressUpdater =
+            d->statusBar.progressUpdater()->startSubtask(1, "", true);
+    // reset state to "completed"
+    d->persistentUnthreadedProgressUpdater->setRange(0,100);
+    d->persistentUnthreadedProgressUpdater->setValue(100);
+
+    d->persistentUnthreadedProgressUpdaterRouter.reset(
+                new KoProgressUpdater(d->persistentUnthreadedProgressUpdater,
+                                      KoProgressUpdater::Unthreaded));
+    d->persistentUnthreadedProgressUpdaterRouter->setAutoNestNames(true);
+
     d->controlFrame.setup(parent);
 
+
     //Check to draw scrollbars after "Canvas only mode" toggle is created.
     this->showHideScrollbars();
 
@@ -281,39 +304,53 @@ KisViewManager::KisViewManager(QWidget *parent, KActionCollection *_actionCollec
             d->controlFrame.paintopBox(), SLOT(slotToolChanged(KoCanvasController*,int)));
 
     connect(&d->nodeManager, SIGNAL(sigNodeActivated(KisNodeSP)),
-            resourceProvider(), SLOT(slotNodeActivated(KisNodeSP)));
-
-    connect(resourceProvider()->resourceManager(), SIGNAL(canvasResourceChanged(int,QVariant)),
-            d->controlFrame.paintopBox(), SLOT(slotCanvasResourceChanged(int,QVariant)));
+            canvasResourceProvider(), SLOT(slotNodeActivated(KisNodeSP)));
 
     connect(KisPart::instance(), SIGNAL(sigViewAdded(KisView*)), SLOT(slotViewAdded(KisView*)));
     connect(KisPart::instance(), SIGNAL(sigViewRemoved(KisView*)), SLOT(slotViewRemoved(KisView*)));
 
     connect(KisConfigNotifier::instance(), SIGNAL(configChanged()), SLOT(slotUpdateAuthorProfileActions()));
+    connect(KisConfigNotifier::instance(), SIGNAL(pixelGridModeChanged()), SLOT(slotUpdatePixelGridAction()));
 
     KisInputProfileManager::instance()->loadProfiles();
 
-    KisConfig cfg;
+    KisConfig cfg(true);
     d->showFloatingMessage = cfg.showCanvasMessages();
-
+    const KoColorSpace *cs = KoColorSpaceRegistry::instance()->rgb8();
+    KoColor foreground(Qt::black, cs);
+    d->canvasResourceProvider.setFGColor(cfg.readKoColor("LastForeGroundColor",foreground));
+    KoColor background(Qt::white, cs);
+    d->canvasResourceProvider.setBGColor(cfg.readKoColor("LastBackGroundColor",background));
 }
 
 
 KisViewManager::~KisViewManager()
 {
-    KisConfig cfg;
-    if (resourceProvider() && resourceProvider()->currentPreset()) {
-        cfg.writeEntry("LastPreset", resourceProvider()->currentPreset()->name());
-        cfg.writeKoColor("LastForeGroundColor",resourceProvider()->fgColor());
-        cfg.writeKoColor("LastBackGroundColor",resourceProvider()->bgColor());
-
+    KisConfig cfg(false);
+    if (canvasResourceProvider() && canvasResourceProvider()->currentPreset()) {
+        cfg.writeKoColor("LastForeGroundColor",canvasResourceProvider()->fgColor());
+        cfg.writeKoColor("LastBackGroundColor",canvasResourceProvider()->bgColor());
     }
 
     cfg.writeEntry("baseLength", KoResourceItemChooserSync::instance()->baseLength());
-
+    cfg.writeEntry("CanvasOnlyActive", false); // We never restart in CavnasOnlyMode
     delete d;
 }
 
+void KisViewManager::initializeResourceManager(KoCanvasResourceProvider *resourceManager)
+{
+    resourceManager->addDerivedResourceConverter(toQShared(new KisCompositeOpResourceConverter));
+    resourceManager->addDerivedResourceConverter(toQShared(new KisEffectiveCompositeOpResourceConverter));
+    resourceManager->addDerivedResourceConverter(toQShared(new KisOpacityResourceConverter));
+    resourceManager->addDerivedResourceConverter(toQShared(new KisFlowResourceConverter));
+    resourceManager->addDerivedResourceConverter(toQShared(new KisSizeResourceConverter));
+    resourceManager->addDerivedResourceConverter(toQShared(new KisLodAvailabilityResourceConverter));
+    resourceManager->addDerivedResourceConverter(toQShared(new KisLodSizeThresholdResourceConverter));
+    resourceManager->addDerivedResourceConverter(toQShared(new KisLodSizeThresholdSupportedResourceConverter));
+    resourceManager->addDerivedResourceConverter(toQShared(new KisEraserModeResourceConverter));
+    resourceManager->addResourceUpdateMediator(toQShared(new KisPresetUpdateMediator));
+}
+
 KActionCollection *KisViewManager::actionCollection() const
 {
     return d->actionCollection;
@@ -321,7 +358,9 @@ KActionCollection *KisViewManager::actionCollection() const
 
 void KisViewManager::slotViewAdded(KisView *view)
 {
-    d->inputManager.addTrackedCanvas(view->canvasBase());
+    // WARNING: this slot is called even when a view from another main windows is added!
+    //          Don't expect \p view be a child of this view manager!
+    Q_UNUSED(view);
 
     if (viewCount() == 0) {
         d->statusBar.showAllStatusBarItems();
@@ -330,11 +369,18 @@ void KisViewManager::slotViewAdded(KisView *view)
 
 void KisViewManager::slotViewRemoved(KisView *view)
 {
-    d->inputManager.removeTrackedCanvas(view->canvasBase());
+    // WARNING: this slot is called even when a view from another main windows is removed!
+    //          Don't expect \p view be a child of this view manager!
+    Q_UNUSED(view);
 
     if (viewCount() == 0) {
         d->statusBar.hideAllStatusBarItems();
     }
+
+    KisConfig cfg(false);
+    if (canvasResourceProvider() && canvasResourceProvider()->currentPreset()) {
+        cfg.writeEntry("LastPreset", canvasResourceProvider()->currentPreset()->name());
+    }
 }
 
 void KisViewManager::setCurrentView(KisView *view)
@@ -347,6 +393,7 @@ void KisViewManager::setCurrentView(KisView *view)
         first = false;
         KisDocument* doc = d->currentImageView->document();
         if (doc) {
+            doc->image()->compositeProgressProxy()->removeProxy(d->persistentImageProgressUpdater);
             doc->disconnect(this);
         }
         d->currentImageView->canvasController()->proxyObject->disconnect(&d->statusBar);
@@ -354,40 +401,52 @@ void KisViewManager::setCurrentView(KisView *view)
     }
 
 
-    d->softProof->setChecked(view->softProofing());
-    d->gamutCheck->setChecked(view->gamutCheck());
-
-    QPointer<KisView>imageView = qobject_cast<KisView*>(view);
+    QPointer<KisView> imageView = qobject_cast<KisView*>(view);
+    d->currentImageView = imageView;
 
     if (imageView) {
+        d->softProof->setChecked(imageView->softProofing());
+        d->gamutCheck->setChecked(imageView->gamutCheck());
+
         // Wait for the async image to have loaded
         KisDocument* doc = view->document();
-        //        connect(canvasController()->proxyObject, SIGNAL(documentMousePositionChanged(QPointF)), d->statusBar, SLOT(documentMousePositionChanged(QPointF)));
 
-        d->currentImageView = imageView;
+        if (KisConfig(true).readEntry<bool>("EnablePositionLabel", false)) {
+            connect(d->currentImageView->canvasController()->proxyObject,
+                    SIGNAL(documentMousePositionChanged(QPointF)),
+                    &d->statusBar,
+                    SLOT(documentMousePositionChanged(QPointF)));
+        }
+
         // Restore the last used brush preset, color and background color.
         if (first) {
-            KisConfig cfg;
             KisPaintOpPresetResourceServer * rserver = KisResourceServerProvider::instance()->paintOpPresetServer();
-            QString lastPreset = cfg.readEntry("LastPreset", QString("Basic_tip_default"));
+            QString defaultPresetName = "basic_tip_default";
+            bool foundTip = false;
+            for (int i=0; i<rserver->resourceCount(); i++) {
+                KisPaintOpPresetSP resource = rserver->resources().at(i);
+                if (resource->name().toLower().contains("basic_tip_default")) {
+                    defaultPresetName = resource->name();
+                    foundTip = true;
+                } else if (foundTip == false && (resource->name().toLower().contains("default") ||
+                                                 resource->filename().toLower().contains("default"))) {
+                    defaultPresetName = resource->name();
+                    foundTip = true;
+                }
+            }
+            KisConfig cfg(true);
+            QString lastPreset = cfg.readEntry("LastPreset", defaultPresetName);
             KisPaintOpPresetSP preset = rserver->resourceByName(lastPreset);
             if (!preset) {
-                preset = rserver->resourceByName("Basic_tip_default");
+                preset = rserver->resourceByName(defaultPresetName);
             }
 
-            if (!preset) {
+            if (!preset && !rserver->resources().isEmpty()) {
                 preset = rserver->resources().first();
             }
             if (preset) {
                 paintOpBox()->restoreResource(preset.data());
             }
-
-            const KoColorSpace *cs = KoColorSpaceRegistry::instance()->rgb8();
-            KoColor foreground(Qt::black, cs);
-            d->canvasResourceProvider.setFGColor(cfg.readKoColor("LastForeGroundColor",foreground));
-            KoColor background(Qt::white, cs);
-            d->canvasResourceProvider.setBGColor(cfg.readKoColor("LastBackGroundColor",background));
-
         }
 
         KisCanvasController *canvasController = dynamic_cast<KisCanvasController*>(d->currentImageView->canvasController());
@@ -413,6 +472,12 @@ void KisViewManager::setCurrentView(KisView *view)
         d->viewConnections.addUniqueConnection(d->softProof, SIGNAL(toggled(bool)), view, SLOT(slotSoftProofing(bool)) );
         d->viewConnections.addUniqueConnection(d->gamutCheck, SIGNAL(toggled(bool)), view, SLOT(slotGamutCheck(bool)) );
 
+        // set up progrress reporting
+        doc->image()->compositeProgressProxy()->addProxy(d->persistentImageProgressUpdater);
+        d->viewConnections.addUniqueConnection(&d->statusBar, SIGNAL(sigCancellationRequested()), doc->image(), SLOT(requestStrokeCancellation()));
+
+        d->viewConnections.addUniqueConnection(d->showPixelGrid, SIGNAL(toggled(bool)), canvasController, SLOT(slotTogglePixelGrid(bool)));
+
         imageView->zoomManager()->setShowRulers(d->showRulersAction->isChecked());
         imageView->zoomManager()->setRulersTrackMouse(d->rulersTrackMouseAction->isChecked());
 
@@ -435,30 +500,30 @@ void KisViewManager::setCurrentView(KisView *view)
         d->currentImageView->notifyCurrentStateChanged(true);
         d->currentImageView->canvasController()->activate();
         d->currentImageView->canvasController()->setFocus();
+
+        d->viewConnections.addUniqueConnection(
+                    image(), SIGNAL(sigSizeChanged(QPointF,QPointF)),
+                    canvasResourceProvider(), SLOT(slotImageSizeChanged()));
+
+        d->viewConnections.addUniqueConnection(
+                    image(), SIGNAL(sigResolutionChanged(double,double)),
+                    canvasResourceProvider(), SLOT(slotOnScreenResolutionChanged()));
+
+        d->viewConnections.addUniqueConnection(
+                    image(), SIGNAL(sigNodeChanged(KisNodeSP)),
+                    this, SLOT(updateGUI()));
+
+        d->viewConnections.addUniqueConnection(
+                    d->currentImageView->zoomManager()->zoomController(),
+                    SIGNAL(zoomChanged(KoZoomMode::Mode,qreal)),
+                    canvasResourceProvider(), SLOT(slotOnScreenResolutionChanged()));
+
     }
 
     d->actionManager.updateGUI();
 
-    d->viewConnections.addUniqueConnection(
-        image(), SIGNAL(sigSizeChanged(const QPointF&, const QPointF&)),
-        resourceProvider(), SLOT(slotImageSizeChanged()));
-
-    d->viewConnections.addUniqueConnection(
-        image(), SIGNAL(sigResolutionChanged(double,double)),
-        resourceProvider(), SLOT(slotOnScreenResolutionChanged()));
-
-    d->viewConnections.addUniqueConnection(
-        image(), SIGNAL(sigNodeChanged(KisNodeSP)),
-        this, SLOT(updateGUI()));
-
-    d->viewConnections.addUniqueConnection(
-        view->zoomManager()->zoomController(),
-        SIGNAL(zoomChanged(KoZoomMode::Mode,qreal)),
-        resourceProvider(), SLOT(slotOnScreenResolutionChanged()));
-
-    resourceProvider()->slotImageSizeChanged();
-    resourceProvider()->slotOnScreenResolutionChanged();
-
+    canvasResourceProvider()->slotImageSizeChanged();
+    canvasResourceProvider()->slotOnScreenResolutionChanged();
 
     Q_EMIT viewChanged();
 }
@@ -480,7 +545,7 @@ KisImageWSP KisViewManager::image() const
     return 0;
 }
 
-KisCanvasResourceProvider * KisViewManager::resourceProvider()
+KisCanvasResourceProvider * KisViewManager::canvasResourceProvider()
 {
     return &d->canvasResourceProvider;
 }
@@ -506,24 +571,19 @@ KisStatusBar * KisViewManager::statusBar() const
     return &d->statusBar;
 }
 
-void KisViewManager::addStatusBarItem(QWidget *widget, int stretch, bool permanent)
-{
-    d->statusBar.addStatusBarItem(widget, stretch, permanent);
-}
-
-void KisViewManager::removeStatusBarItem(QWidget *widget)
-{
-    d->statusBar.removeStatusBarItem(widget);
-}
-
 KisPaintopBox* KisViewManager::paintOpBox() const
 {
     return d->controlFrame.paintopBox();
 }
 
-KoProgressUpdater* KisViewManager::createProgressUpdater(KoProgressUpdater::Mode mode)
+QPointer<KoUpdater> KisViewManager::createUnthreadedUpdater(const QString &name)
 {
-    return new KisProgressUpdater(d->statusBar.progress(), document()->progressProxy(), mode);
+    return d->persistentUnthreadedProgressUpdaterRouter->startSubtask(1, name, false);
+}
+
+QPointer<KoUpdater> KisViewManager::createThreadedUpdater(const QString &name)
+{
+    return d->statusBar.progressUpdater()->startSubtask(1, name, false);
 }
 
 KisSelectionManager * KisViewManager::selectionManager()
@@ -582,10 +642,9 @@ bool KisViewManager::selectionEditable()
 {
     KisLayerSP layer = activeLayer();
     if (layer) {
-        KoProperties properties;
-        QList<KisNodeSP> masks = layer->childNodes(QStringList("KisSelectionMask"), properties);
-        if (masks.size() == 1) {
-            return masks[0]->isEditable();
+        KisSelectionMaskSP mask = layer->selectionMask();
+        if (mask) {
+            return mask->isEditable();
         }
     }
     // global selection is always editable
@@ -604,7 +663,7 @@ KisUndoAdapter * KisViewManager::undoAdapter()
 
 void KisViewManager::createActions()
 {
-    KisConfig cfg;
+    KisConfig cfg(true);
 
     d->saveIncremental = actionManager()->createAction("save_incremental_version");
     connect(d->saveIncremental, SIGNAL(triggered()), this, SLOT(slotSaveIncremental()));
@@ -654,6 +713,7 @@ void KisViewManager::createActions()
     a = actionManager()->createAction("edit_blacklist_cleanup");
     connect(a, SIGNAL(triggered()), this, SLOT(slotBlacklistCleanup()));
 
+    actionManager()->createAction("ruler_pixel_multiple2");
     d->showRulersAction = actionManager()->createAction("view_ruler");
     d->showRulersAction->setChecked(cfg.showRulers());
     connect(d->showRulersAction, SIGNAL(toggled(bool)), SLOT(slotSaveShowRulersState(bool)));
@@ -668,10 +728,19 @@ void KisViewManager::createActions()
     d->zoomOut = actionManager()->createStandardAction(KStandardAction::ZoomOut, 0, "");
 
     d->actionAuthor  = new KSelectAction(KisIconUtils::loadIcon("im-user"), i18n("Active Author Profile"), this);
-    connect(d->actionAuthor, SIGNAL(triggered(const QString &)), this, SLOT(changeAuthorProfile(const QString &)));
+    connect(d->actionAuthor, SIGNAL(triggered(QString)), this, SLOT(changeAuthorProfile(QString)));
     actionCollection()->addAction("settings_active_author", d->actionAuthor);
     slotUpdateAuthorProfileActions();
 
+    d->showPixelGrid = actionManager()->createAction("view_pixel_grid");
+    slotUpdatePixelGridAction();
+
+    d->toggleFgBg = actionManager()->createAction("toggle_fg_bg");
+    connect(d->toggleFgBg, SIGNAL(triggered(bool)), this, SLOT(slotToggleFgBg()));
+
+    d->resetFgBg =  actionManager()->createAction("reset_fg_bg");
+    connect(d->resetFgBg, SIGNAL(triggered(bool)), this, SLOT(slotResetFgBg()));
+
 }
 
 void KisViewManager::setupManagers()
@@ -697,8 +766,6 @@ void KisViewManager::setupManagers()
     d->canvasControlsManager.setup(actionManager());
 
     d->mirrorManager.setup(actionCollection());
-
-    d->scriptManager.setup(actionCollection(), actionManager());
 }
 
 void KisViewManager::updateGUI()
@@ -740,11 +807,6 @@ KisDocument *KisViewManager::document() const
     return 0;
 }
 
-KisScriptManager *KisViewManager::scriptManager() const
-{
-  return &d->scriptManager;
-}
-
 int KisViewManager::viewCount() const
 {
     KisMainWindow *mw  = qobject_cast<KisMainWindow*>(d->mainWindow);
@@ -782,19 +844,25 @@ void KisViewManager::slotCreateTemplate()
 
 void KisViewManager::slotCreateCopy()
 {
-    if (!document()) return;
-    KisDocument *doc = KisPart::instance()->createDocument();
+    KisDocument *srcDoc = document();
+    if (!srcDoc) return;
 
-    QString name = document()->documentInfo()->aboutInfo("name");
+    if (!this->blockUntilOperationsFinished(srcDoc->image())) return;
+
+    KisDocument *doc = 0;
+    {
+        KisImageBarrierLocker l(srcDoc->image());
+        doc = srcDoc->clone();
+    }
+    KIS_SAFE_ASSERT_RECOVER_RETURN(doc);
+
+    QString name = srcDoc->documentInfo()->aboutInfo("name");
     if (name.isEmpty()) {
         name = document()->url().toLocalFile();
     }
     name = i18n("%1 (Copy)", name);
     doc->documentInfo()->setAboutInfo("title", name);
-    KisImageWSP image = document()->image();
-    KisImageSP newImage = new KisImage(doc->createUndoStore(), image->width(), image->height(), image->colorSpace(), name);
-    newImage->setRootLayer(dynamic_cast<KisGroupLayer*>(image->rootLayer()->clone().data()));
-    doc->setCurrentImage(newImage);
+
     KisPart::instance()->addDocument(doc);
     KisMainWindow *mw  = qobject_cast<KisMainWindow*>(d->mainWindow);
     mw->addViewAndNotifyLoadingCompleted(doc);
@@ -819,7 +887,6 @@ void KisViewManager::setQtMainWindow(QMainWindow* newMainWindow)
     d->mainWindow = newMainWindow;
 }
 
-
 void KisViewManager::slotDocumentSaved()
 {
     d->saveIncremental->setEnabled(true);
@@ -830,6 +897,12 @@ void KisViewManager::slotSaveIncremental()
 {
     if (!document()) return;
 
+    if (document()->url().isEmpty()) {
+        KisMainWindow *mw = qobject_cast<KisMainWindow*>(d->mainWindow);
+        mw->saveDocument(document(), true, false);
+        return;
+    }
+
     bool foundVersion;
     bool fileAlreadyExists;
     bool isBackup;
@@ -866,6 +939,7 @@ void KisViewManager::slotSaveIncremental()
         }
         version.remove(0, 1);            //  Trim "_"
     } else {
+        // TODO: this will not work with files extensions like jp2
         // ...else, simply add a version to it so the next loop works
         QRegExp regex2("[.][a-z]{2,4}$");  //  Heuristic to find file extension
         regex2.indexIn(fileName);
@@ -912,7 +986,7 @@ void KisViewManager::slotSaveIncremental()
         return;
     }
     document()->setFileBatchMode(true);
-    document()->saveAs(QUrl::fromUserInput(fileName));
+    document()->saveAs(QUrl::fromUserInput(fileName), document()->mimeType(), true);
     document()->setFileBatchMode(false);
 
     if (mainWindow()) {
@@ -925,6 +999,12 @@ void KisViewManager::slotSaveIncrementalBackup()
 {
     if (!document()) return;
 
+    if (document()->url().isEmpty()) {
+        KisMainWindow *mw = qobject_cast<KisMainWindow*>(d->mainWindow);
+        mw->saveDocument(document(), true, false);
+        return;
+    }
+
     bool workingOnBackup;
     bool fileAlreadyExists;
     QString version = "000";
@@ -983,7 +1063,7 @@ void KisViewManager::slotSaveIncrementalBackup()
             return;
         }
         QFile::copy(fileName, backupFileName);
-        document()->saveAs(QUrl::fromUserInput(fileName));
+        document()->saveAs(QUrl::fromUserInput(fileName), document()->mimeType(), true);
 
         if (mainWindow()) mainWindow()->updateCaption();
     }
@@ -1021,7 +1101,7 @@ void KisViewManager::slotSaveIncrementalBackup()
         // Save both as backup and on current file for interapplication workflow
         document()->setFileBatchMode(true);
         QFile::copy(fileName, backupFileName);
-        document()->saveAs(QUrl::fromUserInput(fileName));
+        document()->saveAs(QUrl::fromUserInput(fileName), document()->mimeType(), true);
         document()->setFileBatchMode(false);
 
         if (mainWindow()) mainWindow()->updateCaption();
@@ -1053,23 +1133,28 @@ void KisViewManager::showStatusBar(bool toggled)
     KisMainWindow *mw = mainWindow();
     if(mw && mw->statusBar()) {
         mw->statusBar()->setVisible(toggled);
-        KisConfig cfg;
+        KisConfig cfg(false);
         cfg.setShowStatusBar(toggled);
     }
 }
 
 void KisViewManager::switchCanvasOnly(bool toggled)
 {
-    KisConfig cfg;
-    KisMainWindow* main = mainWindow();
+    KisConfig cfg(false);
+    KisMainWindow *main = mainWindow();
 
     if(!main) {
         dbgUI << "Unable to switch to canvas-only mode, main window not found";
         return;
     }
 
+    cfg.writeEntry("CanvasOnlyActive", toggled);
+
     if (toggled) {
         d->canvasState = qtMainWindow()->saveState();
+#if QT_VERSION < QT_VERSION_CHECK(5, 10, 0)
+        d->windowFlags = main->windowState();
+#endif
     }
 
     if (cfg.hideStatusbarFullscreen()) {
@@ -1105,11 +1190,19 @@ void KisViewManager::switchCanvasOnly(bool toggled)
         }
     }
 
+    // QT in windows does not return to maximized upon 4th tab in a row
+    // https://bugreports.qt.io/browse/QTBUG-57882, https://bugreports.qt.io/browse/QTBUG-52555, https://codereview.qt-project.org/#/c/185016/
     if (cfg.hideTitlebarFullscreen() && !cfg.fullscreenMode()) {
         if(toggled) {
             main->setWindowState( main->windowState() | Qt::WindowFullScreen);
         } else {
             main->setWindowState( main->windowState() & ~Qt::WindowFullScreen);
+#if QT_VERSION < QT_VERSION_CHECK(5, 10, 0)
+            // If window was maximized prior to fullscreen, restore that
+            if (d->windowFlags & Qt::WindowMaximized) {
+                main->setWindowState( main->windowState() | Qt::WindowMaximized);
+            }
+#endif
         }
     }
 
@@ -1174,18 +1267,11 @@ void KisViewManager::updateIcons()
     if (mainWindow()) {
         QList<QDockWidget*> dockers = mainWindow()->dockWidgets();
         Q_FOREACH (QDockWidget* dock, dockers) {
-            dbgKrita << "name " << dock->objectName();
-            KoDockWidgetTitleBar* titlebar = dynamic_cast<KoDockWidgetTitleBar*>(dock->titleBarWidget());
-            if (titlebar) {
-                titlebar->updateIcons();
-            }
-
             QObjectList objects;
             objects.append(dock);
             while (!objects.isEmpty()) {
                 QObject* object = objects.takeFirst();
                 objects.append(object->children());
-
                 KisIconUtils::updateIconCommon(object);
             }
         }
@@ -1193,7 +1279,7 @@ void KisViewManager::updateIcons()
 }
 void KisViewManager::initializeStatusBarVisibility()
 {
-    KisConfig cfg;
+    KisConfig cfg(true);
     d->mainWindow->statusBar()->setVisible(cfg.showStatusBar());
 }
 
@@ -1212,7 +1298,7 @@ void KisViewManager::guiUpdateTimeout()
 void KisViewManager::showFloatingMessage(const QString &message, const QIcon& icon, int timeout, KisFloatingMessage::Priority priority, int alignment)
 {
     if (!d->currentImageView) return;
-    d->currentImageView->showFloatingMessageImpl(message, icon, timeout, priority, alignment);
+    d->currentImageView->showFloatingMessage(message, icon, timeout, priority, alignment);
 
     emit floatingMessageRequested(message, icon.name());
 }
@@ -1228,7 +1314,7 @@ void KisViewManager::showHideScrollbars()
     if (!d->currentImageView) return;
     if (!d->currentImageView->canvasController()) return;
 
-    KisConfig cfg;
+    KisConfig cfg(true);
     bool toggled = actionCollection()->action("view_show_canvas_only")->isChecked();
 
     if ( (toggled && cfg.hideScrollbarsFullscreen()) || (!toggled && cfg.hideScrollbars()) ) {
@@ -1242,13 +1328,13 @@ void KisViewManager::showHideScrollbars()
 
 void KisViewManager::slotSaveShowRulersState(bool value)
 {
-    KisConfig cfg;
+    KisConfig cfg(false);
     cfg.setShowRulers(value);
 }
 
 void KisViewManager::slotSaveRulersTrackMouseState(bool value)
 {
-    KisConfig cfg;
+    KisConfig cfg(false);
     cfg.setRulersTrackMouse(value);
 }
 
@@ -1259,11 +1345,9 @@ void KisViewManager::setShowFloatingMessage(bool show)
 
 void KisViewManager::changeAuthorProfile(const QString &profileName)
 {
-    KConfigGroup appAuthorGroup(KoGlobal::calligraConfig(), "Author");
-    if (profileName.isEmpty()) {
+    KConfigGroup appAuthorGroup(KSharedConfig::openConfig(), "Author");
+    if (profileName.isEmpty() || profileName == i18nc("choice for author profile", "Anonymous")) {
         appAuthorGroup.writeEntry("active-profile", "");
-    } else if (profileName == i18nc("choice for author profile", "Anonymous")) {
-        appAuthorGroup.writeEntry("active-profile", "anonymous");
     } else {
         appAuthorGroup.writeEntry("active-profile", profileName);
     }
@@ -1280,23 +1364,79 @@ void KisViewManager::slotUpdateAuthorProfileActions()
         return;
     }
     d->actionAuthor->clear();
-    d->actionAuthor->addAction(i18n("Default Author Profile"));
     d->actionAuthor->addAction(i18nc("choice for author profile", "Anonymous"));
 
-    KConfigGroup authorGroup(KoGlobal::calligraConfig(), "Author");
+    KConfigGroup authorGroup(KSharedConfig::openConfig(), "Author");
     QStringList profiles = authorGroup.readEntry("profile-names", QStringList());
+    QString authorInfo = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation) + "/authorinfo/";
+    QStringList filters = QStringList() << "*.authorinfo";
+    QDir dir(authorInfo);
+    Q_FOREACH(QString entry, dir.entryList(filters)) {
+        int ln = QString(".authorinfo").size();
+        entry.chop(ln);
+        if (!profiles.contains(entry)) {
+            profiles.append(entry);
+        }
+    }
     Q_FOREACH (const QString &profile , profiles) {
         d->actionAuthor->addAction(profile);
     }
 
-    KConfigGroup appAuthorGroup(KoGlobal::calligraConfig(), "Author");
+    KConfigGroup appAuthorGroup(KSharedConfig::openConfig(), "Author");
     QString profileName = appAuthorGroup.readEntry("active-profile", "");
 
-    if (profileName == "anonymous") {
-        d->actionAuthor->setCurrentItem(1);
+    if (profileName == "anonymous" || profileName.isEmpty()) {
+        d->actionAuthor->setCurrentItem(0);
     } else if (profiles.contains(profileName)) {
         d->actionAuthor->setCurrentAction(profileName);
-    } else {
-        d->actionAuthor->setCurrentItem(0);
     }
 }
+
+void KisViewManager::slotUpdatePixelGridAction()
+{
+    KIS_SAFE_ASSERT_RECOVER_RETURN(d->showPixelGrid);
+
+    KisSignalsBlocker b(d->showPixelGrid);
+
+    KisConfig cfg(true);
+    d->showPixelGrid->setChecked(cfg.pixelGridEnabled() && cfg.useOpenGL());
+}
+
+void KisViewManager::slotActivateTransformTool()
+{
+    if(KoToolManager::instance()->activeToolId() == "KisToolTransform") {
+        KoToolBase* tool = KoToolManager::instance()->toolById(canvasBase(), "KisToolTransform");
+
+        QSet<KoShape*> dummy;
+        // Start a new stroke
+        tool->deactivate();
+        tool->activate(KoToolBase::DefaultActivation, dummy);
+    }
+
+    KoToolManager::instance()->switchToolRequested("KisToolTransform");
+}
+
+void KisViewManager::slotToggleFgBg()
+{
+
+    KoColor newFg = d->canvasResourceManager.backgroundColor();
+    KoColor newBg = d->canvasResourceManager.foregroundColor();
+
+    /**
+     * NOTE: Some of color selectors do not differentiate foreground
+     *       and background colors, so if one wants them to end up
+     *       being set up to foreground color, it should be set the
+     *       last.
+     */
+    d->canvasResourceManager.setBackgroundColor(newBg);
+    d->canvasResourceManager.setForegroundColor(newFg);
+}
+
+void KisViewManager::slotResetFgBg()
+{
+    // see a comment in slotToggleFgBg()
+    d->canvasResourceManager.setBackgroundColor(KoColor(Qt::white, KoColorSpaceRegistry::instance()->rgb8()));
+    d->canvasResourceManager.setForegroundColor(KoColor(Qt::black, KoColorSpaceRegistry::instance()->rgb8()));
+}
+
+
