@@ -24,6 +24,8 @@
 #include <QQueue>
 #include <klocalizedstring.h>
 #include <QApplication>
+#include <QTouchEvent>
+#include <QElapsedTimer>
 
 #include <KoToolManager.h>
 
@@ -65,32 +67,18 @@ uint qHash(QPointer<T> value) {
     return reinterpret_cast<quintptr>(value.data());
 }
 
-
-#define start_ignore_cursor_events() d->blockMouseEvents()
-#define stop_ignore_cursor_events() d->allowMouseEvents()
-#define break_if_should_ignore_cursor_events() if (d->ignoringQtCursorEvents()) break;
-#define break_if_tablet_active() if (d->tabletActive) break;
-
-// Touch rejection: if touch is disabled on canvas, no need to block mouse press events
-#define touch_start_block_press_events()  d->touchHasBlockedPressEvents = d->disableTouchOnCanvas;
-#define touch_stop_block_press_events()  d->touchHasBlockedPressEvents = false;
-#define break_if_touch_blocked_press_events() if (d->touchHasBlockedPressEvents) break;
-#define touch_eat_one_mouse_press() if (d->disableTouchOnCanvas) d->eatOneMousePress();
-
 KisInputManager::KisInputManager(QObject *parent)
     : QObject(parent), d(new Private(this))
 {
     d->setupActions();
 
-    connect(KoToolManager::instance(), SIGNAL(aboutToChangeTool(KoCanvasController*)),
-            SLOT(slotAboutToChangeTool()));
-    connect(KoToolManager::instance(), SIGNAL(changedTool(KoCanvasController*,int)),
-            SLOT(slotToolChanged()));
+    connect(KoToolManager::instance(), SIGNAL(aboutToChangeTool(KoCanvasController*)), SLOT(slotAboutToChangeTool()));
+    connect(KoToolManager::instance(), SIGNAL(changedTool(KoCanvasController*,int)), SLOT(slotToolChanged()));
     connect(&d->moveEventCompressor, SIGNAL(timeout()), SLOT(slotCompressedMoveEvent()));
 
 
     QApplication::instance()->
-        installEventFilter(new Private::ProximityNotifier(d, this));
+            installEventFilter(new Private::ProximityNotifier(d, this));
 }
 
 KisInputManager::~KisInputManager()
@@ -159,13 +147,7 @@ void KisInputManager::setupAsEventFilter(QObject *receiver)
 
 void KisInputManager::stopIgnoringEvents()
 {
-    stop_ignore_cursor_events();
-}
-
-void KisInputManager::slotFocusOnEnter(bool value)
-{
-    Q_UNUSED(value);
-    // not used anymore
+    d->allowMouseEvents();
 }
 
 #if defined (__clang__)
@@ -210,13 +192,47 @@ bool KisInputManager::eventFilter(QObject* object, QEvent* event)
         // global shortcuts (not the input manager's ones) are not
         // executed, in particular, this line will accept events when the
         // tool is in text editing, preventing shortcut triggering
-        d->toolProxy->processEvent(event);
+        if (d->toolProxy) {
+            d->toolProxy->processEvent(event);
+        }
     }
 
     // Continue with the actual switch statement...
     return eventFilterImpl(event);
 }
 
+// Qt's events do not have copy-ctors yes, so we should emulate them
+// See https://bugreports.qt.io/browse/QTBUG-72488
+
+template <class Event> void copyEventHack(Event *src, QScopedPointer<QEvent> &dst);
+
+template<> void copyEventHack(QMouseEvent *src, QScopedPointer<QEvent> &dst) {
+    QMouseEvent *tmp = new QMouseEvent(src->type(),
+                                       src->localPos(), src->windowPos(), src->screenPos(),
+                                       src->button(), src->buttons(), src->modifiers(),
+                                       src->source());
+    tmp->setTimestamp(src->timestamp());
+    dst.reset(tmp);
+}
+
+template<> void copyEventHack(QTabletEvent *src, QScopedPointer<QEvent> &dst) {
+    QTabletEvent *tmp = new QTabletEvent(src->type(),
+                                         src->posF(), src->globalPosF(),
+                                         src->device(), src->pointerType(),
+                                         src->pressure(),
+                                         src->xTilt(), src->yTilt(),
+                                         src->tangentialPressure(),
+                                         src->rotation(),
+                                         src->z(),
+                                         src->modifiers(),
+                                         src->uniqueId(),
+                                         src->button(), src->buttons());
+    tmp->setTimestamp(src->timestamp());
+    dst.reset(tmp);
+}
+
+
+
 template <class Event>
 bool KisInputManager::compressMoveEventCommon(Event *event)
 {
@@ -235,10 +251,10 @@ bool KisInputManager::compressMoveEventCommon(Event *event)
      */
     if ((event->type() == QEvent::MouseMove ||
          event->type() == QEvent::TabletMove) &&
-        (!d->matcher.supportsHiResInputEvents() ||
-         d->testingCompressBrushEvents)) {
+            (!d->matcher.supportsHiResInputEvents() ||
+             d->testingCompressBrushEvents)) {
 
-        d->compressedMoveEvent.reset(new Event(*event));
+        copyEventHack(event, d->compressedMoveEvent);
         d->moveEventCompressor.start();
 
         /**
@@ -260,12 +276,30 @@ bool KisInputManager::compressMoveEventCommon(Event *event)
     return retval;
 }
 
+bool shouldResetWheelDelta(QEvent * event)
+{
+    return
+        event->type() == QEvent::FocusIn ||
+        event->type() == QEvent::FocusOut ||
+        event->type() == QEvent::MouseButtonPress ||
+        event->type() == QEvent::MouseButtonRelease ||
+        event->type() == QEvent::MouseButtonDblClick ||
+        event->type() == QEvent::TabletPress ||
+        event->type() == QEvent::TabletRelease ||
+        event->type() == QEvent::Enter ||
+        event->type() == QEvent::Leave ||
+        event->type() == QEvent::TouchBegin ||
+        event->type() == QEvent::TouchEnd ||
+        event->type() == QEvent::TouchCancel ||
+        event->type() == QEvent::NativeGesture;
+
+}
+
 bool KisInputManager::eventFilterImpl(QEvent * event)
 {
-    // TODO: Handle touch events correctly.
     bool retval = false;
 
-    if (event->type() != QEvent::Wheel) {
+    if (shouldResetWheelDelta(event)) {
         d->accumulatedScrollDelta = 0;
     }
 
@@ -273,10 +307,7 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
     case QEvent::MouseButtonPress:
     case QEvent::MouseButtonDblClick: {
         d->debugEvent<QMouseEvent, true>(event);
-        //Block mouse press events on Genius tablets
-        break_if_tablet_active();
-        break_if_should_ignore_cursor_events();
-        break_if_touch_blocked_press_events();
+        if (d->touchHasBlockedPressEvents) break;
 
         QMouseEvent *mouseEvent = static_cast<QMouseEvent*>(event);
 
@@ -294,8 +325,7 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
     }
     case QEvent::MouseButtonRelease: {
         d->debugEvent<QMouseEvent, true>(event);
-        break_if_should_ignore_cursor_events();
-        break_if_touch_blocked_press_events();
+        if (d->touchHasBlockedPressEvents) break;
 
         QMouseEvent *mouseEvent = static_cast<QMouseEvent*>(event);
         retval = d->matcher.buttonReleased(mouseEvent->button(), mouseEvent);
@@ -337,7 +367,6 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
     }
     case QEvent::MouseMove: {
         d->debugEvent<QMouseEvent, true>(event);
-        break_if_should_ignore_cursor_events();
 
         QMouseEvent *mouseEvent = static_cast<QMouseEvent*>(event);
         retval = compressMoveEventCommon(mouseEvent);
@@ -347,6 +376,21 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
     case QEvent::Wheel: {
         d->debugEvent<QWheelEvent, false>(event);
         QWheelEvent *wheelEvent = static_cast<QWheelEvent*>(event);
+
+#ifdef Q_OS_MACOS
+        // Some QT wheel events are actually touch pad pan events. From the QT docs:
+        // "Wheel events are generated for both mouse wheels and trackpad scroll gestures."
+
+        // We differentiate between touchpad events and real mouse wheels by inspecting the
+        // event source.
+
+        if (wheelEvent->source() == Qt::MouseEventSource::MouseEventSynthesizedBySystem) {
+            KisAbstractInputAction::setInputManager(this);
+            retval = d->matcher.wheelEvent(KisSingleActionShortcut::WheelTrackpad, wheelEvent);
+            break;
+        }
+#endif
+
         d->accumulatedScrollDelta += wheelEvent->delta();
         KisSingleActionShortcut::WheelAction action;
 
@@ -354,7 +398,7 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
          * Ignore delta 0 events on OSX, since they are triggered by tablet
          * proximity when using Wacom devices.
          */
-#ifdef Q_OS_OSX
+#ifdef Q_OS_MACOS
         if(wheelEvent->delta() == 0) {
             retval = true;
             break;
@@ -391,12 +435,14 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
     }
     case QEvent::Enter:
         d->debugEvent<QEvent, false>(event);
-        d->containsPointer = true;
         //Make sure the input actions know we are active.
         KisAbstractInputAction::setInputManager(this);
+        if (!d->containsPointer) {
+            d->containsPointer = true;
 
-        stop_ignore_cursor_events();
-        touch_stop_block_press_events();
+            d->allowMouseEvents();
+            d->touchHasBlockedPressEvents = false;
+        }
 
         d->matcher.enterEvent();
         break;
@@ -408,8 +454,8 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
          * is hovering above some other widget, so restore cursor
          * events processing right now.
          */
-        stop_ignore_cursor_events();
-        touch_stop_block_press_events();
+        d->allowMouseEvents();
+        d->touchHasBlockedPressEvents = false;
 
         d->matcher.leaveEvent();
         break;
@@ -420,17 +466,17 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
         //Clear all state so we don't have half-matched shortcuts dangling around.
         d->matcher.reinitialize();
 
-        { // Emulate pressing of the key that are already pressed
-            KisExtendedModifiersMapper mapper;
+    { // Emulate pressing of the key that are already pressed
+        KisExtendedModifiersMapper mapper;
 
-            Qt::KeyboardModifiers modifiers = mapper.queryStandardModifiers();
-            Q_FOREACH (Qt::Key key, mapper.queryExtendedModifiers()) {
-                QKeyEvent kevent(QEvent::ShortcutOverride, key, modifiers);
-                eventFilterImpl(&kevent);
-            }
+        Qt::KeyboardModifiers modifiers = mapper.queryStandardModifiers();
+        Q_FOREACH (Qt::Key key, mapper.queryExtendedModifiers()) {
+            QKeyEvent kevent(QEvent::ShortcutOverride, key, modifiers);
+            eventFilterImpl(&kevent);
         }
+    }
 
-        stop_ignore_cursor_events();
+        d->allowMouseEvents();
         break;
 
     case QEvent::FocusOut: {
@@ -438,45 +484,12 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
         KisAbstractInputAction::setInputManager(this);
 
         QPointF currentLocalPos =
-            canvas()->canvasWidget()->mapFromGlobal(QCursor::pos());
+                canvas()->canvasWidget()->mapFromGlobal(QCursor::pos());
 
         d->matcher.lostFocusEvent(currentLocalPos);
 
         break;
     }
-    case QEvent::TabletRelease: {
-#ifdef Q_OS_OSX
-        stop_ignore_cursor_events();
-#endif
-        // break_if_touch_blocked_press_events();
-        d->debugEvent<QTabletEvent, false>(event);
-
-        QTabletEvent *tabletEvent = static_cast<QTabletEvent*>(event);
-        retval = d->matcher.buttonReleased(tabletEvent->button(), tabletEvent);
-        retval = true;
-        event->setAccepted(true);
-        break;
-    }
-
-
-    case QEvent::TabletMove: {
-        d->debugEvent<QTabletEvent, false>(event);
-
-        QTabletEvent *tabletEvent = static_cast<QTabletEvent*>(event);
-        retval = compressMoveEventCommon(tabletEvent);
-
-        /**
-         * The flow of tablet events means the tablet is in the
-         * proximity area, so activate it even when the
-         * TabletEnterProximity event was missed (may happen when
-         * changing focus of the window with tablet in the proximity
-         * area)
-         */
-        start_ignore_cursor_events();
-
-        break;
-    }
-
     case QEvent::TabletPress: {
         d->debugEvent<QTabletEvent, false>(event);
         QTabletEvent *tabletEvent = static_cast<QTabletEvent*>(event);
@@ -486,30 +499,87 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
             //Make sure the input actions know we are active.
             KisAbstractInputAction::setInputManager(this);
             retval = d->matcher.buttonPressed(tabletEvent->button(), tabletEvent);
+            if (!d->containsPointer) {
+                d->containsPointer = true;
+                d->touchHasBlockedPressEvents = false;
+            }
         }
         event->setAccepted(true);
         retval = true;
-        start_ignore_cursor_events();
+        d->blockMouseEvents();
         //Reset signal compressor to prevent processing events before press late
         d->resetCompressor();
         d->eatOneMousePress();
+
+#if defined Q_OS_LINUX && !defined QT_HAS_ENTER_LEAVE_PATCH
+        // remove this hack when this patch is integrated:
+        // https://codereview.qt-project.org/#/c/255384/
+        event->setAccepted(false);
+#endif
+
+        break;
+    }
+    case QEvent::TabletMove: {
+        d->debugEvent<QTabletEvent, false>(event);
+
+        QTabletEvent *tabletEvent = static_cast<QTabletEvent*>(event);
+        retval = compressMoveEventCommon(tabletEvent);
+
+        if (d->tabletLatencyTracker) {
+            d->tabletLatencyTracker->push(tabletEvent->timestamp());
+        }
+
+        /**
+         * The flow of tablet events means the tablet is in the
+         * proximity area, so activate it even when the
+         * TabletEnterProximity event was missed (may happen when
+         * changing focus of the window with tablet in the proximity
+         * area)
+         */
+        d->blockMouseEvents();
+
+#if defined Q_OS_LINUX && !defined QT_HAS_ENTER_LEAVE_PATCH
+        // remove this hack when this patch is integrated:
+        // https://codereview.qt-project.org/#/c/255384/
+        event->setAccepted(false);
+#endif
+
+        break;
+    }
+    case QEvent::TabletRelease: {
+#ifdef Q_OS_MAC
+        d->allowMouseEvents();
+#endif
+        d->debugEvent<QTabletEvent, false>(event);
+
+        QTabletEvent *tabletEvent = static_cast<QTabletEvent*>(event);
+        retval = d->matcher.buttonReleased(tabletEvent->button(), tabletEvent);
+        retval = true;
+        event->setAccepted(true);
+
+#if defined Q_OS_LINUX && !defined QT_HAS_ENTER_LEAVE_PATCH
+        // remove this hack when this patch is integrated:
+        // https://codereview.qt-project.org/#/c/255384/
+        event->setAccepted(false);
+#endif
+
         break;
     }
 
     case QEvent::TouchBegin:
-        touch_start_block_press_events();
-        touch_eat_one_mouse_press();
-        if (d->tryHidePopupPalette()) {
-            retval = true;
-        } else {
+    {
+        if (startTouch(retval)) {
+            QTouchEvent *tevent = static_cast<QTouchEvent*>(event);
             KisAbstractInputAction::setInputManager(this);
-            retval = d->matcher.touchBeginEvent(static_cast<QTouchEvent*>(event));
+            retval = d->matcher.touchBeginEvent(tevent);
             event->accept();
         }
         break;
-    case QEvent::TouchUpdate: {
+    }
+    case QEvent::TouchUpdate:
+    {
         QTouchEvent *tevent = static_cast<QTouchEvent*>(event);
-#ifdef Q_OS_OSX
+#ifdef Q_OS_MAC
         int count = 0;
         Q_FOREACH (const QTouchEvent::TouchPoint &point, tevent->touchPoints()) {
             if (point.state() != Qt::TouchPointReleased) {
@@ -518,31 +588,59 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
         }
 
         if (count < 2 && tevent->touchPoints().length() > count) {
-            touch_stop_block_press_events();
-            d->saveTouchEvent(tevent);
+            d->touchHasBlockedPressEvents = false;
             retval = d->matcher.touchEndEvent(tevent);
-            delete d->lastTouchEvent;
-            d->lastTouchEvent = 0;
         } else {
 #endif
-            touch_start_block_press_events();
+            d->touchHasBlockedPressEvents = KisConfig(true).disableTouchOnCanvas();
             KisAbstractInputAction::setInputManager(this);
             retval = d->matcher.touchUpdateEvent(tevent);
-#ifdef Q_OS_OSX
+#ifdef Q_OS_MACOS
         }
 #endif
         event->accept();
-
         break;
     }
     case QEvent::TouchEnd:
-        touch_stop_block_press_events();
-        d->saveTouchEvent(static_cast<QTouchEvent*>(event));
-        retval = d->matcher.touchEndEvent(static_cast<QTouchEvent*>(event));
+    {
+        endTouch();
+        QTouchEvent *tevent = static_cast<QTouchEvent*>(event);
+        retval = d->matcher.touchEndEvent(tevent);
         event->accept();
-        delete d->lastTouchEvent;
-        d->lastTouchEvent = 0;
         break;
+    }
+
+    case QEvent::NativeGesture:
+    {
+        QNativeGestureEvent *gevent = static_cast<QNativeGestureEvent*>(event);
+        switch (gevent->gestureType()) {
+            case Qt::BeginNativeGesture:
+            {
+                if (startTouch(retval)) {
+                    KisAbstractInputAction::setInputManager(this);
+                    retval = d->matcher.nativeGestureBeginEvent(gevent);
+                    event->accept();
+                }
+                break;
+            }
+            case Qt::EndNativeGesture:
+            {
+                endTouch();
+                retval = d->matcher.nativeGestureEndEvent(gevent);
+                event->accept();
+                break;
+            }
+            default:
+            {
+                KisAbstractInputAction::setInputManager(this);
+                retval = d->matcher.nativeGestureEvent(gevent);
+                event->accept();
+                break;
+            }
+        }
+        break;
+    }
+
     default:
         break;
     }
@@ -550,17 +648,36 @@ bool KisInputManager::eventFilterImpl(QEvent * event)
     return !retval ? d->processUnhandledEvent(event) : true;
 }
 
+bool KisInputManager::startTouch(bool &retval)
+{
+    d->touchHasBlockedPressEvents = KisConfig(true).disableTouchOnCanvas();
+    // Touch rejection: if touch is disabled on canvas, no need to block mouse press events
+    if (KisConfig(true).disableTouchOnCanvas()) {
+        d->eatOneMousePress();
+    }
+    if (d->tryHidePopupPalette()) {
+        retval = true;
+        return false;
+    } else {
+        return true;
+    }
+}
+
+void KisInputManager::endTouch()
+{
+    d->touchHasBlockedPressEvents = false;
+}
 
 void KisInputManager::slotCompressedMoveEvent()
 {
     if (d->compressedMoveEvent) {
-        // touch_stop_block_press_events();
+        // d->touchHasBlockedPressEvents = false;
 
         (void) d->handleCompressedTabletEvent(d->compressedMoveEvent.data());
         d->compressedMoveEvent.reset();
-        dbgKrita << "Compressed move event received.";
+        //dbgInput << "Compressed move event received.";
     } else {
-        dbgKrita << "Unexpected empty move event";
+        //dbgInput << "Unexpected empty move event";
     }
 }
 
@@ -569,46 +686,37 @@ KisCanvas2* KisInputManager::canvas() const
     return d->canvas;
 }
 
-KisToolProxy* KisInputManager::toolProxy() const
+QPointer<KisToolProxy> KisInputManager::toolProxy() const
 {
     return d->toolProxy;
 }
 
-QTouchEvent *KisInputManager::lastTouchEvent() const
-{
-    return d->lastTouchEvent;
-}
-
 void KisInputManager::slotAboutToChangeTool()
 {
-    QPointF currentLocalPos =
-        canvas()->canvasWidget()->mapFromGlobal(QCursor::pos());
-
+    QPointF currentLocalPos;
+    if (canvas() && canvas()->canvasWidget()) {
+        currentLocalPos = canvas()->canvasWidget()->mapFromGlobal(QCursor::pos());
+    }
     d->matcher.lostFocusEvent(currentLocalPos);
 }
 
 void KisInputManager::slotToolChanged()
 {
+    if (!d->canvas) return;
     KoToolManager *toolManager = KoToolManager::instance();
     KoToolBase *tool = toolManager->toolById(canvas(), toolManager->activeToolId());
-    if (tool && tool->isInTextMode()) {
-        d->forwardAllEventsToTool = true;
-        d->matcher.suppressAllActions(true);
-        d->maskSyntheticEvents(tool->maskSyntheticEvents());
-    } else {
-        d->forwardAllEventsToTool = false;
-        d->matcher.suppressAllActions(false);
+    if (tool) {
+        d->setMaskSyntheticEvents(tool->maskSyntheticEvents());
+        if (tool->isInTextMode()) {
+            d->forwardAllEventsToTool = true;
+            d->matcher.suppressAllActions(true);
+        } else {
+            d->forwardAllEventsToTool = false;
+            d->matcher.suppressAllActions(false);
+        }
     }
-
-
 }
 
-QPointF KisInputManager::widgetToDocument(const QPointF& position)
-{
-    const QPointF half = QPointF(.5f, .5f);
-    QPointF pixel = position + half;
-    return d->canvas->coordinatesConverter()->widgetToDocument(pixel);
-}
 
 void KisInputManager::profileChanged()
 {
@@ -631,7 +739,9 @@ void KisInputManager::profileChanged()
                 d->addWheelShortcut(shortcut->action(), shortcut->mode(), shortcut->keys(), shortcut->wheel());
                 break;
             case KisShortcutConfiguration::GestureType:
-                d->addTouchShortcut(shortcut->action(), shortcut->mode(), shortcut->gesture());
+                if (!d->addNativeGestureShortcut(shortcut->action(), shortcut->mode(), shortcut->gesture())) {
+                    d->addTouchShortcut(shortcut->action(), shortcut->mode(), shortcut->gesture());
+                }
                 break;
             default:
                 break;
@@ -639,6 +749,6 @@ void KisInputManager::profileChanged()
         }
     }
     else {
-        dbgKrita << "No Input Profile Found: canvas interaction will be impossible";
+        dbgInput << "No Input Profile Found: canvas interaction will be impossible";
     }
 }
