@@ -46,7 +46,6 @@
 #include <KoID.h>
 #include <KoColorSpaceRegistry.h>
 #include <KoColorProfile.h>
-#include <KoColorSpace.h>
 #include <KoColor.h>
 #include <KoUnit.h>
 
@@ -60,16 +59,17 @@
 #include <kis_transaction.h>
 #include <kis_paint_layer.h>
 #include <kis_group_layer.h>
-#include <metadata/kis_meta_data_io_backend.h>
-#include <metadata/kis_meta_data_store.h>
+#include <kis_meta_data_io_backend.h>
+#include <kis_meta_data_store.h>
 #include <KoColorModelStandardIds.h>
 #include "dialogs/kis_dlg_png_import.h"
 #include "kis_clipboard.h"
+#include <kis_cursor_override_hijacker.h>
+#include "kis_undo_stores.h"
 
 namespace
 {
 
-
 int getColorTypeforColorSpace(const KoColorSpace * cs , bool alpha)
 {
 
@@ -408,7 +408,6 @@ void _flush_fn(png_structp png_ptr)
     Q_UNUSED(png_ptr);
 }
 
-
 KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
 {
     dbgFile << "Start decoding PNG File";
@@ -517,21 +516,21 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
     double greenX, greenY;
     double blueX, blueY;
     png_get_cHRM(png_ptr,info_ptr, &whitePointX, &whitePointY, &redX, &redY, &greenX, &greenY, &blueX, &blueY);
-    qDebug() << "cHRM:" << whitePointX << whitePointY << redX << redY << greenX << greenY << blueX << blueY;
+    dbgFile << "cHRM:" << whitePointX << whitePointY << redX << redY << greenX << greenY << blueX << blueY;
 #endif
 
     // https://www.w3.org/TR/PNG/#11gAMA
 #if defined(PNG_GAMMA_SUPPORTED)
     double gamma;
     png_get_gAMA(png_ptr, info_ptr, &gamma);
-    qDebug() << "gAMA" << gamma;
+    dbgFile << "gAMA" << gamma;
 #endif
 
     // https://www.w3.org/TR/PNG/#11sRGB
 #if defined(PNG_sRGB_SUPPORTED)
     int sRGBIntent;
     png_get_sRGB(png_ptr, info_ptr, &sRGBIntent);
-    qDebug() << "sRGB" << sRGBIntent;
+    dbgFile << "sRGB" << sRGBIntent;
 #endif
 
     bool fromBlender = false;
@@ -550,6 +549,7 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
         }
     }
 
+    bool loadedImageIsHDR = false;
     const KoColorProfile* profile = 0;
     if (png_get_iCCP(png_ptr, info_ptr, &profile_name, &compression_type, &profile_data, &proflen)) {
         QByteArray profile_rawdata;
@@ -564,20 +564,22 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
                 dbgFile << "the profile is not suitable for output and therefore cannot be used in krita, we need to convert the image to a standard profile"; // TODO: in ko2 popup a selection menu to inform the user
             }
         }
+
+        loadedImageIsHDR = strcmp(profile_name, "ITUR_2100_PQ_FULL") == 0;
     }
     else {
         dbgFile << "no embedded profile, will use the default profile";
         if (color_nb_bits == 16 && !fromBlender && !qAppName().toLower().contains("test") && !m_batchMode) {
-            KisConfig cfg;
+            KisConfig cfg(true);
             quint32 behaviour = cfg.pasteBehaviour();
             if (behaviour == PASTE_ASK) {
                 KisDlgPngImport dlg(m_path, csName.first, csName.second);
-                QApplication::restoreOverrideCursor();
+                KisCursorOverrideHijacker hijacker;
+                Q_UNUSED(hijacker);
                 dlg.exec();
                 if (!dlg.profile().isEmpty()) {
                     profile = KoColorSpaceRegistry::instance()->profileByName(dlg.profile());
                 }
-                QApplication::setOverrideCursor(Qt::WaitCursor);
 
             }
         }
@@ -594,35 +596,50 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
     }
 
     // Retrieve a pointer to the colorspace
-    const KoColorSpace* cs;
-    if (profile && profile->isSuitableForOutput()) {
+    KoColorConversionTransformation* transform = 0;
+    const KoColorSpace* cs = 0;
+
+    if (loadedImageIsHDR &&
+        csName.first == RGBAColorModelID.id() &&
+        csName.second == Integer16BitsColorDepthID.id()) {
+
+        const KoColorSpace *p2020PQCS =
+            KoColorSpaceRegistry::instance()->colorSpace(
+                RGBAColorModelID.id(),
+                Integer16BitsColorDepthID.id(),
+                KoColorSpaceRegistry::instance()->p2020PQProfile());
+
+        cs = p2020PQCS;
+
+    } else if (profile && profile->isSuitableForOutput()) {
         dbgFile << "image has embedded profile: " << profile->name() << "\n";
         cs = KoColorSpaceRegistry::instance()->colorSpace(csName.first, csName.second, profile);
     }
     else {
         if (csName.first == RGBAColorModelID.id()) {
             cs = KoColorSpaceRegistry::instance()->colorSpace(csName.first, csName.second, "sRGB-elle-V2-srgbtrc.icc");
-        }
-        else {
+        } else if (csName.first == GrayAColorModelID.id()) {
+            cs = KoColorSpaceRegistry::instance()->colorSpace(csName.first, csName.second, "Gray-D50-elle-V2-srgbtrc.icc");
+        } else {
             cs = KoColorSpaceRegistry::instance()->colorSpace(csName.first, csName.second, 0);
         }
+
+        //TODO: two fixes : one tell the user about the problem and ask for a solution, and two once the kocolorspace include KoColorTransformation, use that instead of hacking a lcms transformation
+        // Create the cmsTransform if needed
+        if (profile) {
+            transform = KoColorSpaceRegistry::instance()->colorSpace(csName.first, csName.second, profile)->createColorConverter(cs, KoColorConversionTransformation::internalRenderingIntent(), KoColorConversionTransformation::internalConversionFlags());
+        }
     }
 
     if (cs == 0) {
         png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
         return KisImageBuilder_RESULT_UNSUPPORTED_COLORSPACE;
     }
-    //TODO: two fixes : one tell the user about the problem and ask for a solution, and two once the kocolorspace include KoColorTransformation, use that instead of hacking a lcms transformation
-    // Create the cmsTransform if needed
-    KoColorTransformation* transform = 0;
-    if (profile && !profile->isSuitableForOutput()) {
-        transform = KoColorSpaceRegistry::instance()->colorSpace(csName.first, csName.second, profile)->createColorConverter(cs, KoColorConversionTransformation::internalRenderingIntent(), KoColorConversionTransformation::internalConversionFlags());
-    }
 
     // Creating the KisImageSP
     if (m_image == 0) {
-        m_image = new KisImage(m_doc->createUndoStore(), width, height, cs, "built image");
-        Q_CHECK_PTR(m_image);
+        KisUndoStore *store = m_doc ? m_doc->createUndoStore() : new KisSurrogateUndoStore();
+        m_image = new KisImage(store, width, height, cs, "built image");
     }
 
     // Read resolution
@@ -644,19 +661,20 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
         dbgFile << "There are " << num_comments << " comments in the text";
         for (int i = 0; i < num_comments; i++) {
             QString key = QString(text_ptr[i].key).toLower();
-            dbgFile << "key is |" << text_ptr[i].key << "| containing " << text_ptr[i].text << " " << (key ==  "Raw profile type exif ");
+            dbgFile << "key: " << text_ptr[i].key
+                    << ", containing: " << text_ptr[i].text
+                    << ": " << (key == "raw profile type exif " ? "isExif" : "something else");
             if (key == "title") {
                 info->setAboutInfo("title", text_ptr[i].text);
             } else if (key == "description") {
                 info->setAboutInfo("comment", text_ptr[i].text);
             } else if (key == "author") {
-        qDebug()<<"Author:"<<text_ptr[i].text;
                 info->setAuthorInfo("creator", text_ptr[i].text);
-            } else if (key.contains("Raw profile type exif")) {
+            } else if (key.contains("raw profile type exif")) {
                 decode_meta_data(text_ptr + i, layer->metaData(), "exif", 6);
-            } else if (key.contains("Raw profile type iptc")) {
+            } else if (key.contains("raw profile type iptc")) {
                 decode_meta_data(text_ptr + i, layer->metaData(), "iptc", 14);
-            } else if (key.contains("Raw profile type xmp")) {
+            } else if (key.contains("raw profile type xmp")) {
                 decode_meta_data(text_ptr + i, layer->metaData(), "xmp", 0);
             } else if (key == "version") {
                 m_image->addAnnotation(new KisAnnotation("kpp_version", "version", QByteArray(text_ptr[i].text)));
@@ -665,16 +683,15 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
             }
         }
     }
-
     // Read image data
-    KisPNGReaderAbstract* reader = 0;
+    QScopedPointer<KisPNGReaderAbstract> reader;
     try {
         if (interlace_type == PNG_INTERLACE_ADAM7) {
-            reader = new KisPNGReaderFullImage(png_ptr, info_ptr, width, height);
+            reader.reset(new KisPNGReaderFullImage(png_ptr, info_ptr, width, height));
         } else {
-            reader = new KisPNGReaderLineByLine(png_ptr, info_ptr, width, height);
+            reader.reset(new KisPNGReaderLineByLine(png_ptr, info_ptr, width, height));
         }
-    } catch (std::bad_alloc& e) {
+    } catch (const std::bad_alloc& e) {
         // new png_byte[] may raise such an exception if the image
         // is invalid / to large.
         dbgFile << "bad alloc: " << e.what();
@@ -717,24 +734,24 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
                 do {
                     quint16 *d = reinterpret_cast<quint16 *>(it->rawData());
                     d[0] = *(src++);
-                    if (transform) transform->transform(reinterpret_cast<quint8*>(d), reinterpret_cast<quint8*>(d), 1);
                     if (hasalpha) {
                         d[1] = *(src++);
                     } else {
                         d[1] = quint16_MAX;
                     }
+                    if (transform) transform->transformInPlace(reinterpret_cast<quint8*>(d), reinterpret_cast<quint8*>(d), 1);
                 } while (it->nextPixel());
             } else  {
                 KisPNGReadStream stream(row_pointer, color_nb_bits);
                 do {
                     quint8 *d = it->rawData();
                     d[0] = (quint8)(stream.nextValue() * coeff);
-                    if (transform) transform->transform(d, d, 1);
                     if (hasalpha) {
                         d[1] = (quint8)(stream.nextValue() * coeff);
                     } else {
                         d[1] = UCHAR_MAX;
                     }
+                    if (transform) transform->transformInPlace(d, d, 1);
                 } while (it->nextPixel());
             }
             // FIXME:should be able to read 1 and 4 bits depth and scale them to 8 bits"
@@ -748,9 +765,9 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
                     d[2] = *(src++);
                     d[1] = *(src++);
                     d[0] = *(src++);
-                    if (transform) transform->transform(reinterpret_cast<quint8 *>(d), reinterpret_cast<quint8*>(d), 1);
                     if (hasalpha) d[3] = *(src++);
                     else d[3] = quint16_MAX;
+                    if (transform) transform->transformInPlace(reinterpret_cast<quint8 *>(d), reinterpret_cast<quint8*>(d), 1);
                 } while (it->nextPixel());
             } else {
                 KisPNGReadStream stream(row_pointer, color_nb_bits);
@@ -759,9 +776,9 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
                     d[2] = (quint8)(stream.nextValue() * coeff);
                     d[1] = (quint8)(stream.nextValue() * coeff);
                     d[0] = (quint8)(stream.nextValue() * coeff);
-                    if (transform) transform->transform(d, d, 1);
                     if (hasalpha) d[3] = (quint8)(stream.nextValue() * coeff);
                     else d[3] = UCHAR_MAX;
+                    if (transform) transform->transformInPlace(d, d, 1);
                 } while (it->nextPixel());
             }
             break;
@@ -794,8 +811,6 @@ KisImageBuilder_Result KisPNGConverter::buildImage(QIODevice* iod)
 
     // Freeing memory
     png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
-
-    delete reader;
     return KisImageBuilder_RESULT_OK;
 
 }
@@ -846,8 +861,7 @@ bool KisPNGConverter::saveDeviceToStore(const QString &filename, const QRect &im
 
         if (dev->colorSpace()->id() != "RGBA") {
             dev = new KisPaintDevice(*dev.data());
-            KUndo2Command *cmd = dev->convertTo(KoColorSpaceRegistry::instance()->rgb8());
-            delete cmd;
+            dev->convertTo(KoColorSpaceRegistry::instance()->rgb8());
         }
 
         bool success = pngconv.buildFile(&io, imageRect, xRes, yRes, dev, annotIt, annotIt, options, metaDataStore);
@@ -902,20 +916,44 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
 
     if (device->colorSpace()->colorDepthId() == Float16BitsColorDepthID
             || device->colorSpace()->colorDepthId() == Float32BitsColorDepthID
-            || device->colorSpace()->colorDepthId() == Float64BitsColorDepthID) {
-        const KoColorSpace *dstcs = KoColorSpaceRegistry::instance()->colorSpace(device->colorSpace()->colorModelId().id(), Integer8BitsColorDepthID.id(), "");
-        KisPaintDeviceSP tmp = new KisPaintDevice(dstcs);
-        KisPainter gc(tmp);
-        gc.bitBlt(imageRect.topLeft(), device, imageRect);
-        gc.end();
+            || device->colorSpace()->colorDepthId() == Float64BitsColorDepthID
+            || options.saveAsHDR) {
+
+        const KoColorSpace *dstCS =
+            KoColorSpaceRegistry::instance()->colorSpace(
+                device->colorSpace()->colorModelId().id(),
+                Integer16BitsColorDepthID.id(),
+                device->colorSpace()->profile());
+
+        if (options.saveAsHDR) {
+            dstCS =
+                KoColorSpaceRegistry::instance()->colorSpace(
+                        RGBAColorModelID.id(),
+                        Integer16BitsColorDepthID.id(),
+                        KoColorSpaceRegistry::instance()->p2020PQProfile());
+        }
+
+        KisPaintDeviceSP tmp = new KisPaintDevice(device->colorSpace());
+        tmp->makeCloneFromRough(device, imageRect);
+        tmp->convertTo(dstCS);
+
         device = tmp;
+
     }
 
-    if (options.forceSRGB) {
+    KIS_SAFE_ASSERT_RECOVER(!options.saveAsHDR || !options.forceSRGB) {
+        options.forceSRGB = false;
+    }
+
+    KIS_SAFE_ASSERT_RECOVER(!options.saveAsHDR || !options.tryToSaveAsIndexed) {
+        options.tryToSaveAsIndexed = false;
+    }
+
+    QStringList colormodels = QStringList() << RGBAColorModelID.id() << GrayAColorModelID.id();
+    if (options.forceSRGB || !colormodels.contains(device->colorSpace()->colorModelId().id())) {
         const KoColorSpace* cs = KoColorSpaceRegistry::instance()->colorSpace(RGBAColorModelID.id(), device->colorSpace()->colorDepthId().id(), "sRGB built-in - (lcms internal)");
         device = new KisPaintDevice(*device);
-        KUndo2Command *cmd = device->convertTo(cs);
-        delete cmd;
+        device->convertTo(cs);
     }
 
     // Initialize structures
@@ -968,15 +1006,15 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
     Q_ASSERT(color_type > -1);
 
     // Try to compute a table of color if the colorspace is RGB8f
-    png_colorp palette = 0;
+    QScopedArrayPointer<png_color> palette;
     int num_palette = 0;
     if (!options.alpha && options.tryToSaveAsIndexed && KoID(device->colorSpace()->id()) == KoID("RGBA")) { // png doesn't handle indexed images and alpha, and only have indexed for RGB8
-        palette = new png_color[255];
+        palette.reset(new png_color[255]);
 
         KisSequentialIterator it(device, imageRect);
 
         bool toomuchcolor = false;
-        do {
+        while (it.nextPixel()) {
             const quint8* c = it.oldRawData();
             bool findit = false;
             for (int i = 0; i < num_palette; i++) {
@@ -997,7 +1035,7 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
                 palette[num_palette].blue = c[0];
                 num_palette++;
             }
-        }  while (it.nextPixel());
+        }
 
         if (!toomuchcolor) {
             dbgFile << "Found a palette of " << num_palette << " colors";
@@ -1012,12 +1050,14 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
                 color_nb_bits = 8;
             }
         } else {
-            delete [] palette;
+            palette.reset();
         }
     }
 
     int interlacetype = options.interlace ? PNG_INTERLACE_ADAM7 : PNG_INTERLACE_NONE;
 
+    KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(color_type >= 0, KisImageBuilder_RESULT_FAILURE);
+
     png_set_IHDR(png_ptr, info_ptr,
                  imageRect.width(),
                  imageRect.height(),
@@ -1036,9 +1076,47 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
     /*if (!options.saveSRGBProfile && sRGB) {
         png_set_sRGB_gAMA_and_cHRM(png_ptr, info_ptr, PNG_sRGB_INTENT_PERCEPTUAL);
     }*/
+
+
+    /** TODO: Firefox still opens the image incorrectly if there is gAMA+cHRM tags
+     * present. According to the standard it should use iCCP tag with higher priority,
+     * but it doesn't:
+     *
+     * "When the iCCP chunk is present, PNG decoders that recognize it and are capable
+     *  of colour management [ICC] shall ignore the gAMA and cHRM chunks and use
+     *  the iCCP chunk instead and interpret it according to [ICC-1] and [ICC-1A]"
+     */
+
+#if 0
+    if (options.saveAsHDR) {
+        // https://www.w3.org/TR/PNG/#11gAMA
+#if defined(PNG_GAMMA_SUPPORTED)
+        // the values are set in accurdance of HDR-PNG standard:
+        // https://www.w3.org/TR/png-hdr-pq/
+
+        png_set_gAMA_fixed(png_ptr, info_ptr, 15000);
+        dbgFile << "gAMA" << "(Rec 2100)";
+#endif
+
+#if defined PNG_cHRM_SUPPORTED
+        png_set_cHRM_fixed(png_ptr, info_ptr,
+                           31270, 32900, // white point
+                           70800, 29200, // red
+                           17000, 79700, // green
+                           13100, 4600 // blue
+                           );
+        dbgFile << "cHRM" << "(Rec 2100)";
+#endif
+    }
+#endif
+
+
+    // we should ensure we don't access non-existing palette object
+    KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(palette || color_type != PNG_COLOR_TYPE_PALETTE, KisImageBuilder_RESULT_FAILURE);
+
     // set the palette
     if (color_type == PNG_COLOR_TYPE_PALETTE) {
-        png_set_PLTE(png_ptr, info_ptr, palette, num_palette);
+        png_set_PLTE(png_ptr, info_ptr, palette.data(), num_palette);
     }
     // Save annotation
     vKisAnnotationSP_it it = annotationsStart;
@@ -1075,31 +1153,57 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
     const KoColorProfile* colorProfile = device->colorSpace()->profile();
     QByteArray colorProfileData = colorProfile->rawData();
     if (!sRGB || options.saveSRGBProfile) {
+
 #if PNG_LIBPNG_VER_MAJOR >= 1 && PNG_LIBPNG_VER_MINOR >= 5
-        png_set_iCCP(png_ptr, info_ptr, (char*)"icc", PNG_COMPRESSION_TYPE_BASE, (const png_bytep)colorProfileData.constData(), colorProfileData . size());
+        const char *typeString = !options.saveAsHDR ? "icc" : "ITUR_2100_PQ_FULL";
+        png_set_iCCP(png_ptr, info_ptr, (png_const_charp)typeString, PNG_COMPRESSION_TYPE_BASE, (png_const_bytep)colorProfileData.constData(), colorProfileData . size());
 #else
-        png_set_iCCP(png_ptr, info_ptr, (char*)"icc", PNG_COMPRESSION_TYPE_BASE, (char*)colorProfileData.constData(), colorProfileData . size());
+        // older version of libpng has a problem with constness on the parameters
+        char typeStringICC[] = "icc";
+        char typeStringHDR[] = "ITUR_2100_PQ_FULL";
+        char *typeString = !options.saveAsHDR ? typeStringICC : typeStringHDR;
+        png_set_iCCP(png_ptr, info_ptr, typeString, PNG_COMPRESSION_TYPE_BASE, colorProfileData.data(), colorProfileData . size());
 #endif
     }
 
-    // read comments from the document information
-    // warning: according to the official png spec, the keys need to be capitalised!
+    // save comments from the document information
+    // warning: according to the official png spec, the keys need to be capitalized!
     if (m_doc) {
-        png_text texts[3];
+        png_text texts[4];
         int nbtexts = 0;
         KoDocumentInfo * info = m_doc->documentInfo();
         QString title = info->aboutInfo("title");
-        if (!title.isEmpty()) {
+        if (!title.isEmpty() && options.storeMetaData) {
             fillText(texts + nbtexts, "Title", title);
             nbtexts++;
         }
-        QString abstract = info->aboutInfo("comment");
-        if (!abstract.isEmpty()) {
+        QString abstract = info->aboutInfo("subject");
+        if (abstract.isEmpty()) {
+            abstract = info->aboutInfo("abstract");
+        }
+        if (!abstract.isEmpty() && options.storeMetaData) {
+            QString keywords = info->aboutInfo("keyword");
+            if (!keywords.isEmpty()) {
+                abstract = abstract + " keywords: " + keywords;
+            }
             fillText(texts + nbtexts, "Description", abstract);
             nbtexts++;
         }
+
+        QString license = info->aboutInfo("license");
+        if (!license.isEmpty() && options.storeMetaData) {
+            fillText(texts + nbtexts, "Copyright", license);
+            nbtexts++;
+        }
+
         QString author = info->authorInfo("creator");
-        if (!author.isEmpty()) {
+        if (!author.isEmpty() && options.storeAuthor) {
+            if (!info->authorContactInfo().isEmpty()) {
+                QString contact = info->authorContactInfo().at(0);
+                if (!contact.isEmpty()) {
+                    author = author+"("+contact+")";
+                }
+            }
             fillText(texts + nbtexts, "Author", author);
             nbtexts++;
         }
@@ -1166,26 +1270,48 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
     // Write the PNG
     //     png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, 0);
 
+    struct RowPointersStruct {
+        RowPointersStruct(const QSize &size, int pixelSize)
+            : numRows(size.height())
+        {
+            rows = new png_byte*[numRows];
+
+            for (int i = 0; i < numRows; i++) {
+                rows[i] = new png_byte[size.width() * pixelSize];
+            }
+        }
+
+        ~RowPointersStruct() {
+            for (int i = 0; i < numRows; i++) {
+                delete[] rows[i];
+            }
+            delete[] rows;
+        }
+
+        const int numRows = 0;
+        png_byte** rows = 0;
+    };
+
+
     // Fill the data structure
-    png_byte** row_pointers = new png_byte*[imageRect.height()];
+    RowPointersStruct rowPointers(imageRect.size(), device->pixelSize());
+
     int row = 0;
     for (int y = imageRect.y(); y < imageRect.y() + imageRect.height(); y++, row++) {
         KisHLineConstIteratorSP it = device->createHLineConstIteratorNG(imageRect.x(), y, imageRect.width());
 
-        row_pointers[row] = new png_byte[imageRect.width() * device->pixelSize()];
-
         switch (color_type) {
         case PNG_COLOR_TYPE_GRAY:
         case PNG_COLOR_TYPE_GRAY_ALPHA:
             if (color_nb_bits == 16) {
-                quint16 *dst = reinterpret_cast<quint16 *>(row_pointers[row]);
+                quint16 *dst = reinterpret_cast<quint16 *>(rowPointers.rows[row]);
                 do {
                     const quint16 *d = reinterpret_cast<const quint16 *>(it->oldRawData());
                     *(dst++) = d[0];
                     if (options.alpha) *(dst++) = d[1];
                 } while (it->nextPixel());
             } else {
-                quint8 *dst = row_pointers[row];
+                quint8 *dst = rowPointers.rows[row];
                 do {
                     const quint8 *d = it->oldRawData();
                     *(dst++) = d[0];
@@ -1196,7 +1322,7 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
         case PNG_COLOR_TYPE_RGB:
         case PNG_COLOR_TYPE_RGB_ALPHA:
             if (color_nb_bits == 16) {
-                quint16 *dst = reinterpret_cast<quint16 *>(row_pointers[row]);
+                quint16 *dst = reinterpret_cast<quint16 *>(rowPointers.rows[row]);
                 do {
                     const quint16 *d = reinterpret_cast<const quint16 *>(it->oldRawData());
                     *(dst++) = d[2];
@@ -1205,7 +1331,7 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
                     if (options.alpha) *(dst++) = d[3];
                 } while (it->nextPixel());
             } else {
-                quint8 *dst = row_pointers[row];
+                quint8 *dst = rowPointers.rows[row];
                 do {
                     const quint8 *d = it->oldRawData();
                     *(dst++) = d[2];
@@ -1216,7 +1342,7 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
             }
             break;
         case PNG_COLOR_TYPE_PALETTE: {
-            quint8 *dst = row_pointers[row];
+            quint8 *dst = rowPointers.rows[row];
             KisPNGWriteStream writestream(dst, color_nb_bits);
             do {
                 const quint8 *d = it->oldRawData();
@@ -1233,26 +1359,17 @@ KisImageBuilder_Result KisPNGConverter::buildFile(QIODevice* iodevice, const QRe
         }
             break;
         default:
-            delete[] row_pointers;
             return KisImageBuilder_RESULT_UNSUPPORTED;
         }
     }
 
-    png_write_image(png_ptr, row_pointers);
+    png_write_image(png_ptr, rowPointers.rows);
 
     // Writing is over
     png_write_end(png_ptr, info_ptr);
 
     // Free memory
     png_destroy_write_struct(&png_ptr, &info_ptr);
-    for (int y = 0; y < imageRect.height(); y++) {
-        delete[] row_pointers[y];
-    }
-    delete[] row_pointers;
-
-    if (color_type == PNG_COLOR_TYPE_PALETTE) {
-        delete [] palette;
-    }
     return KisImageBuilder_RESULT_OK;
 }
 
