@@ -24,12 +24,11 @@
 #include <klocalizedstring.h>
 
 #include <KoPointerEvent.h>
-#include <KoCanvasResourceManager.h>
+#include <KoCanvasResourceProvider.h>
 
 #include "kis_algebra_2d.h"
 #include "kis_distance_information.h"
 #include "kis_painting_information_builder.h"
-#include "kis_recording_adapter.h"
 #include "kis_image.h"
 #include "kis_painter.h"
 #include <brushengine/kis_paintop_preset.h>
@@ -40,6 +39,11 @@
 #include "KisStabilizerDelayedPaintHelper.h"
 #include "kis_config.h"
 
+#include "kis_random_source.h"
+#include "KisPerStrokeRandomSource.h"
+
+#include "strokes/freehand_stroke.h"
+#include "strokes/KisFreehandStrokeInfo.h"
 
 #include <math.h>
 
@@ -50,10 +54,17 @@
 // airbrush rate, which can improve responsiveness.
 const qreal AIRBRUSH_INTERVAL_FACTOR = 0.5;
 
+// The amount of time, in milliseconds, to allow between updates of the spacing information. Only
+// used when spacing updates between dabs are enabled.
+const qreal SPACING_UPDATE_INTERVAL = 50.0;
+
+// The amount of time, in milliseconds, to allow between updates of the timing information. Only
+// used when airbrushing.
+const qreal TIMING_UPDATE_INTERVAL = 50.0;
+
 struct KisToolFreehandHelper::Private
 {
     KisPaintingInformationBuilder *infoBuilder;
-    KisRecordingAdapter *recordingAdapter;
     KisStrokesFacade *strokesFacade;
 
     KUndo2MagicString transactionText;
@@ -66,7 +77,7 @@ struct KisToolFreehandHelper::Private
     QTime strokeTime;
     QTimer strokeTimeoutTimer;
 
-    QVector<PainterInfo*> painterInfos;
+    QVector<KisFreehandStrokeInfo*> strokeInfos;
     KisResourcesSnapshotSP resources;
     KisStrokeId strokeId;
 
@@ -75,6 +86,10 @@ struct KisToolFreehandHelper::Private
 
     KisSmoothingOptionsSP smoothingOptions;
 
+    // fake random sources for hovering outline *only*
+    KisRandomSourceSP fakeDabRandomSource;
+    KisPerStrokeRandomSourceSP fakeStrokeRandomSource;
+
     // Timer used to generate paint updates periodically even without input events. This is only
     // used for paintops that depend on timely updates even when the cursor is not moving, e.g. for
     // airbrushing effects.
@@ -88,11 +103,14 @@ struct KisToolFreehandHelper::Private
     KisPaintOpUtils::PositionHistory lastCursorPos;
 
     // Stabilizer data
+    bool usingStabilizer;
     QQueue<KisPaintInformation> stabilizerDeque;
     QTimer stabilizerPollTimer;
     KisStabilizedEventsSampler stabilizedSampler;
     KisStabilizerDelayedPaintHelper stabilizerDelayedPaintHelper;
 
+    QTimer asynchronousUpdatesThresholdTimer;
+
     int canvasRotation;
     bool canvasMirroredH;
 
@@ -102,21 +120,24 @@ struct KisToolFreehandHelper::Private
 
 KisToolFreehandHelper::KisToolFreehandHelper(KisPaintingInformationBuilder *infoBuilder,
                                              const KUndo2MagicString &transactionText,
-                                             KisRecordingAdapter *recordingAdapter,
                                              KisSmoothingOptions *smoothingOptions)
     : m_d(new Private())
 {
     m_d->infoBuilder = infoBuilder;
-    m_d->recordingAdapter = recordingAdapter;
     m_d->transactionText = transactionText;
     m_d->smoothingOptions = KisSmoothingOptionsSP(
                 smoothingOptions ? smoothingOptions : new KisSmoothingOptions());
     m_d->canvasRotation = 0;
 
+    m_d->fakeDabRandomSource = new KisRandomSource();
+    m_d->fakeStrokeRandomSource = new KisPerStrokeRandomSource();
+
     m_d->strokeTimeoutTimer.setSingleShot(true);
     connect(&m_d->strokeTimeoutTimer, SIGNAL(timeout()), SLOT(finishStroke()));
     connect(&m_d->airbrushingTimer, SIGNAL(timeout()), SLOT(doAirbrushing()));
+    connect(&m_d->asynchronousUpdatesThresholdTimer, SIGNAL(timeout()), SLOT(doAsynchronousUpdate()));
     connect(&m_d->stabilizerPollTimer, SIGNAL(timeout()), SLOT(stabilizerPollAndPaint()));
+    connect(m_d->smoothingOptions.data(), SIGNAL(sigSmoothingTypeChanged()), SLOT(slotSmoothingTypeChanged()));
 
     m_d->stabilizerDelayedPaintHelper.setPaintLineCallback(
                 [this](const KisPaintInformation &pi1, const KisPaintInformation &pi2) {
@@ -154,9 +175,9 @@ QPainterPath KisToolFreehandHelper::paintOpOutline(const QPointF &savedCursorPos
     qreal startAngle = KisAlgebra2D::directionBetweenPoints(prevPoint, savedCursorPos, 0);
     info.setCanvasRotation(m_d->canvasRotation);
     info.setCanvasHorizontalMirrorState( m_d->canvasMirroredH );
-    KisDistanceInformation distanceInfo(prevPoint, 0, startAngle);
+    KisDistanceInformation distanceInfo(prevPoint, startAngle);
 
-    if (!m_d->painterInfos.isEmpty()) {
+    if (!m_d->strokeInfos.isEmpty()) {
         settings = m_d->resources->currentPaintOpPreset()->settings();
         if (m_d->stabilizerDelayedPaintHelper.running() &&
                 m_d->stabilizerDelayedPaintHelper.hasLastPaintInformation()) {
@@ -175,7 +196,7 @@ QPainterPath KisToolFreehandHelper::paintOpOutline(const QPointF &savedCursorPos
          * object.
          */
         KisDistanceInformation *buddyDistance =
-            m_d->painterInfos.first()->buddyDragDistance();
+            m_d->strokeInfos.first()->buddyDragDistance();
 
         if (buddyDistance) {
             /**
@@ -185,20 +206,21 @@ QPainterPath KisToolFreehandHelper::paintOpOutline(const QPointF &savedCursorPos
              * local copy of the coordinates.
              */
             distanceInfo = *buddyDistance;
-            distanceInfo.overrideLastValues(prevPoint, 0, startAngle);
+            distanceInfo.overrideLastValues(prevPoint, startAngle);
 
-        } else if (m_d->painterInfos.first()->dragDistance->isStarted()) {
-            distanceInfo = *m_d->painterInfos.first()->dragDistance;
+        } else if (m_d->strokeInfos.first()->dragDistance->isStarted()) {
+            distanceInfo = *m_d->strokeInfos.first()->dragDistance;
         }
     }
 
     KisPaintInformation::DistanceInformationRegistrar registrar =
         info.registerDistanceInformation(&distanceInfo);
 
+    info.setRandomSource(m_d->fakeDabRandomSource);
+    info.setPerStrokeRandomSource(m_d->fakeStrokeRandomSource);
+
     QPainterPath outline = settings->brushOutline(info, mode);
 
-
-
     if (m_d->resources &&
         m_d->smoothingOptions->smoothingType() == KisSmoothingOptions::STABILIZER &&
         m_d->smoothingOptions->useDelayDistance()) {
@@ -219,7 +241,7 @@ void KisToolFreehandHelper::cursorMoved(const QPointF &cursorPos)
 
 void KisToolFreehandHelper::initPaint(KoPointerEvent *event,
                                       const QPointF &pixelCoords,
-                                      KoCanvasResourceManager *resourceManager,
+                                      KoCanvasResourceProvider *resourceManager,
                                       KisImageWSP image, KisNodeSP currentNode,
                                       KisStrokesFacade *strokesFacade,
                                       KisNodeSP overrideNode,
@@ -248,15 +270,13 @@ bool KisToolFreehandHelper::isRunning() const
 
 void KisToolFreehandHelper::initPaintImpl(qreal startAngle,
                                           const KisPaintInformation &pi,
-                                          KoCanvasResourceManager *resourceManager,
+                                          KoCanvasResourceProvider *resourceManager,
                                           KisImageWSP image,
                                           KisNodeSP currentNode,
                                           KisStrokesFacade *strokesFacade,
                                           KisNodeSP overrideNode,
                                           KisDefaultBoundsBaseSP bounds)
 {
-    Q_UNUSED(overrideNode);
-
     m_d->strokesFacade = strokesFacade;
 
     m_d->haveTangent = false;
@@ -266,37 +286,41 @@ void KisToolFreehandHelper::initPaintImpl(qreal startAngle,
 
     m_d->previousPaintInformation = pi;
 
-    createPainters(m_d->painterInfos,
-                   m_d->previousPaintInformation.pos(),
-                   m_d->previousPaintInformation.currentTime(),
-                   startAngle);
-
     m_d->resources = new KisResourcesSnapshot(image,
                                               currentNode,
                                               resourceManager,
                                               bounds);
-
     if(overrideNode) {
         m_d->resources->setCurrentNode(overrideNode);
     }
 
-    if(m_d->recordingAdapter) {
-        m_d->recordingAdapter->startStroke(image, m_d->resources);
-    }
+    const bool airbrushing = m_d->resources->needsAirbrushing();
+    const bool useSpacingUpdates = m_d->resources->needsSpacingUpdates();
+
+    KisDistanceInitInfo startDistInfo(m_d->previousPaintInformation.pos(),
+                                      startAngle,
+                                      useSpacingUpdates ? SPACING_UPDATE_INTERVAL : LONG_TIME,
+                                      airbrushing ? TIMING_UPDATE_INTERVAL : LONG_TIME,
+                                      0);
+    KisDistanceInformation startDist = startDistInfo.makeDistInfo();
+
+    createPainters(m_d->strokeInfos,
+                   startDist);
 
     KisStrokeStrategy *stroke =
-        new FreehandStrokeStrategy(m_d->resources->needsIndirectPainting(),
-                                   m_d->resources->indirectPaintingCompositeOp(),
-                                   m_d->resources, m_d->painterInfos, m_d->transactionText);
+        new FreehandStrokeStrategy(m_d->resources, m_d->strokeInfos, m_d->transactionText);
 
     m_d->strokeId = m_d->strokesFacade->startStroke(stroke);
 
     m_d->history.clear();
     m_d->distanceHistory.clear();
 
-    if(m_d->resources->needsAirbrushing()) {
+    if (airbrushing) {
         m_d->airbrushingTimer.setInterval(computeAirbrushTimerInterval());
         m_d->airbrushingTimer.start();
+    } else if (m_d->resources->presetNeedsAsynchronousUpdates()) {
+        m_d->asynchronousUpdatesThresholdTimer.setInterval(80 /* msec */);
+        m_d->asynchronousUpdatesThresholdTimer.start();
     }
 
     if (m_d->smoothingOptions->smoothingType() == KisSmoothingOptions::STABILIZER) {
@@ -304,9 +328,9 @@ void KisToolFreehandHelper::initPaintImpl(qreal startAngle,
     }
 
     // If airbrushing, paint an initial dab immediately. This is a workaround for an issue where
-    // some paintops (Dyna, Particle, Sketch) might never initialize their spacing information until
-    // paintAt is called.
-    if (m_d->resources->needsAirbrushing()) {
+    // some paintops (Dyna, Particle, Sketch) might never initialize their spacing/timing
+    // information until paintAt is called.
+    if (airbrushing) {
         paintAt(pi);
     }
 }
@@ -350,8 +374,8 @@ void KisToolFreehandHelper::paintBezierSegment(KisPaintInformation pi1, KisPaint
             intersection.manhattanLength() > maxSanePoint) {
 
             intersection = 0.5 * (pi1.pos() + pi2.pos());
-//            dbgKrita << "WARINING: there is no intersection point "
-//                     << "in the basic smoothing algoriths";
+//            dbgKrita << "WARNING: there is no intersection point "
+//                     << "in the basic smoothing algorithms";
         }
 
         controlTarget1 = intersection;
@@ -439,7 +463,7 @@ void KisToolFreehandHelper::paint(KisPaintInformation &info)
      * 2) There is no 'Quality' parameter, since the number of samples
      *    is calculated automatically
      *
-     * 3) 'Tail Aggressiveness' is used for controling the end of the
+     * 3) 'Tail Aggressiveness' is used for controlling the end of the
      *    stroke
      *
      * 4) The formila is a little bit different: 'Distance' parameter
@@ -561,7 +585,11 @@ void KisToolFreehandHelper::paint(KisPaintInformation &info)
             m_d->previousTangent = newTangent;
         }
         m_d->olderPaintInformation = m_d->previousPaintInformation;
-        m_d->strokeTimeoutTimer.start(100);
+
+        // Enable stroke timeout only when not airbrushing.
+        if (!m_d->airbrushingTimer.isActive()) {
+            m_d->strokeTimeoutTimer.start(100);
+        }
     }
     else if (m_d->smoothingOptions->smoothingType() == KisSmoothingOptions::NO_SMOOTHING){
         paintLine(m_d->previousPaintInformation, info);
@@ -569,6 +597,11 @@ void KisToolFreehandHelper::paint(KisPaintInformation &info)
 
     if (m_d->smoothingOptions->smoothingType() == KisSmoothingOptions::STABILIZER) {
         m_d->stabilizedSampler.addEvent(info);
+        if (m_d->stabilizerDelayedPaintHelper.running()) {
+            // Paint here so we don't have to rely on the timer
+            // This is just a tricky source for a relatively stable 7ms "timer"
+            m_d->stabilizerDelayedPaintHelper.paintSome();
+        }
     } else {
         m_d->previousPaintInformation = info;
     }
@@ -591,6 +624,10 @@ void KisToolFreehandHelper::endPaint()
         m_d->airbrushingTimer.stop();
     }
 
+    if (m_d->asynchronousUpdatesThresholdTimer.isActive()) {
+        m_d->asynchronousUpdatesThresholdTimer.stop();
+    }
+
     if (m_d->smoothingOptions->smoothingType() == KisSmoothingOptions::STABILIZER) {
         stabilizerEnd();
     }
@@ -601,14 +638,13 @@ void KisToolFreehandHelper::endPaint()
      * Please note that we are not in MT here, so no mutex
      * is needed
      */
-    m_d->painterInfos.clear();
+    m_d->strokeInfos.clear();
+
+    // last update to complete rendering if there is still something pending
+    doAsynchronousUpdate(true);
 
     m_d->strokesFacade->endStroke(m_d->strokeId);
     m_d->strokeId.clear();
-
-    if(m_d->recordingAdapter) {
-        m_d->recordingAdapter->endStroke();
-    }
 }
 
 void KisToolFreehandHelper::cancelPaint()
@@ -621,6 +657,10 @@ void KisToolFreehandHelper::cancelPaint()
         m_d->airbrushingTimer.stop();
     }
 
+    if (m_d->asynchronousUpdatesThresholdTimer.isActive()) {
+        m_d->asynchronousUpdatesThresholdTimer.stop();
+    }
+
     if (m_d->stabilizerPollTimer.isActive()) {
         m_d->stabilizerPollTimer.stop();
     }
@@ -630,15 +670,11 @@ void KisToolFreehandHelper::cancelPaint()
     }
 
     // see a comment in endPaint()
-    m_d->painterInfos.clear();
+    m_d->strokeInfos.clear();
 
     m_d->strokesFacade->cancelStroke(m_d->strokeId);
     m_d->strokeId.clear();
 
-    if(m_d->recordingAdapter) {
-        //FIXME: not implemented
-        //m_d->recordingAdapter->cancelStroke();
-    }
 }
 
 int KisToolFreehandHelper::elapsedStrokeTime() const
@@ -648,6 +684,7 @@ int KisToolFreehandHelper::elapsedStrokeTime() const
 
 void KisToolFreehandHelper::stabilizerStart(KisPaintInformation firstPaintInfo)
 {
+    m_d->usingStabilizer = true;
     // FIXME: Ugly hack, this is no a "distance" in any way
     int sampleSize = qRound(m_d->effectiveSmoothnessDistance());
     sampleSize = qMax(3, sampleSize);
@@ -659,14 +696,14 @@ void KisToolFreehandHelper::stabilizerStart(KisPaintInformation firstPaintInfo)
     }
 
     // Poll and draw regularly
-    KisConfig cfg;
+    KisConfig cfg(true);
     int stabilizerSampleSize = cfg.stabilizerSampleSize();
     m_d->stabilizerPollTimer.setInterval(stabilizerSampleSize);
     m_d->stabilizerPollTimer.start();
 
-    int delayedPaintInterval = cfg.stabilizerDelayedPaintInterval();
-    if (delayedPaintInterval < stabilizerSampleSize) {
-        m_d->stabilizerDelayedPaintHelper.start(delayedPaintInterval, firstPaintInfo);
+    bool delayedPaintEnabled = cfg.stabilizerDelayedPaint();
+    if (delayedPaintEnabled) {
+        m_d->stabilizerDelayedPaintHelper.start(firstPaintInfo);
     }
 
     m_d->stabilizedSampler.clear();
@@ -700,14 +737,14 @@ KisToolFreehandHelper::getStabilizedPaintInfo(const QQueue<KisPaintInformation>
         if (m_d->smoothingOptions->stabilizeSensors()) {
             while (it != end) {
                 qreal k = qreal(i - 1) / i; // coeff for uniform averaging
-                result = KisPaintInformation::mixWithoutTime(k, *it, result);
+                result.KisPaintInformation::mixOtherWithoutTime(k, *it);
                 it++;
                 i++;
             }
         } else{
             while (it != end) {
                 qreal k = qreal(i - 1) / i; // coeff for uniform averaging
-                result = KisPaintInformation::mixOnlyPosition(k, *it, result);
+                result.KisPaintInformation::mixOtherOnlyPosition(k, *it);
                 it++;
                 i++;
             }
@@ -798,11 +835,23 @@ void KisToolFreehandHelper::stabilizerEnd()
     if (m_d->stabilizerDelayedPaintHelper.running()) {
         m_d->stabilizerDelayedPaintHelper.end();
     }
+    m_d->usingStabilizer = false;
 }
 
-const KisPaintOp* KisToolFreehandHelper::currentPaintOp() const
+void KisToolFreehandHelper::slotSmoothingTypeChanged()
 {
-    return !m_d->painterInfos.isEmpty() ? m_d->painterInfos.first()->painter->paintOp() : 0;
+    if (!isRunning()) {
+        return;
+    }
+    KisSmoothingOptions::SmoothingType currentSmoothingType =
+            m_d->smoothingOptions->smoothingType();
+    if (m_d->usingStabilizer
+            && (currentSmoothingType != KisSmoothingOptions::STABILIZER)) {
+        stabilizerEnd();
+    } else if (!m_d->usingStabilizer
+            && (currentSmoothingType == KisSmoothingOptions::STABILIZER)) {
+        stabilizerStart(m_d->previousPaintInformation);
+    }
 }
 
 void KisToolFreehandHelper::finishStroke()
@@ -823,7 +872,7 @@ void KisToolFreehandHelper::finishStroke()
 void KisToolFreehandHelper::doAirbrushing()
 {
     // Check that the stroke hasn't ended.
-    if (!m_d->painterInfos.isEmpty()) {
+    if (!m_d->strokeInfos.isEmpty()) {
 
         // Add a new painting update at a point identical to the previous one, except for the time
         // and speed information.
@@ -841,40 +890,38 @@ void KisToolFreehandHelper::doAirbrushing()
     }
 }
 
+void KisToolFreehandHelper::doAsynchronousUpdate(bool forceUpdate)
+{
+    m_d->strokesFacade->addJob(m_d->strokeId,
+                               new FreehandStrokeStrategy::UpdateData(forceUpdate));
+}
+
 int KisToolFreehandHelper::computeAirbrushTimerInterval() const
 {
     qreal realInterval = m_d->resources->airbrushingInterval() * AIRBRUSH_INTERVAL_FACTOR;
     return qMax(1, qFloor(realInterval));
 }
 
-void KisToolFreehandHelper::paintAt(int painterInfoId,
+void KisToolFreehandHelper::paintAt(int strokeInfoId,
                                     const KisPaintInformation &pi)
 {
     m_d->hasPaintAtLeastOnce = true;
     m_d->strokesFacade->addJob(m_d->strokeId,
-                               new FreehandStrokeStrategy::Data(m_d->resources->currentNode(),
-                                                                painterInfoId, pi));
+                               new FreehandStrokeStrategy::Data(strokeInfoId, pi));
 
-    if(m_d->recordingAdapter) {
-        m_d->recordingAdapter->addPoint(pi);
-    }
 }
 
-void KisToolFreehandHelper::paintLine(int painterInfoId,
+void KisToolFreehandHelper::paintLine(int strokeInfoId,
                                       const KisPaintInformation &pi1,
                                       const KisPaintInformation &pi2)
 {
     m_d->hasPaintAtLeastOnce = true;
     m_d->strokesFacade->addJob(m_d->strokeId,
-                               new FreehandStrokeStrategy::Data(m_d->resources->currentNode(),
-                                                                painterInfoId, pi1, pi2));
+                               new FreehandStrokeStrategy::Data(strokeInfoId, pi1, pi2));
 
-    if(m_d->recordingAdapter) {
-        m_d->recordingAdapter->addLine(pi1, pi2);
-    }
 }
 
-void KisToolFreehandHelper::paintBezierCurve(int painterInfoId,
+void KisToolFreehandHelper::paintBezierCurve(int strokeInfoId,
                                              const KisPaintInformation &pi1,
                                              const QPointF &control1,
                                              const QPointF &control2,
@@ -906,21 +953,15 @@ void KisToolFreehandHelper::paintBezierCurve(int painterInfoId,
 
     m_d->hasPaintAtLeastOnce = true;
     m_d->strokesFacade->addJob(m_d->strokeId,
-                               new FreehandStrokeStrategy::Data(m_d->resources->currentNode(),
-                                                                painterInfoId,
+                               new FreehandStrokeStrategy::Data(strokeInfoId,
                                                                 pi1, control1, control2, pi2));
 
-    if(m_d->recordingAdapter) {
-        m_d->recordingAdapter->addCurve(pi1, control1, control2, pi2);
-    }
 }
 
-void KisToolFreehandHelper::createPainters(QVector<PainterInfo*> &painterInfos,
-                                           const QPointF &lastPosition,
-                                           int lastTime,
-                                           qreal lastAngle)
+void KisToolFreehandHelper::createPainters(QVector<KisFreehandStrokeInfo*> &strokeInfos,
+                                           const KisDistanceInformation &startDist)
 {
-    painterInfos << new PainterInfo(lastPosition, lastTime, lastAngle);
+    strokeInfos << new KisFreehandStrokeInfo(startDist);
 }
 
 void KisToolFreehandHelper::paintAt(const KisPaintInformation &pi)
@@ -960,4 +1001,3 @@ void KisToolFreehandHelper::setCanvasHorizontalMirrorState(bool mirrored)
 {
    m_d->canvasMirroredH = mirrored;
 }
-
