@@ -172,10 +172,6 @@ KisPropertiesConfigurationSP KisTIFFOptions::toProperties() const
     compToIndex[COMPRESSION_JPEG] = 1;
     compToIndex[COMPRESSION_DEFLATE] = 2;
     compToIndex[COMPRESSION_LZW] = 3;
-    compToIndex[COMPRESSION_JP2000] = 4;
-    compToIndex[COMPRESSION_CCITTRLE] = 5;
-    compToIndex[COMPRESSION_CCITTFAX3] = 6;
-    compToIndex[COMPRESSION_CCITTFAX4] = 7;
     compToIndex[COMPRESSION_PIXARLOG] = 8;
 
     KisPropertiesConfigurationSP cfg = new KisPropertiesConfiguration();
@@ -186,7 +182,6 @@ KisPropertiesConfigurationSP KisTIFFOptions::toProperties() const
     cfg->setProperty("flatten", flatten);
     cfg->setProperty("quality", jpegQuality);
     cfg->setProperty("deflate", deflateCompress);
-    cfg->setProperty("faxmode", faxMode - 1);
     cfg->setProperty("pixarlog", pixarLogCompress);
     cfg->setProperty("saveProfile", saveProfile);
 
@@ -200,10 +195,9 @@ void KisTIFFOptions::fromProperties(KisPropertiesConfigurationSP cfg)
     indexToComp[1] = COMPRESSION_JPEG;
     indexToComp[2] = COMPRESSION_DEFLATE;
     indexToComp[3] = COMPRESSION_LZW;
-    indexToComp[4] = COMPRESSION_JP2000;
-    indexToComp[5] = COMPRESSION_CCITTRLE;
-    indexToComp[6] = COMPRESSION_CCITTFAX3;
-    indexToComp[7] = COMPRESSION_CCITTFAX4;
+    indexToComp[4] = COMPRESSION_PIXARLOG;
+
+    // old value that might be still stored in a config (remove after Krita 5.0 :) )
     indexToComp[8] = COMPRESSION_PIXARLOG;
 
     compressionType =
@@ -216,7 +210,6 @@ void KisTIFFOptions::fromProperties(KisPropertiesConfigurationSP cfg)
     flatten = cfg->getBool("flatten", true);
     jpegQuality = cfg->getInt("quality", 80);
     deflateCompress = cfg->getInt("deflate", 6);
-    faxMode = cfg->getInt("faxmode", 0) + 1;
     pixarLogCompress = cfg->getInt("pixarlog", 6);
     saveProfile = cfg->getBool("saveProfile", true);
 }
@@ -352,10 +345,13 @@ KisImageBuilder_Result KisTIFFConverter::readTIFFDirectory(TIFF* image)
 
     // Do not use the linear gamma profile for 16 bits/channel by default, tiff files are usually created with
     // gamma correction. XXX: Should we ask the user?
-    if (!profile && colorSpaceIdTag.first == RGBAColorModelID.id() && colorSpaceIdTag.second == Integer16BitsColorDepthID.id()) {
-        profile = KoColorSpaceRegistry::instance()->profileByName("sRGB-elle-V2-srgbtrc.icc");
-        dbgFile << "Getting srgb profile" << profile;
-     }
+    if (!profile) {
+        if (colorSpaceIdTag.first == RGBAColorModelID.id()) {
+            profile = KoColorSpaceRegistry::instance()->profileByName("sRGB-elle-V2-srgbtrc.icc");
+        } else if (colorSpaceIdTag.first == GrayAColorModelID.id()) {
+            profile = KoColorSpaceRegistry::instance()->profileByName("Gray-D50-elle-V2-srgbtrc.icc");
+        }
+    }
 
     // Retrieve a pointer to the colorspace
     const KoColorSpace* cs = 0;
@@ -384,12 +380,17 @@ KisImageBuilder_Result KisTIFFConverter::readTIFFDirectory(TIFF* image)
     int8 alphapos = -1; // <- no alpha
     // Check which extra is alpha if any
     dbgFile << "There are" << nbchannels << " channels and" << extrasamplescount << " extra channels";
-    if (sampleinfo) { // index images don't have any sampleinfo, and therefor sampleinfo == 0
+    if (sampleinfo) { // index images don't have any sampleinfo, and therefore sampleinfo == 0
         for (int i = 0; i < extrasamplescount; i ++) {
-            dbgFile << i << "" << extrasamplescount << ""  << (cs->colorChannelCount()) <<  nbchannels << "" << sampleinfo[i];
+            dbgFile << "sample" << i << "extra sample count" << extrasamplescount << "color channel count" << (cs->colorChannelCount()) << "Number of channels" <<  nbchannels << "sample info" << sampleinfo[i];
+            if (sampleinfo[i] == EXTRASAMPLE_UNSPECIFIED) {
+                qWarning() << "Extra sample type not defined for this file, assuming unassociated alpha.";
+                alphapos = i;
+            }
             if (sampleinfo[i] == EXTRASAMPLE_ASSOCALPHA) {
                 // XXX: dangelo: the color values are already multiplied with
                 // the alpha value.  This needs to be reversed later (postprocessor?)
+                qWarning() << "Associated alpha in this file: krita does not handle plremultiplied alpha.";
                 alphapos = i;
             }
 
@@ -515,10 +516,10 @@ KisImageBuilder_Result KisTIFFConverter::readTIFFDirectory(TIFF* image)
         uint16 position;
         TIFFGetFieldDefaulted(image, TIFFTAG_YCBCRPOSITIONING, &position);
         if (dstDepth == 8) {
-            tiffReader = new KisTIFFYCbCrReaderTarget8Bit(layer->paintDevice(), layer->image()->width(), layer->image()->height(), poses, alphapos, depth, sampletype, nbcolorsamples, extrasamplescount, transform, postprocessor, hsubsampling, vsubsampling, (KisTIFFYCbCr::Position)position);
+            tiffReader = new KisTIFFYCbCrReaderTarget8Bit(layer->paintDevice(), layer->image()->width(), layer->image()->height(), poses, alphapos, depth, sampletype, nbcolorsamples, extrasamplescount, transform, postprocessor, hsubsampling, vsubsampling);
         }
         else if (dstDepth == 16) {
-            tiffReader = new KisTIFFYCbCrReaderTarget16Bit(layer->paintDevice(), layer->image()->width(), layer->image()->height(), poses, alphapos, depth, sampletype, nbcolorsamples, extrasamplescount, transform, postprocessor, hsubsampling, vsubsampling, (KisTIFFYCbCr::Position)position);
+            tiffReader = new KisTIFFYCbCrReaderTarget16Bit(layer->paintDevice(), layer->image()->width(), layer->image()->height(), poses, alphapos, depth, sampletype, nbcolorsamples, extrasamplescount, transform, postprocessor, hsubsampling, vsubsampling);
         }
     }
     else if (dstDepth == 8) {
@@ -728,6 +729,7 @@ KisImageBuilder_Result KisTIFFConverter::buildFile(const QString &filename, KisI
         TIFFClose(image);
         return KisImageBuilder_RESULT_FAILURE;
     }
+
     KisTIFFWriterVisitor* visitor = new KisTIFFWriterVisitor(image, &options);
     if (!visitor->visit(root)) {
         TIFFClose(image);
